/*
 * YARK - Yet Another Real-time Kernel
 * Copyright (C) 1995 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	Author: Bryan Ford
 */
#ifndef _MACH_YARK_IPC_H_
#define _MACH_YARK_IPC_H_

#include <mach/machine/vm_types.h>
#include <mach/machine/yark/ipc.h>


/* Yark thread ID, opaque to general user processes.  */
typedef integer_t yark_tid_t;


/* Yark message descriptor, used by the message functions below.  */
typedef struct yark_msg
{
	/* Simple integer values passed in registers.
	   Machine-dependent code depends on this being first.  */
	integer_t	reg[YARK_REGS];

	/* Thread to send to/thread received from.  */
	yark_tid_t	tid;

	/* Larger data buffer passed by copying.
	   If none, buf_size is 0 and buf is dont-care.  */
	vm_size_t	buf_size;
	void *		buf;

} yark_msg_t;


/* Machine-independent IPC invocation routines.
   yark_call, yark_return, and yark_receive return recv->reg[0]
   as the function result as well as in the message buffer.  */
integer_t yark_call(yark_msg_t *send, yark_msg_t *recv);
integer_t yark_return(yark_msg_t *send, yark_msg_t *recv);
integer_t yark_receive(yark_msg_t *recv);
void yark_reply(yark_msg_t *send);


#endif _MACH_YARK_IPC_H_
