/*
 * YARK - Yet Another Real-time Kernel
 * Copyright (C) 1995 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	Author: Bryan Ford
 */
#ifndef _MACH_YARK_KMSG_H_
#define _MACH_YARK_KMSG_H_

/* Call and reply codes used in messages to and from the kernel.  */

/* Hardware interrupt dispatched to a thread.
   param1 = oid
   param2 = KC_INTERRUPT
   param3 = IRQ number (0-15).
   The interrupt is now disabled; it will be re-enabled on return.
   No data is provided; the leaked registers are trash.  */
#define KC_INTERRUPT	0x00000001

/* A trap occurred in the calling thread.
   param1 = oid
   param2 = trap number + KC_TRAP_BASE
   param3 = error code for traps that provide one (trash otherwise)
   param4 = linear address, if page fault (trash otherwise)
   On return, the faulting instruction will be restarted.  */
#define KC_TRAP_BASE	0x00000010
/* XXX this is really machine-dependent */


/* Reply codes used in kernel messages.  */
/* XXX use Mach codes */

/* The target thread has been terminated (or perhaps never existed).  */
#define KR_TARGET_DEAD		0x00000101

/* The target object has been terminated (or perhaps never existed).  */
#define KR_INVALID_OBJECT	0x00000102

#define KR_INTERRUPTED		0x00000103


#endif _MACH_YARK_KMSG_H_
