/*
 * Mach Operating System
 * Copyright (c) 1991,1990,1989 Carnegie Mellon University.
 * Copyright (c) 1993,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON, THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF
 * THIS SOFTWARE IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY
 * OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF
 * THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 *
 *      Utah $Hdr: def_pager_setup.c 1.1.1.1 94/09/09$
 */

#include <ufs_pager.h>
#if UFS_PAGER

#include <file_io.h>
/*#include <mach_interface.h>*/
#include <mach_user_internal.h>
#include <mach_port_internal.h>
#include <boot_printf.h>


#include <kern/kalloc.h>
#include <kern/thread.h>
#include <kern/task.h>
#include <kern/current.h>
#include <vm/vm_map.h>

/*
 * Set up default pager
 */
char	paging_file_name[128] = "\0";

extern void	create_default_partition();
decl_simple_lock_data(extern ,default_partition_lock);

extern char *strbuild();

void
default_pager_setup(master_host_port, master_device_port, rootname)
	mach_port_t master_host_port;
	mach_port_t master_device_port;
	char	*rootname;
{
	register struct file		*fp;
	register struct file_direct	*fdp;
	register int	result;

	simple_lock_init(&default_partition_lock);
	fp = (struct file *) kalloc(sizeof *fp);
	bzero((char *)fp, sizeof *fp);

#if	iPSC860 || PARAGON860
	{ extern char *	export_paging;

	(void) strbuild(paging_file_name,
			"/dev/",
			rootname,
			"/mach_servers/paging_file",
			export_paging,
			(char *)0);
	}
#else	/* iPSC860 || PARAGON860 */

	(void) strbuild(paging_file_name,
			"/dev/",
			rootname,
			"/mach_servers/paging_file",
			(char *)0);

#endif	/* iPSC860 || PARAGON860 */

	while (TRUE) {
	    result = open_file(master_device_port,
			       paging_file_name,
			       fp);
	    if (result == 0)
		break;
	    printf("Can't open paging file %s: %d\n",
		   paging_file_name,
		   result);

	    bzero(paging_file_name, sizeof(paging_file_name));
	    printf("Paging file name ? ");
	    gets(paging_file_name, sizeof(paging_file_name));

	    if (paging_file_name[0] == 0) {
		printf("*** WARNING: running without paging area!\n");
		return;
	    }
	}

	printf("Paging file %s found\n", paging_file_name);

	fdp = (struct file_direct *) kalloc(sizeof *fdp);
	bzero((char *)fdp, sizeof *fdp);

	result = open_file_direct(fp->f_dev, fdp);
	if (result)
	    panic("Can't open paging file\n");

	result = add_file_direct(fdp, fp);
	if (result)
	    panic("Can't read disk addresses: %d\n", result);

	/*
	 * Wire the buffers, since they will be used by the default pager.
	 * Set flag indicating that all implicitly created memory is
	 *	to be wired.
	 */
	current_task()->map->wiring_required = TRUE;

	result = file_wire_direct(master_host_port, fdp);
	if (result)
	    panic("Can't wire down file buffers: %d", result);

	/*
	 * Set up the default paging partition
	 */
	create_default_partition(fdp->fd_size * fdp->fd_fs->fs_bsize,
				 page_read_file_direct,
				 page_write_file_direct,
				 (char *)fdp);

	/*
	 * Our caller will become the default pager - later
	 */
}

#endif UFS_PAGER
