/* 
 * Copyright (c) 1987-1994, The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: dca.c 1.18 94/12/16$
 */

/*
 * Copyright (c) 1982, 1986 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)dca.c	7.1 (Berkeley) 6/5/86
 */

#include "dca.h"
#if NDCA > 0
/*
 *  Driver for National Semiconductor INS8250/NS16550AF/WD16C552 UARTs.
 *  Includes:
 *	98626/98644/internal serial interface on hp300/hp400
 *	internal serial ports on hp700
 *
 *  N.B. On the hp700, there is a "secret bit" with undocumented behavior.
 *  The third bit of the Modem Control Register (MCR_IEN == 0x08) must be
 *  set to enable interrupts.  If DEBUG, we'll check at various places to
 *  make sure it actually is enabled.
 */
#ifdef MACH_KERNEL
#include <sys/types.h>
#include <sys/ioctl.h>
#include <device/tty.h>
#include <device/errno.h>
#include <device/io_req.h>
#include <machine/machspl.h>
#include <mach_glue.h>	/* XXX */
#include <mach_kdb.h>
#else
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/tty.h>
#include <sys/conf.h>
#include <sys/systm.h>
#include <sys/uio.h>
#include <sys/syslog.h>
#endif
#ifdef hp700
#include <machine/asp.h>
#endif
#include <machine/cpu.h>
#ifdef MACH_KERNEL
#include <device.h>
#include <dcareg.h>
#else
#include <hpdev/device.h>
#include <hpdev/dcareg.h>
#endif

#ifndef DEFAULT_BAUD_RATE
#define DEFAULT_BAUD_RATE 9600
#endif

#ifndef CRTSCTS
#define CRTSCTS MDMBUF		/* XXX - hack waiting for 4.4 */
#endif

#ifdef KGDB_LAN		/* XXX */
#undef KGDB
#endif

int	dcaprobe();
struct	driver dcadriver = {
	dcaprobe, "dca",
};

int	dcastart(), dcaintr();
#ifdef	MACH_KERNEL
int	dcastop();
#endif
int	ttrstrt();
int	dcasoftCAR = 0;
int	dca_active = 0;
int	ndca = NDCA;
int	dcaconsole = -1;
int	dcadefaultrate = DEFAULT_BAUD_RATE;
int	dcadelay[NDCA] = { 0 };
struct	dcadevice *dca_addr[NDCA] = { 0 };
struct	tty dca_tty[NDCA] = { 0 };

int	dcadefaultfifotrig = FIFO_TRIGGER_14;
int	dca_hasfifo = 0;
int	dcafifot[NDCA];

#ifdef hp300
#include "machine/isr.h"
struct	isr dcaisr[NDCA];
int	dcaoflows[NDCA];
int	dcafastservice;
#ifdef KGDB
#define kgdb_break()	asm("trap #15;")
#endif
#endif

int	dca_speeds[] = {
	0, DCABRD(50), DCABRD(75), DCABRD(110), DCABRD(134), DCABRD(150),
	DCABRD(200), DCABRD(300), DCABRD(600), DCABRD(1200), DCABRD(1800),
	DCABRD(2400), DCABRD(4800), DCABRD(9600), DCABRD(19200), DCABRD(38400)
};

#ifdef KGDB
extern int kgdb_dev;
extern int kgdb_rate;
extern int kgdb_debug_init;
#ifdef MACH_KERNEL
#define KGDB_ON	{ extern int kgdb_enabled; kgdb_enabled = 1; }
#else
#define KGDB_ON 
#endif
#endif

#define	UNIT(x)		(x & 0x7)
#define DCA_NOFIFODEV(x) ((x) & 0x10)

#ifdef MMAP
#include "user.h"
#include "proc.h"
#define DCA_MAPDEV(x)	(x & 0x80)
struct	dcammap {
	int	isopen;
	struct	proc *proc;
} dcammap[NDCA];
#endif

#ifdef DEBUG
long	fifoin[17];
long	fifoout[17];
long	dcaintrcount[16];
long	dcamintcount[16];
long	mcrintclear[NDCA][16];
#endif

#if defined(hp700) && defined(DEBUG)
#define CHECKMCRINT(ix, u, dca) \
	if (((dca)->dca_mcr & MCR_IEN) == 0) { \
		mcrintclear[u][ix]++; \
		(dca)->dca_mcr |= MCR_IEN; \
	}
#else
#define CHECKMCRINT(ix, u, dca)
#endif

static int
rate2speed(baudrate)
	register int baudrate;
{
	static struct r2s {
		int	r;
		int	s;
	} rates[] = {
		{38400, EXTB}, {19200, EXTA}, {9600, B9600}, {4800, B4800},
		{2400, B2400}, {1200, B1200}, {600, B600}, {300, B300},
		{200, B200}, {150, B150}, {134, B134}, {110, B110},
		{75, B75}, {50, B50}, {0, B0}
	};
	register struct r2s *rp = rates;

	while (baudrate < rp->r)
		++rp;

	return (rp->s);
}

dcaprobe(hd)
	register struct hp_device *hd;
{
	register struct dcadevice *dca;
	register int unit;

	dca = (struct dcadevice *)hd->hp_addr;
#ifdef hp300
	if (badbaddr((char *)&dca->dca_id) ||
	    (dca->dca_id != DCAID0 && dca->dca_id != DCAREMID0 &&
	     dca->dca_id != DCAID1 && dca->dca_id != DCAREMID1))
		return (0);
#endif
	unit = hd->hp_unit;
	if (unit == dcaconsole)
		DELAY(100000);
#ifdef hp300
	dca->dca_reset = 0xFF;
	DELAY(100);
#endif

	/* look for a NS 16550AF UART with FIFOs */
	dcafifot[unit] = dcadefaultfifotrig;
	dca->dca_fifo = FIFO_ENABLE|FIFO_RCV_RST|FIFO_XMT_RST|dcafifot[unit];
	DELAY(100);
	if ((dca->dca_iir & IIR_FIFO_MASK) == IIR_FIFO_MASK)
		dca_hasfifo |= 1 << unit;

	dca_addr[unit] = dca;
#ifdef hp300
	hd->hp_ipl = DCAIPL(dca->dca_ic);
	dcaisr[unit].isr_ipl = hd->hp_ipl;
	dcaisr[unit].isr_arg = unit;
	dcaisr[unit].isr_intr = dcaintr;
	isrlink(&dcaisr[unit]);
#endif
	dca_active |= 1 << unit;
	if (hd->hp_flags)
		dcasoftCAR |= (1 << unit);
#if defined(KGDB) && !defined(MACH_KERNEL)
	if (kgdb_dev == makedev(1, unit)) {
		if (dcaconsole == unit)
			kgdb_dev = -1;	/* can't debug over console port */
		else {
			dcainit(unit, kgdb_rate);
			printf("dca%d: kgdb at %d baud ", unit, kgdb_rate);
			if (kgdb_debug_init) {
				printf("waiting...");
				/* trap into kgdb */
				kgdb_break();
				printf("connected.\n");
			} else
				printf("enabled\n");
		}
	}
#endif
#ifdef hp300
	dca->dca_ic = IC_IE;
#endif
	/*
	 * May need to reset baud rate, etc. of console in case console
	 * defaults are different.  Also make sure console is always
	 * "hardwired".
	 */
	if (unit == dcaconsole) {
		dcainit(unit, dcadefaultrate);
		dcasoftCAR |= (1 << unit);
	}
	return (1);
}

#ifdef	MACH_KERNEL
dcaopen(dev, flag, ior)
	dev_t dev;
	int flag;
	io_req_t ior;
#else
dcaopen(dev, flag)
	dev_t dev;
	int flag;
#endif
{
	register struct tty *tp;
	register int unit;
#ifdef hp300
	int error;
#endif
 
	unit = UNIT(dev);
	if (unit >= NDCA || (dca_active & (1 << unit)) == 0)
		return (ENXIO);
	tp = &dca_tty[unit];
#ifdef MMAP
	/*
	 * Disallow open if device is already opened for mapping
	 */
	if (dcammap[unit].isopen)
		return (EBUSY);
	/*
	 * If this is the mapped device:
	 *   disallow if already open or the console device
	 *   disable interrupts
	 */
	if (DCA_MAPDEV(dev)) {
		int s;

		if ((tp->t_state & TS_ISOPEN) || unit == dcaconsole)
			return (EBUSY);
		s = spltty();
#ifdef hp300
		dca_addr[unit]->dca_ic &= ~IC_IE;	/* XXX needed? */
#endif
		dca_addr[unit]->dca_ier = 0;
		dcammap[unit].isopen = 1;
		splx(s);
		return(0);
	}
#endif
	tp->t_oproc = dcastart;
#ifdef	MACH_KERNEL
	tp->t_stop = dcastop;
#endif
	if ((tp->t_state & TS_ISOPEN) == 0) {
		ttychars(tp);
		tp->t_state |= TS_HUPCLS;
		tp->t_ispeed = rate2speed(dcadefaultrate);
		tp->t_ospeed = rate2speed(dcadefaultrate);
		tp->t_flags = (EVENP | ODDP | ECHO);
		if (unit == dcaconsole)
			tp->t_flags |= XTABS|CRMOD;
#ifndef	MACH_KERNEL
		if (DCA_NOFIFODEV(dev))
			dcafifot[unit] = FIFO_TRIGGER_1;
		else
			dcafifot[unit] = dcadefaultfifotrig;
#endif
		dcaparam(unit);
	}
#ifndef	MACH_KERNEL
	else {
		if (tp->t_state&TS_XCLUDE && !suser())
			return (EBUSY);
		/*
		 * Can only allow one of the fifo/nofifo devices to be open
		 * at the same time.
		 */
		if (DCA_NOFIFODEV(dev) && dcafifot[unit] != FIFO_TRIGGER_1 ||
		    !DCA_NOFIFODEV(dev) && dcafifot[unit] == FIFO_TRIGGER_1)
			return (EBUSY);
	}
#endif
	(void) dcamctl(dev, MCR_DTR | MCR_RTS, DMSET);
	if ((dcasoftCAR & (1 << unit)) || (dcamctl(dev, 0, DMGET) & MSR_DCD))
		tp->t_state |= TS_CARR_ON;
#ifdef	MACH_KERNEL
	return (char_open(dev, tp, flag, ior));
#else
	(void) spltty();
	while ((tp->t_state & TS_CARR_ON) == 0) {
		tp->t_state |= TS_WOPEN;
		sleep((caddr_t)&tp->t_rawq, TTIPRI);
	}
	(void) spl0();
#ifdef hp300
	error = (*linesw[tp->t_line].l_open)(dev, tp);
	/*
	 * XXX hack to speed up unbuffered builtin port.
	 * If dca_fastservice is set, a level 5 interrupt
	 * will be directed to dcaintr first.
	 */
	if (error == 0 && unit == 0 && (dca_hasfifo & 1) == 0)
		dcafastservice = 1;
	return (error);
#else
	return ((*linesw[tp->t_line].l_open)(dev, tp));
#endif
#endif
}
 
dcaclose(dev, flag)
	dev_t dev;
{
	register struct tty *tp;
	register struct dcadevice *dca;
	register int unit;
 
	unit = UNIT(dev);
#ifdef hp300
	if (unit == 0)
		dcafastservice = 0;
#endif
	dca = dca_addr[unit];
#ifdef MMAP
	/*
	 * If this is the mapped interface, then we haven't gone
	 * through the normal tty open sequence.  All we need to
	 * do is ensure that interrupts are cleared (but enabled).
	 */
	if (DCA_MAPDEV(dev)) {
		int s = spltty();

		dcammap[unit].isopen = 0;
		dcammap[unit].proc = 0;
		dca->dca_ier = 0;
#ifdef hp300
		dca->dca_ic = IC_IE;	/* XXX needed? */
#endif
		splx(s);
		return (0);
	}
#endif
	tp = &dca_tty[unit];
#ifdef	MACH_KERNEL
	(void) spltty();
	simple_lock(&tp->t_lock);
#else
	(*linesw[tp->t_line].l_close)(tp);
#endif
	dca->dca_cfcr &= ~CFCR_SBREAK;
#ifdef KGDB
	/* do not disable interrupts if debugging */
	if (kgdb_dev != makedev(1, unit))
#endif
		dca->dca_ier = 0;
	(void) dcamctl(dev, 0, DMSET);
	if (tp->t_state & TS_HUPCLS)
		(*linesw[tp->t_line].l_modem)(tp, 0);
	ttyclose(tp);
#ifdef	MACH_KERNEL
	simple_unlock(&tp->t_lock);
	(void) spl0();
#endif
	return (0);
}
 
dcaread(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	int unit = UNIT(dev);
	register struct tty *tp = &dca_tty[unit];
#ifdef hp300
	int error;
	int of = dcaoflows[unit];
#endif
 
#ifdef MMAP
	if (DCA_MAPDEV(dev))
		return (EIO);
#endif
#ifdef hp300
	error = (*linesw[tp->t_line].l_read)(tp, uio);
	/*
	 * XXX hardly a reasonable thing to do, but reporting overflows
	 * at interrupt time just exacerbates the problem.
	 */
	if (dcaoflows[unit] != of)
		log(LOG_WARNING, "dca%d: silo overflow\n", unit);
	return (error);
#else
	return ((*linesw[tp->t_line].l_read)(tp, uio));
#endif
}
 
dcawrite(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	int unit = UNIT(dev);
	register struct tty *tp = &dca_tty[unit];
 
#ifdef MMAP
	if (DCA_MAPDEV(dev))
		return (EIO);
#endif
#ifndef	MACH_KERNEL
	/*
	 * (XXX) We disallow virtual consoles if the physical console is
	 * a serial port.  This is in case there is a display attached that
	 * is not the console.  In that situation we don't need/want the X
	 * server taking over the console.
	 */
	{
		extern struct tty *constty;

		if (constty && unit == dcaconsole)
			constty = NULL;
	}
#endif
	return ((*linesw[tp->t_line].l_write)(tp, uio));
}
 
/*
 * Map the registers.  We assume that one page is enough.
 */
dcamap(dev, off, prot)
	dev_t dev;
	register int off;
{
#ifdef MMAP
	int unit = UNIT(dev);

	if (!DCA_MAPDEV(dev) || off != 0)
		return (EINVAL);
	if (dcammap[unit].proc == 0)
		dcammap[unit].proc = u.u_procp;
	return (kvtop((u_int)dca_addr[unit]) >> PGSHIFT);
#endif
}

dcaintr(unit)
	register int unit;
{
	register struct dcadevice *dca;
	register u_char code;
	register struct tty *tp;

	dca = dca_addr[unit];
#ifdef hp300
	if ((dca->dca_ic & (IC_IR|IC_IE)) != (IC_IR|IC_IE))
		return (0);
#endif
#ifdef MMAP
	if (dcammap[unit].proc) {
#ifdef hp300
		dca->dca_ic &= ~IC_IE;
#else
		dca->dca_ier = 0;
#endif
		psignal(dcammap[unit].proc, SIGUSR2);
		return(1);
	}
#endif
	tp = &dca_tty[unit];
	while (1) {
		code = dca->dca_iir;
#ifdef DEBUG
		dcaintrcount[code & IIR_IMASK]++;
#endif
		switch (code & IIR_IMASK) {
		case IIR_NOPEND:
			CHECKMCRINT(0, unit, dca);
			return (1);
		case IIR_RXTOUT:
		case IIR_RXRDY:
			/* do time-critical read in-line */
/*
 * Process a received byte.  Inline for speed...
 */
#ifdef KGDB
#define	RCVBYTE() \
			code = dca->dca_data; \
			if ((tp->t_state & TS_ISOPEN) == 0) { \
				if (kgdb_dev == makedev(1, unit) && \
				    (code == '!' || code == '\003')) { \
					printf("kgdb trap from dca%d\n", unit); \
					KGDB_ON; \
					kgdb_break(); \
				} \
			} else \
				(*linesw[tp->t_line].l_rint)(code, tp)
#else
#define	RCVBYTE() \
			code = dca->dca_data; \
			if ((tp->t_state & TS_ISOPEN) != 0) \
				(*linesw[tp->t_line].l_rint)(code, tp)
#endif
			RCVBYTE();
			if (dca_hasfifo & (1 << unit)) {
#ifdef DEBUG
				register int fifocnt = 1;
#endif
				while ((code = dca->dca_lsr) & LSR_RCV_MASK) {
					if (code == LSR_RXRDY) {
						RCVBYTE();
					} else
						dcaeint(unit, code, dca);
#ifdef DEBUG
					fifocnt++;
#endif
				}
				CHECKMCRINT(1, unit, dca);
#ifdef DEBUG
				if (fifocnt > 16)
					fifoin[0]++;
				else
					fifoin[fifocnt]++;
#endif
			}
			break;
		case IIR_TXRDY:
			tp->t_state &=~ (TS_BUSY|TS_FLUSH);
			if (tp->t_line)
				(*linesw[tp->t_line].l_start)(tp);
			else
				dcastart(tp);
			break;
		case IIR_RLS:
			dcaeint(unit, dca->dca_lsr, dca);
			CHECKMCRINT(2, unit, dca);
			break;
		default:
			if (code & IIR_NOPEND)
				return (1);
			log(LOG_WARNING, "dca%d: weird interrupt: 0x%x\n",
			    unit, code);
			/* fall through */
		case IIR_MLSC:
			dcamint(unit, dca);
			break;
		}
	}
	/*NOTREACHED*/
}

dcaeint(unit, stat, dca)
	register int unit, stat;
	register struct dcadevice *dca;
{
	register struct tty *tp;
	register int c;

	tp = &dca_tty[unit];
	c = dca->dca_data;
	if ((tp->t_state & TS_ISOPEN) == 0) {
#ifdef KGDB
		/* we don't care about parity errors */
		if (((stat & (LSR_BI|LSR_FE|LSR_PE)) == LSR_PE) &&
		    kgdb_dev == makedev(1, unit) && (c == '!' || c == '\003')) {
			printf("kgdb trap from dca%d\n", unit);
			/* trap into kgdb */
			kgdb_break();
		}
#endif
#if	MACH_KDB
		if (stat & LSR_BI)
			kdb_kintr();
#endif
		return;
	}
	if (stat & (LSR_BI | LSR_FE)) {
#ifdef	MACH_KERNEL
		c = tp->t_breakc;
#else
		if (tp->t_flags & RAW)
			c = 0;
		else
			c = tp->t_intrc;
#endif
	}
	else if (stat & LSR_PE) {
		if (((tp->t_flags & (EVENP|ODDP)) == EVENP)
		    || ((tp->t_flags & (EVENP|ODDP)) == ODDP))
			return;
	}
	else if (stat & LSR_OE) {
#ifdef hp300
		dcaoflows[unit]++;
#else
		log(LOG_WARNING, "dca%d: silo overflow\n", unit);
#endif
	}
	(*linesw[tp->t_line].l_rint)(c, tp);
}

dcamint(unit, dca)
	register int unit;
	register struct dcadevice *dca;
{
	register struct tty *tp;
	register u_char stat;

	tp = &dca_tty[unit];
	stat = dca->dca_msr;
#ifdef DEBUG
	dcamintcount[stat & 0xf]++;
#endif
	if ((stat & MSR_DDCD) && (dcasoftCAR & (1 << unit)) == 0) {
		if (stat & MSR_DCD)
			(void)(*linesw[tp->t_line].l_modem)(tp, 1);
		else if ((*linesw[tp->t_line].l_modem)(tp, 0) == 0)
			dca->dca_mcr &= ~(MCR_DTR | MCR_RTS);
	} else if ((stat & MSR_DCTS) && (tp->t_state & TS_ISOPEN) &&
		   (tp->t_flags & CRTSCTS)) {
		/* the line is up and we want to do rts/cts flow control */
		if (stat & MSR_CTS) {
			tp->t_state &=~ TS_TTSTOP;
			ttstart(tp);
		} else
			tp->t_state |= TS_TTSTOP;
	}
}

#ifdef	MACH_KERNEL
int
dcareset()
{
}

int
dcaportdeath(dev, port)
	dev_t   dev;
	mach_port_t port;
{
	return (tty_portdeath(&dca_tty[minor(dev)], port));
}

io_return_t dcagetstat(dev, flavor, data, count)
	dev_t		dev;
	int		flavor;
	int *		data;		/* pointer to OUT array */
	unsigned int	*count;		/* out */
{
	register struct tty *tp;
	register int	unit = minor(dev);

	tp = &dca_tty[unit];

	switch (flavor) {
	    case TTY_MODEM:
		break;

	    default:
		return (tty_get_status(tp, flavor, data, count));
	}
	return (D_SUCCESS);
}

io_return_t dcasetstat(dev, flavor, data, count)
	dev_t		dev;
	int		flavor;
	int *		data;
	unsigned int	count;
{
	register struct tty *tp;
	register int	unit = minor(dev);
	register struct dcadevice *dca;
	int error = 0;

	tp = &dca_tty[unit];
	dca = dca_addr[unit];

	switch (flavor) {
	    case TTY_MODEM:

	    case TTY_SET_BREAK:
		dca->dca_cfcr |= CFCR_SBREAK;
		break;

	    case TTY_CLEAR_BREAK:
		dca->dca_cfcr &= ~CFCR_SBREAK;
		break;

	    case TTY_STATUS:
		error = tty_set_status(tp, flavor, data, count);
		/*
		 * Change HW characteristics.
		 * XXX don't change console modes, dubious but consistent
		 *     with Acorn.
		 */
		if (error == 0 && unit != dcaconsole)
			dcaparam(unit);
		return (error);

	    default:
		return (tty_set_status(tp, flavor, data, count));
	}
	return (D_SUCCESS);
}
#else
dcaioctl(dev, cmd, data, flag)
	dev_t dev;
	caddr_t data;
{
	register struct tty *tp;
	register int unit = UNIT(dev);
	register struct dcadevice *dca;
	register int error;
 
#ifdef MMAP
	if (DCA_MAPDEV(dev))
		return (EIO);
#endif
	tp = &dca_tty[unit];
	error = (*linesw[tp->t_line].l_ioctl)(tp, cmd, data, flag);
	if (error >= 0)
		return (error);
	error = ttioctl(tp, cmd, data, flag);
	if (error >= 0) {
		if (cmd == TIOCSETP || cmd == TIOCSETN ||
		    cmd == TIOCLBIS || cmd == TIOCLBIC ||
		    cmd == TIOCLSET) {
			dcaparam(unit);
#ifdef hp700
			/*
			 * If something is doing these things back-to-back
			 * and then sets FLUSHO, the hp700 can fiddle with
			 * the DCA so quickly that we may never see another
			 * RXRDY (rcvd data) due to the higher-priority
			 * TXRDY (transmitter empty) interrupts.
			 */
			if (tp->t_flags & FLUSHO)
				DELAY(100000);
#endif
		}
		return (error);
	}
	dca = dca_addr[unit];
	switch (cmd) {

	case TIOCSBRK:
		dca->dca_cfcr |= CFCR_SBREAK;
		break;

	case TIOCCBRK:
		dca->dca_cfcr &= ~CFCR_SBREAK;
		break;

	case TIOCSDTR:
		(void) dcamctl(dev, MCR_DTR | MCR_RTS, DMBIS);
		break;

	case TIOCCDTR:
		(void) dcamctl(dev, MCR_DTR | MCR_RTS, DMBIC);
		break;

	case TIOCMSET:
		(void) dcamctl(dev, *(int *)data, DMSET);
		break;

	case TIOCMBIS:
		(void) dcamctl(dev, *(int *)data, DMBIS);
		break;

	case TIOCMBIC:
		(void) dcamctl(dev, *(int *)data, DMBIC);
		break;

	case TIOCMGET:
		*(int *)data = dcamctl(dev, 0, DMGET);
		break;

	default:
		return (ENOTTY);
	}
	return (0);
}
#endif

dcaparam(unit)
	register int unit;
{
	register struct tty *tp;
	register struct dcadevice *dca;
	register int mode = 0;
	register int s = spltty();
 
	tp = &dca_tty[unit];
	dca = dca_addr[unit];
	dca->dca_fifo = 0;	/* clear FIFOs so we dont lose TXRDY intrs */
	dca->dca_ier = IER_ERXRDY | IER_ETXRDY | IER_ERLS | IER_EMSC;
	CHECKMCRINT(3, unit, dca);
#ifdef hp700
	dca->dca_mcr |= MCR_IEN;
#endif
	if (tp->t_ispeed == 0) {
		(void) dcamctl(unit, 0, DMSET);	/* hang up line */
		splx(s);
		return;
	}
	dca->dca_cfcr |= CFCR_DLAB;
	dca->dca_data = dca_speeds[tp->t_ispeed] & 0xFF;
	dca->dca_ier = dca_speeds[tp->t_ispeed] >> 8;
	if (tp->t_flags & (RAW|LITOUT|PASS8))
		mode |= CFCR_8BITS;
	else {
		mode |= CFCR_7BITS | CFCR_PENAB;
		if ((tp->t_flags & EVENP) == 0)
			mode |= CFCR_PODD;
		else if ((tp->t_flags & ODDP) == 0)
			mode |= CFCR_PEVEN;
		else
			mode |= CFCR_PZERO;
	}
	if (tp->t_ispeed == B110)
		mode |= CFCR_STOPB;
	dca->dca_cfcr = mode;
	if (dca_hasfifo & (1 << unit))
		dca->dca_fifo = FIFO_ENABLE | dcafifot[unit];
	splx(s);
}
 
dcastart(tp)
	register struct tty *tp;
{
	register struct dcadevice *dca;
	int s, unit, c;
 
	unit = UNIT(tp->t_dev);
	dca = dca_addr[unit];
	s = spltty();
	if (tp->t_state & (TS_TIMEOUT|TS_TTSTOP))
		goto out;
	if (tp->t_outq.c_cc <= TTLOWAT(tp)) {
#ifdef	MACH_KERNEL
		tt_write_wakeup(tp);
#else
		if (tp->t_state&TS_ASLEEP) {
			tp->t_state &= ~TS_ASLEEP;
			wakeup((caddr_t)&tp->t_outq);
		}
		selwakeup(&tp->t_wsel);
#endif
	}
	if (tp->t_outq.c_cc == 0)
		goto out;
#ifdef CCA_PAGE
	if (!(tp->t_flags & (RAW|CBREAK|LITOUT)) &&
	    tp->t_pagemode && (firstc(&tp->t_outq) & 0177) == '\n') {
		ttpage(tp);
		if (tp->t_state & TS_TTSTOP)
			goto out;
	}
#endif CCA_PAGE
	if (dca->dca_lsr & LSR_TXRDY) {
		c = getc(&tp->t_outq);
		tp->t_state |= TS_BUSY;
		dca->dca_data = c;
		if (dca_hasfifo & (1 << unit)) {
			for (c = 1; c < 16 && tp->t_outq.c_cc; ++c)
				dca->dca_data = getc(&tp->t_outq);
#ifdef DEBUG
			if (c > 16)
				fifoout[0]++;
			else
				fifoout[c]++;
#endif
		}
	}
	CHECKMCRINT(4, unit, dca);
out:
	CHECKMCRINT(5, unit, dca);
	splx(s);
}
 
/*
 * Stop output on a line.
 */
dcastop(tp, flag)
	register struct tty *tp;
{
	register int s;

	s = spltty();
	if (tp->t_state & TS_BUSY) {
		if ((tp->t_state&TS_TTSTOP)==0)
			tp->t_state |= TS_FLUSH;
	}
	splx(s);
}
 
dcamctl(dev, bits, how)
	dev_t dev;
	int bits, how;
{
	int unit = UNIT(dev);
	register struct dcadevice *dca = dca_addr[unit];
	int s;

#ifdef hp700
	/*
	 * Always make sure MCR_IEN is set (unless setting to 0)
	 */
#ifdef KGDB
	if (how == DMSET && kgdb_dev == makedev(1, unit))
		bits |= MCR_IEN;
	else
#endif
	if (how == DMBIS || (how == DMSET && bits))
		bits |= MCR_IEN;
	else if (how == DMBIC)
		bits &= ~MCR_IEN;
#endif
	s = spltty();
	CHECKMCRINT(6, unit, dca);
	switch (how) {

	case DMSET:
		dca->dca_mcr = bits;
		break;

	case DMBIS:
		dca->dca_mcr |= bits;
		break;

	case DMBIC:
		dca->dca_mcr &= ~bits;
		break;

	case DMGET:
		bits = dca->dca_msr;
		break;
	}
	(void) splx(s);
	return(bits);
}

/*
 * Following are all routines needed for DCA to act as console
 */
#ifdef MACH_KERNEL
#include <cons.h>
#else
#include <hpdev/cons.h>
#endif
#ifdef hp700
#include <machine/iomod.h>
#endif

dcacnprobe(cp)
	struct consdev *cp;
{
	int unit, maj;
	extern int dcaopen();

	/* XXX: ick */
	unit = CONUNIT;
#ifdef hp300
	dca_addr[CONUNIT] = (struct dcadevice *) sctova(CONSCODE);

	/* make sure hardware exists */
	if (badbaddr((char *)&dca_addr[unit]->dca_id)) {
		cp->cn_pri = CN_DEAD;
		return;
	}
#endif
#ifdef hp700
	{
		extern struct iodc_data cons_iodc;	/* defined in... */
		extern struct device_path cons_dp;	/* vm_machdep.c */
		extern int isgecko;
		struct modtab *m, *getmod();

#if 0
/*
 * There is no way to tell the gecko boot rom to use the RS232 port as
 * console.  Enable this code if you want to force the kernel's console
 * to the RS232 port.
 */
		m = getmod(IODC_TP_FIO, SVMOD_FIO_GRS232, 0);
		if (m) {
			dca_addr[0] = GPORT1;
			unit = 0;
			cp->cn_pri = CN_REMOTE;
		} else
#endif
		/*
		 * If the console we booted on was an RS232 port, then
		 * keep it as our console (CN_REMOTE).  Otherwise the
		 * DCA is just another console possibility (CN_NORMAL)
		 * and we choose the first one found.
		 */
		if (cons_iodc.iodc_sv_model == SVMOD_FIO_RS232) {
			if (cons_dp.dp_mod == 5) {	/* XXX? */
				dca_addr[1] = PORT2;
				unit = 1;
			} else {
				dca_addr[0] = PORT1;
				unit = 0;
			}
			cp->cn_pri = CN_REMOTE;
		} else if (cons_iodc.iodc_sv_model == SVMOD_FIO_GRS232) {
			dca_addr[0] = GPORT1;
			unit = 0;
			cp->cn_pri = CN_REMOTE;
		} else {
			m = getmod(IODC_TP_FIO, SVMOD_FIO_RS232, 0);
			if (m == 0)
				m = getmod(IODC_TP_FIO, SVMOD_FIO_GRS232, 0);
			if (m) {
				unit = (m->m_fixed == 5) ? 1 : 0;
				dca_addr[unit] = (struct dcadevice *)m->m_hpa;
				cp->cn_pri = CN_NORMAL;
			} else
				cp->cn_pri = CN_DEAD;
		}
	}
#endif
#ifdef MACH_KERNEL
	/* No major number of because of indirection table */
	maj = 0;
#else
	/* locate the major number */
	for (maj = 0; maj < nchrdev; maj++)
		if (cdevsw[maj].d_open == dcaopen)
			break;
#endif
	/* initialize required fields */
	cp->cn_dev = makedev(maj, unit);
#ifdef hp300
	switch (dca_addr[unit]->dca_id) {
	case DCAID0:
	case DCAID1:
		cp->cn_pri = CN_NORMAL;
		break;
	case DCAREMID0:
	case DCAREMID1:
		cp->cn_pri = CN_REMOTE;
		break;
	default:
		cp->cn_pri = CN_DEAD;
		break;
	}
#endif
}

dcacninit(cp)
	struct consdev *cp;
{
	int unit;

	unit = UNIT(cp->cn_dev);
	dcainit(unit, dcadefaultrate);
	dcaconsole = unit;
	return(0);
}

dcainit(unit, ubrate)
	int unit, ubrate;
{
	register struct dcadevice *dca;
	int s;
	u_char stat;
	int rate = DCABRD(ubrate);

	dca = dca_addr[unit];
	dcafifot[unit] = dcadefaultfifotrig;
	dcadelay[unit] = 10000000 / ubrate;
	s = splhigh();
#ifdef hp300
	dca->dca_reset = 0xFF;
	DELAY(100);
	dca->dca_ic = IC_IE;
#endif
	dca->dca_cfcr = CFCR_DLAB;
	dca->dca_data = rate & 0xFF;
	dca->dca_ier = rate >> 8;
	dca->dca_cfcr = CFCR_8BITS;
	dca->dca_ier = IER_ERXRDY | IER_ETXRDY;
#ifdef hp700
	dca->dca_mcr |= MCR_IEN;
#endif
	dca->dca_fifo = FIFO_ENABLE|FIFO_RCV_RST|FIFO_XMT_RST|dcafifot[unit];
	DELAY(100);
	stat = dca->dca_iir;
	splx(s);
}

dcacngetc(dev, wait)
	dev_t dev;
	int wait;
{
	register struct dcadevice *dca = dca_addr[UNIT(dev)];
	register u_char stat;
	int c, s;

	s = splhigh();
	while (((stat = dca->dca_lsr) & LSR_RXRDY) == 0)
		if (!wait) {
			splx(s);
			return(0);
		}
	c = dca->dca_data;
	stat = dca->dca_iir;
	CHECKMCRINT(7, UNIT(dev), dca);
	splx(s);
	return(c);
}

/*
 * Console kernel output character routine.
 * XXX assumes that DELAY is working.
 */
dcacnputc(dev, c)
	dev_t dev;
	register int c;
{
	register struct dcadevice *dca;
	register int timo;
	register u_char stat;
	int unit = UNIT(dev);
	int s = splhigh();

#ifdef KGDB
	if (dev == kgdb_dev) {
		/*
		 * In case someone sets kgdb_dev after boot time.
		 */
		if (kgdb_debug_init == 0) {
			dcainit(unit, kgdb_rate);
			kgdb_debug_init = 1;
		}
	} else
#endif
	if (dcaconsole == -1) {
#ifdef DEBUG
		dcainit(unit, dcadefaultrate);
		panic("yow! cnputc called before dcaconsole set\n");
#endif
		return;
	}
	dca = dca_addr[unit];
	/* wait one character time for any pending transmission to finish */
	timo = dcadelay[unit];
	while (((stat = dca->dca_lsr) & LSR_TXRDY) == 0 && --timo)
		DELAY(1);
	dca->dca_data = c;
	/* wait for this transmission to complete */
	timo = dcadelay[unit];
	while (((stat = dca->dca_lsr) & LSR_TXRDY) == 0 && --timo)
		DELAY(1);
	/*
	 * If the "normal" interface was busy transfering a character
	 * we must let our interrupt through to keep things moving.
	 * Otherwise, we clear the interrupt that we have caused.
	 */
	if ((dca_tty[unit].t_state & TS_BUSY) == 0)
		stat = dca->dca_iir;
	CHECKMCRINT(8, unit, dca);
	splx(s);
}
#endif
