/* 
 * Copyright (c) 1994, The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: gkd.c 1.6 94/12/16$
 */
#include "gkd.h"
#if NGKD > 0
/*
 *  Driver for IBM PS2 compatible keyboard/mouse driver
 */
#ifdef MACH_KERNEL
#include <sys/types.h>
#include <sys/ioctl.h>
#include <device/tty.h>
#include <device/cirbuf.h>
#include <device/errno.h>
#include <device/io_req.h>
#include <mach_glue.h>	/* XXX */
#include <mach_kdb.h>
#include <device.h>
#include <gkdreg.h>
#include <gkdvar.h>
#include <gkdhpux.h>
#include <itevar.h>
#include <kbdmap.h>
#include <machine/machspl.h>
#else
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/tty.h>
#include <sys/conf.h>
#include <sys/systm.h>
#include <sys/uio.h>
#include <sys/syslog.h>
#include <sys/fcntl.h>
#include <sys/user.h>
#include <hpdev/device.h>
#include <hpsgc/gkdreg.h>
#include <hpsgc/gkdvar.h>
#include <hpsgc/gkdhpux.h>
#include <hpdev/itevar.h>
#include <hpdev/kbdmap.h>
#include <machine/spl.h>
#endif
#ifdef hp700
#include <machine/asp.h>
#endif
#include <machine/cpu.h>

int	gkdprobe();
int	gkdstart(), gkdintr();

/*
 * table describing software view of the hardware
 * Note, at this time, we only support 1 ps2 interface 
 * with exactly one mouse and one keyboard
 */
struct	gkd_dev	gkd_device[GKD_MAX_PORTS];
static  int  gkd_kbd_index, gkd_mouse_index; 

/*
 * Commands that we can issue to devices
 */
static char gkd_kbd_set_leds[] = { 2, 0, GKD_KBD_LEDS, 0};
static char gkd_mouse_stream[] = { 1, 0, 0xEA};
static char gkd_mouse_resolution[] = { 2, 0, 0xE8, 0};
static char gkd_device_ident[] = { 1, 2, 0xF2};
static char gkd_device_disable[] = { 1, 0, 0xF5};
static char gkd_device_enable[] = { 1, 0, 0xF4};
static char gkd_device_reset[] = { 1, 0, 0xFF};
static char gkd_device_rate[] = { 2, 0, 0xF3, 0};
static char gkd_device_setdefault[] = { 1, 0, 0xF6 };
static char gkd_kbd_scancode[] = { 2, 0, 0xF0, 0};
static char gkd_kbd_all_tmat_mkbrk[] = { 1, 0, 0xFA };
static char gkd_kbd_all_tmat[] = { 1, 0, 0xF7 };
static char gkd_kbd_all_mkbrk[] = { 1, 0, 0xF8 };
static char gkd_kbd_all_mk[] = { 1, 0, 0xF9 };
static char gkd_mouse_promptmode[] = { 1, 0, 0xF0 };

/*
 * Probe to see if we have a PS/2 style keyboard and mouse
 */
gkdprobe(hd)
	register struct hp_device *hd;
{
	struct gkd_reg *regs = (struct gkd_reg *) hd->hp_addr;

	if (regs->id_reset == 0 && gkd_next(regs)->id_reset ==1) {
	  return(1);
	}

	return (0);
}

/*
 * Attach the keyboard
 */
gkdattach(hd)
	register struct hp_device *hd;
{
        extern int asp2procibit[], gkdintr();
	int gkd_kbd_led();
	int ibit;

	gkd_device[0].regs =	(struct gkd_reg *) hd->hp_addr;
	gkd_device[1].regs =	(struct gkd_reg *) gkd_next(hd->hp_addr);

	gkd_kbd_index = 0;      /* XXX we're hard coding this */
	gkd_mouse_index = 1;	/* XXX we're hard coding this */

	gkd_init(gkd_device);
	gkd_config(&gkd_device[gkd_kbd_index]);

	ibit = aspitab(INT_PS2, SPLHIL, gkdintr, 0, 0);
}

/*
 * initialize this interface
 */
gkd_init(dev)
	struct gkd_dev *dev;
{	
	int nport;

	for (nport = 0; nport < GKD_MAX_PORTS; nport++ ) {

		if (dev[nport].flags & F_GKD_INITED)
			continue;
		
		/*
		 * Verify the ports are correct (i.e. double 
		 *        check the probe routine)
		 */	
		if (dev[nport].regs->id_reset != nport ) {
			/* panic here eventually */
			printf("yikes! something wrong in gkd_init\n");
		}

		dev[nport].flags |= F_GKD_INITED;
	}
}

	
/*
 * Open the keyboard or mouse
 */
gkdopen(dev, flag)
	dev_t dev;
	int flag;
{
	struct gkd_dev *dptr;
	int m = minor(dev);

#ifdef GKDDEBUG
	printf("gkdopen(0x%x, 0x%x)\n", dev, flag);
#endif
	if (m >= GKD_MAX_PORTS)
		return(ENXIO);	/* no such device */

	dptr = &gkd_device[m];
	if ((dptr->flags & F_GKD_INITED) == 0)
		return(ENXIO);	/* device wasn't attached */

	if (dptr->flags & F_GKD_OPEN)
		return(EBUSY);	/* device already opened */

	if (flag & FNDELAY)
		dptr->flags |= F_GKD_NOBLOCK;

	dptr->flags |= F_GKD_OPEN;

#ifndef MACH_KERNEL
	/*
	 * It is safe to flush the read buffer as we are guaranteed
	 * that no one else is using it.
	 */
	ndflush(&dptr->hd_queue, dptr->hd_queue.c_cc);
#endif

	gkd_enable(dptr->regs);	/* enable the device */
	return(0);		/* should be all set */
}
 
gkdclose(dev, flag)
	dev_t dev;
{
	struct gkd_dev *dptr;
	int m = minor(dev);

#ifdef  GKDDEBUG
	printf("gkdclose called dev 0x%x, flag 0x%x\n", dev, flag);
#endif
	if (m >= GKD_MAX_PORTS)
		return(ENXIO);	/* no such device */

	dptr = &gkd_device[m];
	if ((dptr->flags & F_GKD_INITED) == 0)
		return(ENXIO);	/* device wasn't attached */

	dptr->flags &= ~ (F_GKD_OPEN | F_GKD_NOBLOCK); /* shut off open flag */
	
	gkd_disable(dptr->regs);     /* shut off device */

#if 0
	if (m == gkd_kbd_index)	kbdenable(0); /* ITE console needs this */
#endif

	return(0);
}

#ifdef	MACH_KERNEL
/* XXX fix this */
io_return_t
gkdread(dev, ior)
	dev_t dev;
	register io_req_t ior;
{
	return (EIO);
}
#else
/*
 * Read from the device
 */ 
gkdread(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	struct gkd_dev *dptr;
	int m = minor(dev), error, s, cc;
	char buf[GKDBUFSIZE];

	if (m >= GKD_MAX_PORTS)
		return(ENXIO);	/* no such device */

	dptr = &gkd_device[m];
	if ((dptr->flags & F_GKD_OPEN) == 0)
		return(ENXIO);	/* device wasn't opened?? is this possible??*/

	s = splgkd();
	while (dptr->hd_queue.c_cc == 0) {
		if (dptr->flags & F_GKD_NOBLOCK) {
			splx(s);
			return(EWOULDBLOCK);
		}
		dptr->flags |= F_GKD_ASLEEP;
		sleep((caddr_t)dptr, TTIPRI);
	}
	splx(s);

	error = 0;
	while (uio->uio_resid > 0 && error == 0) {
		cc = q_to_b(&dptr->hd_queue, buf,
			       MIN(uio->uio_resid, GKDBUFSIZE));
		if (cc <= 0)
			break;
		error = uiomove(buf, cc, uio);
	}
	return(error);
} 
#endif
 
/*
 * Map the registers.  We assume that one page is enough.
 */
gkdmap(dev, off, prot)
	dev_t dev;
	register int off;
{
#ifdef GKDDEBUG
	printf("gkdmap called dev 0x%x off 0x%x prot 0x%x\n", dev, off, prot);
#endif
}

gkdselect(dev, rw)
	dev_t dev;
{
	struct gkd_dev *dptr;
	int m = minor(dev), s;
	
	if (rw == FWRITE)
		return (1);

	dptr = &gkd_device[m];
#ifdef	MACH_KERNEL
	return (dptr->tty.t_inq.c_cc != 0);
#else
	s = splgkd();
	if (dptr->hd_queue.c_cc) {
		splx(s);
		return(1);
	}
	selrecord(u.u_procp, &dptr->gkd_selr);
	splx(s);
	return (0);
#endif
}

/*
 * Interrupt routine -- respond to keyboard/mouse interrupts!
 */ 
gkdintr(unit)
	register int unit;
{
 	unsigned char scancode, status;
	struct gkd_dev *dev_kbd = &gkd_device[gkd_kbd_index], 
	               *dev_mouse = &gkd_device[gkd_mouse_index];
	struct gkd_reg *kbd_regs = dev_kbd->regs, 
	               *mouse_regs = dev_mouse->regs;
	  
	/*
	 * The HP Gecko hardware turns the composite interrupt bit
	 * on for both the mouse and the keyboard no matter which device
	 * causes the interrupt
	 */
	while (kbd_regs->status & GKD_STATUS_CINT) {

		/* check for keyboard interrupt */
		if (kbd_regs->status & GKD_STATUS_RBNE) 
			gkd_kbd_intr(dev_kbd);

		/* check for mouse interrupt */
		if (mouse_regs->status & GKD_STATUS_RBNE) 
			gkd_mouse_intr(dev_mouse);
	}
}

/*
 * Handle a keyboard interrupt
 *
 */
gkd_kbd_intr(dev)
	struct gkd_dev *dev;
{
 	unsigned char scancode, status;
	struct gkd_reg *regs = dev->regs;

	while ( regs->status & GKD_STATUS_RBNE ) {

		scancode = regs->data;	/* get the data */

#ifdef GKDDEBUG
		if (dev->flags & F_GKD_SCRLK) 
			printf("(0x%x) ", scancode);
#endif

		/* if we're processing a command this takes priority */
		if (dev->in_cmd) {
			(dev->cmd_func)(dev, scancode);
			continue;
		}

		if (dev->flags & F_GKD_OPEN) {
			/*
			 * A user program has the buffer open,
			 * default to that
			 */
#ifdef MACH_KERNEL
			int s;
			b_to_q(&scancode, 1, &dev->tty.t_inq);
			s = spltty();
			tty_queue_completion(&dev->tty.t_delayed_read);
			splx(s);
#else
			b_to_q(&scancode, 1, &dev->hd_queue);
			if (dev->flags & F_GKD_ASLEEP) {
				dev->flags &= ~F_GKD_ASLEEP;
				wakeup((caddr_t)dev);
			}
			selwakeup(&dev->gkd_selr);
#endif
			continue;
		}

		status = GKBD_KEY; /* assume just a key pressed */

		/*
		 * We should be on the ITE console, process the keys
		 */
		
		if ((scancode == GKD_SPEC_KEY1) || (scancode == GKD_SPEC_KEY2) || dev->spec) {
			if (gkd_process_special(dev, &scancode) != GKD_DONE)
				continue;
			if (!scancode)
				continue; /* nothing more to do */
		}

		if (scancode == GKD_KEY_UP) {
			dev->released = 1;
			continue;
		} 

		switch (scancode) {
		case GKD_CAPS: 
			if (dev->released) {
				dev->released = 0;
				continue;
			}
			/* caps lock was pressed -- convert */
			scancode = KBD_CAPSLOCK;
			if (dev->flags & F_GKD_CAPS) {
				dev->flags &= ~F_GKD_CAPS;
				gkd_kbd_led(dev, dev->leds &= ~KBD_CAP_LOCK);
			} else {
				dev->flags |= F_GKD_CAPS;
				gkd_kbd_led(dev, dev->leds |= KBD_CAP_LOCK);
			}
			break;

		case GKD_ALT:
			if (dev->released) {
				dev->flags &= ~F_GKD_ALT;
				scancode = KBD_EXT_LEFT_UP;
			} else {
				dev->flags |= F_GKD_ALT;
				scancode = KBD_EXT_LEFT_DOWN;
			}
			dev->released = 0; 
			continue;

		case GKD_CTRL:
			if (dev->released)
				dev->flags &= ~F_GKD_CTRL;
			else
				dev->flags |=  F_GKD_CTRL;
			dev->released = 0;
			continue;

		case GKD_NUM_LCK:
			if (dev->released) {
				dev->released = 0;
				continue;
			}
			/* num lock was pressed -- what do we do?? */
			if (dev->flags & F_GKD_NUMLK) {
				dev->flags &= ~F_GKD_NUMLK;
				gkd_kbd_led(dev, dev->leds &= ~KBD_NUM_LOCK);
			} else {
				dev->flags |= F_GKD_NUMLK;
				gkd_kbd_led(dev, dev->leds |= KBD_NUM_LOCK);
			}
			continue;

		case GKD_SCROLL_LCK:
			if (dev->released) {
				dev->released = 0;
				continue;
			}
			/* scroll lock was pressed -- what do we do?? */
			if (dev->flags & F_GKD_SCRLK) {
				dev->flags &= ~F_GKD_SCRLK;
				gkd_kbd_led(dev, dev->leds &= ~KBD_SCRL_LOCK);
			} else {
				dev->flags |= F_GKD_SCRLK;
				gkd_kbd_led(dev, dev->leds |= KBD_SCRL_LOCK);
			}
			continue;
			
		case GKD_RIGHT_SHIFT:
		case GKD_LEFT_SHIFT:
			if (dev->released)
				dev->flags &= ~F_GKD_SHIFT;
			else
				dev->flags |=  F_GKD_SHIFT;
			dev->released = 0;
			continue;

		default:
			switch (dev->flags & F_KBD_SCC) {
			case F_GKD_SHIFT:
				status = GKBD_SHIFT;
				break;
			case F_GKD_CTRL:
				status = GKBD_CTRL;
				break;
			case F_GKD_CTRL|F_GKD_SHIFT:
				status = GKBD_CTRLSHIFT;
				break;
			}
		}
			
		if (dev->released) {
			dev->released = 0;
		} else {
			if (scancode == GKD_KEY_UP) {
#ifdef GKDDEBUG
				printf("\n\n ACK!! scancode is KEY_UP!\n\n");
#endif
				scancode = 0;
			}
#ifdef GKDDEBUG
			if (scancode > 0x87) 
				printf("\n\n ACK: SCANCODE 0x%x\n\n", scancode);
			else
#endif
				gkditefilter(dev, status, scancode);
		}
	}
}

/*
 * Handle a mouse interrupt
 */
gkd_mouse_intr(dev)
	struct gkd_dev *dev;
{
 	unsigned char data, status;
	struct gkd_reg *regs = dev->regs;

	while ( regs->status & GKD_STATUS_RBNE ) {

		data = regs->data;	/* get the data */

		/* if we're processing a command this takes priority */
		if (dev->in_cmd) {
			(dev->cmd_func)(dev, data);
			continue;
		}

		if ((dev->flags & F_GKD_OPEN) == 0)
			continue; /* no one is reading the mouse */

#ifdef MACH_KERNEL
		b_to_q(&data, 1, &dev->tty.t_inq);
#else
		b_to_q(&data, 1, &dev->hd_queue);
#endif
		if (dev->flags & F_GKD_ASLEEP) {
			dev->flags &= ~F_GKD_ASLEEP;
			wakeup((caddr_t)dev);
		}
#ifdef MACH_KERNEL
		{
			int s = spltty();
			tty_queue_completion(&dev->tty.t_delayed_read);
			splx(s);
		}
#else
		selwakeup(&dev->gkd_selr);
#endif
	}
}

/*
 * Handle the special keys from the keyboard (ala keyboard arrows etc )
 */
gkd_process_special(dev, scancode)
	struct gkd_dev *dev;
	unsigned char *scancode;
{
	unsigned char c = *scancode;


	if (!dev->spec) {
		dev->spec = c;
		return GKD_NEED_MORE;
	}

	if (c == GKD_KEY_UP) {
		dev->released = 1;
		return GKD_NEED_MORE;
	}

	/*
	 * Looks like high key numbers are single codes
	 * need to watch out for SysReq however..
	 */
	if (c != 0x12) {

		if ((dev->spec == GKD_SPEC_KEY2) && (c != 0x77)) {
			/* processing of Pause/Break key is not done yet */
			return GKD_NEED_MORE;
		} else if (dev->spec == GKD_SPEC_KEY2) {
			*scancode = 0;
			if (dev->released) dev->released = 0;
		} else if (dev->spec == GKD_SPEC_KEY3) {
			if (c == GKD_SPEC_KEY1)
				return GKD_NEED_MORE;
			*scancode = 0;
			if (dev->released) dev->released = 0;
		}

		dev->spec = 0;
		return GKD_DONE;
	} else {
		/* yikes!!  we have a sysreq -- what to do? */

		dev->spec == GKD_SPEC_KEY3;
		return GKD_NEED_MORE;
	}
}

/*
 * check up on a numlock scancode and return the appropriate value
 */
gkd_chk_numlock(code, c)
        char *code, c;
{

        static char gkd_num_lck_table[] = {'1','\0','4','7','\0','\0','\0',
                                           '0','.','2','5','6','8','\0','\0',
                                           '\0','+','3','-','*','9'};
        /*
         * check to see if numlock character and return if so
         */

        if (c < 0x69 || c > 0x7d)
                return;         /* nope, it ain't */

        *code = gkd_num_lck_table[ c - 0x69];
}


/*
 * Config keyboard device
 */
gkd_config(kbd)
     struct gkd_dev *kbd;
{
	struct kbdmap *km;
	int  lang_code = KBD_US; /* XXX fix this someday */
	int  s;
	extern struct kbdmap gkbd_map[];

	s = splgkd();
	for (km = gkbd_map; km->kbd_code; km++) {
		if (km->kbd_code == lang_code) {
			kbd_keymap = km->kbd_keymap;
			kbd_shiftmap = km->kbd_shiftmap;
			kbd_ctrlmap = km->kbd_ctrlmap;
			kbd_ctrlshiftmap = km->kbd_ctrlshiftmap;
			kbd_stringmap = km->kbd_stringmap;
		}
	}
	splx(s);
}

/*
 * Handle keyboard commands to turn on/off the leds
 */
gkd_kbd_led(dev, val)
	struct gkd_dev *dev;
	int val;		/* either scancode or command arg */
{

	gkd_kbd_set_leds[3] = val; /* XXX this is known to be here! */
	gkd_send_cmd(dev, gkd_kbd_set_leds);
#ifdef no

	if (!dev->in_cmd) {
		/* 
		 * We're starting the command, setup the state
		 */
		dev->in_cmd = 1; /* commands takes priority over key strokes */
		dev->cmd_state = 0; /* initial state */
		dev->cmd_func = gkd_kbd_led;
		dev->cmd_buf[0] = GKD_KBD_LEDS;
		dev->cmd_buf[1] = val;

		/*
		 * Send the first byte
		 */
		gkd_write_cmd(dev->regs, dev->cmd_buf[0]);
		return;
	}

	/*
	 * in the middle of processing this command
	 */
	switch (dev->cmd_state) {
		/* 
		 * State 0: receive ack, go to state 1
                 *          else resend command, loop in state 0
		 */
	case 0:		
		if (val == GKD_DEV_ACK) {
			/* command was acked, go on */
			dev->cmd_state = 1;
			gkd_write_cmd(dev->regs, dev->cmd_buf[1]);
		} else if (val == GKD_DEV_RETRY) {
			/* command needs to be sent again, do it */
			gkd_write_cmd(dev->regs, dev->cmd_buf[1]);
			return;
		} else {
			printf("gkd ack:unknown val %d in state %d\n", val, 0);
		}
		break;
	case 1:
		/*
		 * State 1: receive ack, finished
		 *          else, resend data arg, loop in state 1
		 */
		if (val == GKD_DEV_ACK) {
			/* val was acked, we're done */
			dev->in_cmd = 0; /* no longer processing */
			return;
		} else if (val == GKD_DEV_RETRY) {
			/* command needs to be sent again, do it */
			gkd_write_cmd(dev->regs, dev->cmd_buf[1]);
			return;
		} else {
			printf("gkd ack:unknown val %d in state %d\n", val, 1);
		}
	}
#endif
}

/*
 * Send a command to the specified device
 */
gkd_send_cmd(dev, cmd2send)
	struct gkd_dev *dev;
	char *cmd2send;
{
	int gkd_continue_cmd();

	if (dev->in_cmd)	/* we can only do one command at a time */
		return(-1);

	/* 
	 * We're starting the command, setup the state
	 */
	dev->in_cmd = 1; /* commands take priority over key strokes */
	dev->cmd_state = 0; /* initial state */
	dev->cmd_func = gkd_continue_cmd;
	dev->cmd_ptr = cmd2send;
	dev->cmd_serr = dev->cmd_terr = dev->cmd_nread = dev->cmd_nindx = 0;

	/*
	 * Send the first byte
	 */
	gkd_write_cmd(dev->regs, GKD_GSB(dev));
	return 0;
}

/*
 * continue processing a command
 */
gkd_continue_cmd(dev, val)
	struct gkd_dev *dev;
	int val;		/* byte read from the port */
{

#ifdef GKDDEBUG
	printf("gkd_continue_cmd dev %d, val 0x%x\n",
	       dev - &gkd_device[0], val);
#endif
	
	/*
	 *  Assume positive -- the command accepted okay
	 */
	if (val == GKD_DEV_ACK) {
		/* command was acked, go on */
		dev->cmd_state += 1;
		if (dev->cmd_state == GKD_SNUM(dev) && 
		                   dev->cmd_nindx == GKD_GSN(dev)) {
			/* we're done! */
cdone:
			dev->in_cmd = 0;
#ifdef GKDDEBUG
			printf("gkd_continue_cmd: command done\n");
#endif
			if (dev->flags & F_GKD_CSLEEP) {
				dev->flags &= ~F_GKD_CSLEEP;
				wakeup(dev->cmd_ptr);
			}
			return;
		}
		gkd_write_cmd(dev->regs, GKD_GSB(dev));
		return;
	}

	if (val == GKD_DEV_RETRY) {
		/* oh great! send the command again */
		dev->cmd_serr++;
		dev->cmd_terr++;
		if (dev->cmd_serr > 5)
			goto cdone;  /* too many errs -- bail! */
		gkd_write_cmd(dev->regs, GKD_GSB(dev));
		return;
	}

	/*
	 * anything else, check to see if we want the info otherwise
	 * silently drop
	 */
	if (GKD_GSN(dev) > dev->cmd_nindx)
		dev->cmd_buf[dev->cmd_nindx++] = val;

	/*
	 * XXX special case: IDENT command for mouse only returns 1
	 */
	if ((dev->cmd_ptr[2] == (char)0xF2) /*XXX ident command*/  && (val == 0)) {
		/* this is a mouse and we're done */
		goto cdone;
	}
		
}

/*
 * Give a command and let the interrupt routine carry it out.
 * Called from the top half of the driver 
 */
gkd_give_cmd(dev, cmd)
	struct gkd_dev *dev;
	char *cmd;
{
	int s = splgkd();

#ifdef GKDDEBUG
	printf("gkd_give_cmd: spl s is 0x%x\n", s);
#endif
	if (gkd_send_cmd(dev, cmd)) {
		splx(s);
		return(-1);	/* this failed, no use continuing */
	}

	dev->flags |= F_GKD_CSLEEP;
	while ((dev->flags & F_GKD_CSLEEP) && dev->in_cmd)
		sleep((caddr_t)cmd, TTIPRI);

	if (dev->flags & F_GKD_CSLEEP) {
		printf("gkd_give_cmd: command completed before sleep!\n");
		dev->flags &= ~F_GKD_CSLEEP;
	}

	splx(s);
	return(0);
}

/*
 * Drive a command in "poll" mode (i.e. simulate the interrupt routine)
 */
gkd_issue_cmd(dev, cmd)
	struct gkd_dev *dev;
	char *cmd;
{
	unsigned char data;
	struct gkd_reg *regs = dev->regs;
	int s = splgkd();
	
	/* chuck everything to start */
	while (regs->status & GKD_STATUS_RBNE) {
		data = regs->data;
		DELAY(30000);
	}

	/* call the function to kick things off */
	if (gkd_send_cmd(dev, cmd))
		return(-1);	/* yikes, how can this be? */

	while (dev->in_cmd) {
		while (!(regs->status & GKD_STATUS_RBNE)) DELAY(30000);
		data = regs->data;
		gkd_continue_cmd(dev, data);
	}
	   
	/* chuck everything else */
	while (regs->status & GKD_STATUS_RBNE) {
		data = regs->data;
		DELAY(30000);
	}
	splx(s);
}
	

/*
 * write this command out
 */
gkd_write_cmd(regs, byte)
	struct gkd_reg *regs;
	unsigned char byte;
{	
	int max_count = 2000;	/* max tries */

	while (regs->status & GKD_STATUS_TBNE) {
		if (!max_count--) {
			printf("gkd: warning! can't talk with the keyboard!\n");
			return;
		}
		DELAY(500);
	}

#ifdef GKDDEBUG		
	printf("gkd_write_cmd 0x%x\n", byte);
#endif
	regs->data = byte;
        gkd_enable(regs);	/* send the command */
}

#ifdef MACH_KERNEL
/*
 * Mach splits ioctls into those that write data into the kernel, and
 * those that read data out of the kernel (yep, very annoying).
 * Provide some glue around hilioctl().
 */

io_return_t
gkdgetstat(dev, flavor, data, count)
	dev_t		dev;
	int		flavor;
	int *		data;		/* pointer to OUT array */
	unsigned int	*count;		/* out */
{
	int error;

	error = gkdioctl(dev, flavor, data, 0);
	*count = ((flavor >> 16) & 0x7F) / sizeof(int);
	return(error);
}

io_return_t
gkdsetstat(dev, flavor, data, count)
	dev_t		dev;
	int		flavor;
	int *		data;
	unsigned int	count;
{
	return(gkdioctl(dev, flavor, data, 0));
}
#endif

gkdioctl(dev, cmd, data, flag)
	dev_t dev;
	caddr_t data;
{
	int m = minor(dev), arg, err = 0;
	struct gkd_dev *dptr;

#ifdef GKDDEBUG
	printf("gkdioctl called dev 0x%x cmd 0x%x data 0x%x flag 0x%x\n",
	       dev, cmd, data, flag);
#endif

	if (m >= GKD_MAX_PORTS)
		return(ENXIO);	/* no such device */

	dptr = &gkd_device[m];

	switch (cmd) {
	case PS2_SETDEFAULT:
		if (gkd_give_cmd(dptr, gkd_device_setdefault))
			err = EIO;
		break;

	case PS2_DISABLE:
		if (gkd_give_cmd(dptr, gkd_device_disable))
			err = EIO;
		break;

	case PS2_RESET:
		if (gkd_give_cmd(dptr, gkd_device_reset))
		         err = EIO;
		break;

	case PS2_SAMPLERATE:
		/* only for mouse */
		arg = ((char *)data)[0];

#ifdef GKDDEBUG
		printf("PS2_SAMPLERATE, arg is %d\n", arg);
#endif
		gkd_device_rate[3] = arg;
		if (gkd_give_cmd(dptr, gkd_device_rate))
			err = EIO;
		break;

	case PS2_RATEDELAY:
		/* only for keyboard */
		arg = ((char *)data)[0];

		gkd_device_rate[3] = arg;
		if (gkd_give_cmd(dptr, gkd_device_rate))
			err = EIO;
		break;

	case PS2_STREAMMODE:
		/* for mouse only */
		if (gkd_give_cmd(dptr, gkd_mouse_stream))
			err = EIO;
		break;

	case PS2_RESOLUTION:
		/* for mouse only */
		arg = ((char *)data)[0];

                gkd_mouse_resolution[3] = arg;
#ifdef GKDDEBUG
		printf("gkdioctl: PS2_RESOLUTION val 0x%x\n", arg);
#endif
				 
		if (gkd_give_cmd(dptr, gkd_mouse_resolution))
			err = EIO;	 

		break;

	case PS2_ENABLE:
		if (gkd_give_cmd(dptr, gkd_device_enable))
			err = EIO;
	        break;

	case PS2_SCANCODE:
		/* for keyboard only */
		arg = ((char *)data)[0];

		gkd_kbd_scancode[3] = arg;
#ifdef GKDDEBUG
		printf("gkdioctl: PS2_SCANCODE val 0x%x\n", arg); 
#endif
		
		if (gkd_give_cmd(dptr, gkd_kbd_scancode))
			err = EIO;
		break;

	case PS2_INDICATORS:
		/* for keyboard only */
		arg = ((char *)data)[0];
#ifdef GKDDEBUG
		printf("gkdioctl: PS2_INDICATORS *data is 0x%x\n", arg);
#endif

		if (arg < KBD_NONE || arg > KBD_CAP_LOCK) {
			err = EINVAL;
			goto done;
		}

		gkd_kbd_set_leds[3] = arg;
		if (gkd_give_cmd(dptr, gkd_kbd_set_leds))
			err = EIO;
		break;

	case PS2_ALL_TMAT_MKBRK:
		
		if (gkd_give_cmd(dptr, gkd_kbd_all_tmat_mkbrk))
			err = EIO;
		break;
	       
	case PS2_ALL_TMAT:
		
		if (gkd_give_cmd(dptr, gkd_kbd_all_tmat))
			err = EIO;
		break;

	case PS2_PORTSTAT:

		/* XXX */
		if (m == gkd_kbd_index) {
			data[0] = 2;	/* PS2_KEYBD */
			data[1] = 3;	/* HAS_FIRST_KEYBD | HAS_ITE */
		} else if (m == gkd_mouse_index) {
			data[0] = 1;	/* PS2_MOUSE */
			data[1] = 5;	/* HAS_FIRST_MOUSE | HAS_ITE */
		} else
			data[0] = data[1] = 0;
		break;
		
	case PS2_IDENT:

		/* XXX */
		if (m == gkd_kbd_index) {
			data[0] = 0xAB;	/* so says the HP-UX man page */
			data[1] = 0x83;	/* so says the HP-UX man page */
		} else if (m == gkd_mouse_index)
			data[0] = 0x00;	/* so says the HP-UX man page */
		else
			data[0] = data[1] = 0;
		break;
	}

done:
	return(err);
}

/*
 * Cooked keyboard functions for ite driver.
 */

/*
 * Ring the bell.
 */
gkdkbdbell()
{
  /* nothing to do yet (maybe when audio driver works...) */
}

/*
 * Enable the keyboard
 */
gkdkbdenable(unit)
     int unit;
{
     int s = splgkd();
     struct gkd_dev *dptr = &gkd_device[gkd_kbd_index];
	
     gkd_enable(dptr->regs);	/* enable interrupts */

     gkd_kbd_scancode[3] = 2;	/* reset scancode */
     gkd_issue_cmd(dptr, gkd_kbd_scancode);

     gkd_kbd_set_leds[3] = 0;   /* reset LEDS and state */
     dptr->flags = F_GKD_INITED;
     gkd_issue_cmd(&gkd_device[0], gkd_kbd_set_leds);

     splx(s);
}

/*
 * Disable the keyboard
 */
gkdkbddisable(unit)
     int unit;
{
     int s = splgkd();
     struct gkd_dev *dptr = &gkd_device[gkd_kbd_index];
     
     gkd_disable(dptr->regs);	/* disable the device */

     splx(s);
}

/*
 * Get a character
 */
gkdkbdgetc()
{
  /* need to write this */
}
#endif /* NGKD > 0 */
