/* 
 * Copyright (c) 1994, The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: gkdreg.h 1.4 94/12/14$
 */

/*
 *  Structure describing the registers of the PS2 interface on
 *  HP Geckos.
 *
 *         Offset      Register
 *           0           id      (r)
 *           0           reset   (w)
 *           4           data    (r/w)
 *           8           control (r/w)
 *          12           status  (r)
 */

struct gkd_reg {
	volatile u_int           :24,
                         id_reset:8;  /* Read returns ID, write resets*/
	volatile u_int           :24,
	                     data:8;  /* read: receive, write: transmit */ 
	volatile u_int           :24,
	                  control:8;  /* Control register */
	volatile u_int           :24,
                           status:8;  /* Status register */
};

/*
 *  Some constants
 */
#define   GKD_PADDING    256

/*
 *  Given an address, return next port
 */
#define   gkd_next(a)    ((struct gkd_reg *)((caddr_t)a + GKD_PADDING))

/*
 *  Enable/Disable a port given a pointer
 */
#define   gkd_disable(a)  ((struct gkd_reg *)a)->control = 0
#define   gkd_enable(a)   ((struct gkd_reg *)a)->control = 1

/*
 *  Bit fields for status register
 */
#define   GKD_STATUS_RBNE  0x01		/* receive buffer not empty */
#define   GKD_STATUS_TBNE  0x02 	/* transmit buffer not empty */
#define   GKD_STATUS_TERR  0x04 	/* timeout error detected */
#define   GKD_STATUS_PERR  0x08         /* parity error detcted */
#define   GKD_STATUS_CINT  0x10	        /* composite interrupt */
	                                /* reserved */
#define   GKD_STATUS_SDAT  0x40         /* shadow of external data line */
#define   GKD_STATUS_SCLK  0x80         /* shadow of external clock line */

/*
 * Commands to keyboard
 */
#define   GKD_KBD_LEDS     0xED	        /* Set LEDS */

/*
 * Values returned
 */
#define   GKD_DEV_ACK      0xFA
#define   GKD_DEV_RETRY    0XFE 

