/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * Definitions for SCSI support
 */

#ifndef _HP_SCSI_H_
#define _HP_SCSI_H_ 1

#ifdef __cplusplus
    extern "C" {
#endif

#ifdef	KERNEL
#ifdef	MACH_KERNEL
#include <sys/types.h>
#include <device/device_types.h>
#include <device/buf.h>
#else
#include <sys/buf.h>
#endif
#endif

#include <sys/ioctl.h>

/*
 * NEXUS (bus,tid,lun) scsi device identifier
 */
struct scsi_device {
	unsigned char	bus;            /* bus index (starts at 0) */
	unsigned char	target;         /* bus target */
	unsigned char	lun;            /* target logical unit */
	unsigned char	unused;         /* Not currently used */
};

#ifdef KERNEL
/* various structures and commands defined by the SCSI interface */

/* actions for scsi_command */
#define SCSI_QUEUE_HEAD		1
#define SCSI_QUEUE_TAIL		2
#define SCSI_QUEUE_ANYWHERE	3
#define SCSI_FLUSH_ALL		4
#define SCSI_FLUSH_COMMAND	5

/* The various generic formats of comand descriptor blocks */
union scsi_cdb {
	struct {
		unsigned	opcode:8;
		unsigned	lun:3;
		unsigned	block_addr:21;
		unsigned	length:8;
		unsigned	control:8;
	} c6;
	struct {
		unsigned	opcode:8;
		unsigned	lun:3;
		unsigned	:4;
		unsigned	reladr:1;
		unsigned	block_addr_hi:16;
		unsigned	block_addr_lo:16;
		unsigned	:8;
		unsigned	length_hi:8;
		unsigned	length_lo:8;
		unsigned	control:8;
	} c10;
	struct {
		unsigned	opcode:8;
		unsigned	lun:3;
		unsigned	:4;
		unsigned	reladr:1;
		unsigned	block_addr_hi:16;
		unsigned	block_addr_lo:16;
		unsigned	length_hi:16;
		unsigned	length_lo:16;
		unsigned	:8;
		unsigned	control:8;
	} c12;
	struct {
		unsigned	group:3;
		unsigned	command:5;
	} code;
	unsigned char	command[12];
};

#define	SCSI_SET_BADDR10(_cc, _baddr)				\
	(_cc)->cdb.c10.block_addr_hi = (unsigned)(_baddr) >> 16,\
	(_cc)->cdb.c10.block_addr_lo = (_baddr)
#define	SCSI_SET_BLEN10(_cc, _blen)				\
	(_cc)->cdb.c10.length_hi = (unsigned)(_blen) >> 8,\
	(_cc)->cdb.c10.length_lo = (_blen)
#define	SCSI_SET_BADDR12(_cc, _baddr)				\
	(_cc)->cdb.c12.block_addr_hi = (unsigned)(_baddr) >> 16,\
	(_cc)->cdb.c12.block_addr_lo = (_baddr)
#define	SCSI_SET_BLEN12(_cc, _blen)				\
	(_cc)->cdb.c12.length_hi = (unsigned)(_blen) >> 16,\
	(_cc)->cdb.c12.length_lo = (_blen)

/* The comand structure that is passed to the device dependant SCSI layer */
struct scsi_cmd {
	unsigned long		flags;			/* various flags */
	unsigned long		data_len;		/* length of data in/out area */
	vm_offset_t		data_addr;		/* pointer to data in/out area */
	vm_map_t		data_map;		/* map data in/out area resides in, NULL is kernel map */
#ifdef	__STDC__
	void			(*io_done)(		/* routine to call when I/O is complete */
					struct scsi_cmd *cmd);
#else
	void			(*io_done)();		/* routine to call when I/O is complete */
#endif /* __STDC__ */
	vm_offset_t		driver_info;		/* usually a pointer to a buf structure */
	vm_offset_t		controller_info;	/* usually a pointer to a controller specific per request data */
	struct scsi_cmd		*next;			/* ptr to next command in target's queue */
	union scsi_cdb		cdb;			/* the scsi command to send */
	unsigned long		cdb_len;		/* command length (filled in by scsi_command) */
	struct scsi_device	device;			/* device number for driver */
	unsigned long		id;			/* command id (filled in by scsi_command) */
	unsigned long		scsi_status;		/* scsi return status (SCSI_CHECK_CONDITION implies look at sense status & info) */
	unsigned long		scsi_sense_status;	/* scsi sense status */
	struct scsi_request_sense *sense;		/* sense if there was an error (must be scsi_sense_free'd) */
};

/* for the flags field in the scsi_cmd */
#define SCSI_DATA_OUT   0x01            /* data is outbound from initiator */
#define SCSI_DATA_IN    0x02            /* data is inbound from initiator */
#define SCSI_WAIT       0x04            /* do a wakeup on the command buffer */
#define SCSI_CALL       0x08            /* call the completion function for target */
#define SCSI_INPROGRESS	0x10            /* The command is being processed */
#define SCSI_FLUSH	0x20            /* The command has been flushed from the queue */

#define SCSI_DEVICE_FLAGS	0x00ff0000	/* mask of flags that device drivers may use */
/* macro for defining device flags */
#define SCSI_DEVICE_FLAG(b)	(((b)<<16)&SCSI_DEVICE_FLAGS)

#define SCSI_CONTROLLER_FLAGS	0xff000000	/* mask of flags that controllers may use */
/* macro for defining controller flags */
#define SCSI_CONTROLLER_FLAG(b) (((b)<<24)&SCSI_CONTROLLER_FLAGS)

#define SCSI_PHYSIO             0x80000000

/* Macros to check the scsi_status code in a scsi_cmd structure. */
#define SCSI_GOOD			0x00
#define SCSI_CHECK_CONDITION		0x02
#define SCSI_CONDITION_MET		0x04
#define SCSI_BUSY			0x08
#define SCSI_INTERMEDIATE_GOOD		0x10
#define SCSI_INTERMEDIATE_CONDITION_MET	0x14
#define SCSI_RESERVATION_CONFLICT	0x18
#define SCSI_QUEUE_FULL			0x28
#define SCSI_DATA_DIRECTION_ERROR       0xFB
#define SCSI_DMA_UNDERRUN               0xFC
#define SCSI_DMA_OVERRUN                0xFD
#define SCSI_DEVICE_NA                  0xFE    /* Device not available */
#define SCSI_UNDEF_STAT                 0xFF    /* Undefined status */

/* command codes and their structures (all devices) */
#define SCSI_CHANGE_DEFINITION		0x40
#define SCSI_COMPARE			0x39
#define SCSI_COPY			0x18
#define SCSI_COPY_AND_VERIFY		0x3a
#define SCSI_INQUIRY			0x12
#define SCSI_MODE_SELECT_6		0x15
#define SCSI_MODE_SELECT_10		0x55
#define SCSI_MODE_SENSE_6		0x1a
#define SCSI_MODE_SENSE_10		0x5a
#define SCSI_READ_BUFFER		0x3c
#define SCSI_READ_LOG			0x5f
#define SCSI_RECEIVE_DIAGNOSTIC_RESULTS	0x1c
#define SCSI_REQUEST_SENSE		0x03
#define SCSI_SEND_DIAGNOSTIC		0x1d
#define SCSI_TEST_UNIT_READY		0x00
#define SCSI_WRITE_BUFFER		0x3b

/* command codes for various devices */
#define SCSI_PREVENT_REMOVAL		0x1e
#define SCSI_READ			0x08
#define SCSI_RELEASE			0x17
#define SCSI_RESERVE			0x16
#define SCSI_VERIFY			0x13
#define SCSI_WRITE			0x0a

/* commands only for disks */
#define SCSI_FORMAT_UNIT		0x04
#define SCSI_LOCK_CACHE			0x36
#define SCSI_PRE_FETCH			0x34
#define SCSI_READ_10			0x28
#define SCSI_READ_CAPACITY		0x25
#define SCSI_READ_DEFECT_DATA		0x37
#define SCSI_READ_LONG			0x3e
#define SCSI_REASSIGN_BLOCKS		0x07
#define SCSI_REZERO_UNIT		0x01
#define SCSI_SEARCH_DATA_EQUAL		0x31
#define SCSI_SEARCH_DATA_HIGH		0x30
#define SCSI_SEARCH_DATA_LOW		0x32
#define SCSI_SEEK			0x0b
#define SCSI_SEEK_10			0x2b
#define SCSI_SET_LIMITS			0x33
#define SCSI_START_UNIT			0x1b
#define SCSI_SYNC_CACHE			0x35
#define SCSI_WRITE_10			0x2a
#define SCSI_WRITE_AND_VERIFY		0x2e
#define SCSI_WRITE_LONG			0x3f
#define SCSI_WRITE_SAME			0x41

/* commands only for tapes */
#define SCSI_ACCESS_LOG			0x5f
#define SCSI_ERASE			0x19
#define SCSI_LOAD			0x1e
#define SCSI_LOCATE			0x28
#define SCSI_READ_BLOCK_LIMITS		0x05
#define SCSI_READ_POSITION		0x35
#define SCSI_READ_REVERSE		0x0f
#define SCSI_RECOVER_BUFFERED_DATA	0x14
#define SCSI_REWIND			0x01
#define SCSI_SPACE			0x11
#define SCSI_WRITE_FILEMARK		0x10

#endif /* KERNEL */

/* The structure returned by a INQUIRY command */
struct scsi_inquiry {
	unsigned	peripheral_qualifier:3;
	unsigned	peripheral_device_type:5;
	unsigned	removable_media:1;
	unsigned	device_type_qualifier:7;
	unsigned	iso_version:2;
	unsigned	ecma_version:3;
	unsigned	ansi_version:3;
	unsigned	supports_async_event_notification:1;
	unsigned	:3;
	unsigned	response_data_format:4;
	unsigned	additional_length:8;
	unsigned	:16;
	unsigned	supports_relative_address:1;
	unsigned	supports_32bit_transfers:1;
	unsigned	supports_16bit_transfers:1;
	unsigned	supports_sync_transfers:1;
	unsigned	supports_linked_operations:1;
	unsigned	supports_cacheing:1;
	unsigned	supports_command_queuing:1;
	unsigned	:1;
	char		vendor_identification[8];
	char		product_identification[16];
	char		product_revision[4];
	unsigned char	vendor_unique[20];
	unsigned char	reserved1[40];
	unsigned char	vendor_parameters[1];
};

/* values of peripheral qualifier field */
#define SCSI_DEVICE_CONNECTED		0x00
#define SCSI_DEVICE_NOT_CONNECTED	0x01
#define SCSI_NO_PHYSICAL_DEVICE		0x03

/* values of peripheral device type field */
#define SCSI_DIRECT_ACCESS	0x00
#define SCSI_SEQUENTIAL_ACCESS	0x01
#define SCSI_PRINTER_DEVICE	0x02
#define SCSI_PROCESSOR_DEVICE	0x03
#define SCSI_WORM_DEVICE	0x04
#define SCSI_ROM_DEVICE		0x05
#define SCSI_SCANNER_DEVICE	0x06
#define SCSI_OPTICAL_DEVICE	0x07
#define SCSI_CHANGER_DEVICE	0x08
#define SCSI_COMM_DEVICE	0x09
#define SCSI_UNKNOWN_DEVICE	0x1f

/* The following structure is used by the READ BLOCK LIMITS command */
struct scsi_block_limits {
	unsigned	:8;
	unsigned	max_block_length:24;
	unsigned	min_block_length:16;
};

/* The following structure is used by the READ CAPACITY command */
struct scsi_read_capacity {
        unsigned long	logical_block_address;
        unsigned long	block_length;
};

/* The following structures are used by the MODE_SENSE/MODE_SELECT command */
struct scsi_mode1_header {
	unsigned 	data_length:8;
	unsigned 	medium_type:8;
	unsigned 	device_specific_params:8;
	unsigned 	block_descriptor_length:8;
};

struct scsi_mode2_header {
	unsigned	data_length:16;
	unsigned	medium_type:8;
	unsigned	device_specific_params:8;
	unsigned	reserved:16;
	unsigned	block_descriptor_length:16;
};

struct scsi_mode_block {
	unsigned	density_code:8;
	unsigned	number_of_blocks:24;
	unsigned	reserved:8;
	unsigned	block_length:24;
};

/* Default block length for variable record sized sequential devices */
#define SCSI_SEQUENTIAL_DEFAULT_BLOCK_LENGTH	512

/*
 * This page header is at the top of every sense/select page.
 */
#define SCSI_MODE_PAGE_HEADER_SIZE	2
struct scsi_mode_page_header {
	unsigned	parameters_savable:1;
	unsigned	:1;
	unsigned	page_code:6;
	unsigned	page_parameter_length:8;
};

#define SCSI_MODE_ERROR_RECOVERY_PAGE	0x01
struct scsi_mode_error_recovery_page {
	unsigned	parameters_savable:1;
	unsigned	:1;
	unsigned	page_code:6;
	unsigned	page_parameter_length:8;
	unsigned	awre:1;			/* Automatic Write REallocation */
	unsigned	arre:1;			/* Automatic Read REallocation */
	unsigned	tb:1;			/* Transfer Block */
	unsigned	rc:1;			/* Read Continuous */
	unsigned	eer:1;			/* Enable Early Recovery */
	unsigned	per:1;			/* Post ERror */
	unsigned	dte:1;			/* Disable Transfer on Error */
	unsigned	dcr:1;			/* Disable CoRrection */
	unsigned	read_retry_count:8;
	unsigned	correction_span:8;
	unsigned	head_offset_count:8;
	unsigned	data_strobe_offset_count:8;
	unsigned	:8;
	/* added CCS fields */
	unsigned	write_retry_count:8;
	unsigned	:8;
	unsigned	recovery_time_limit:16;
};

#define SCSI_MODE_CONNECT_PAGE	0x02
struct scsi_mode_connect_page {
	unsigned	parameters_savable:1;
	unsigned	:1;
	unsigned	page_code:6;
	unsigned	page_parameter_length:8;
	unsigned	buffer_full_ratio:8;
	unsigned	buffer_empty_ratio:8;
	unsigned	bus_inactivity_limit:16;
	unsigned	bus_disconnect_time_limit:16;
	unsigned	bus_connect_time_limit:16;
	unsigned	maximum_burst_size:16;
};

#define SCSI_MODE_FORMAT_PAGE	0x03
struct scsi_mode_format_page {
	unsigned	parameters_savable:1;
	unsigned	:1;
	unsigned	page_code:6;
	unsigned	page_parameter_length:8;
	unsigned	tracks_per_zone:16;
	unsigned	alt_sectors_per_zone:16;
	unsigned	alt_tracks_per_zone:16;
	unsigned	alt_tracks_per_lu:16;
	unsigned	sectors_per_track:16;
	unsigned	bytes_per_sector:16;
	unsigned	interleave:16;
	unsigned	track_skew_factor:16;
	unsigned	cylinder_skew_factor:16;
	unsigned	ssec:1;				/* softsector */
	unsigned	hsec:1;				/* hardsector */
	unsigned	rmb:1;				/* removeable */
	unsigned	surf:1;				/* surface */
	unsigned	:28;
};

#define SCSI_MODE_RIGID_DISK_GEOMETRY_PAGE	0x04
struct scsi_mode_rigid_disk_geometry_page {
	unsigned	parameters_savable:1;
	unsigned	:1;
	unsigned	page_code:6;
	unsigned	page_parameter_length:8;
	unsigned char	number_of_cylinders[3];
	unsigned	number_of_heads:8;
	unsigned char	starting_cylinder_wpc[3];
	unsigned char	starting_cylinder_rwc[3];
	unsigned	drive_step_rate:16;
	unsigned char	landing_zone_cylinder[3];
	unsigned	:6;
	unsigned	rpl:2;
	unsigned	rotational_offset:8;
	unsigned	:8;
	unsigned	rotation_rate:16;
	unsigned	:8;
	unsigned	:8;
};
#define SCSI_24BIT_FIELD(p, f) \
	(((p)->f[0] << 16) | ((p)->f[1] << 8) | (p)->f[2])

#define SCSI_MODE_FLEX_DISK_PARAM_PAGE	0x05
struct scsi_mode_flex_disk_param_page {
	unsigned	parameters_savable:1;
	unsigned	:1;
	unsigned	page_code:6;
	unsigned	page_parameter_length:8;
	unsigned	transfer_rate:16;
	unsigned	number_of_heads:8;
	unsigned	sectors_per_track:8;
	unsigned	bytes_per_sector:16;
	unsigned	number_of_cylinders:16;
	unsigned	starting_cylinder_wpc:16;
	unsigned	starting_cylinder_rwc:16;
	unsigned	drive_step_rate:16;
	unsigned	drive_step_pulse_width:8;
	unsigned	head_settle_delay:16;
	unsigned	motor_on_delay:8;
	unsigned	motor_off_delay:8;
	unsigned	true_rdy:1;
	unsigned	start_sector_number:1;
	unsigned	motor_on:1;
	unsigned	:5;
	unsigned	:4;
	unsigned	step_pulse_per_cyl:4;
	unsigned	write_precomp_value:8;
	unsigned	head_load_delay:8;
	unsigned	head_unload_delay:8;
	unsigned	pin_34:4;
	unsigned	pin_2:4;
	unsigned	pin_4:4;
	unsigned	pin_1:4;
	unsigned	:32;
};

#define SCSI_MODE_CACHE_CONTROL_PAGE	0x08
struct scsi_mode_cache_control_page {
	unsigned	parameters_savable:1;
	unsigned	:1;
	unsigned	page_code:6;
	unsigned	page_parameter_length:8;
	unsigned	:5;
	unsigned	wce:1;				/* Write Cache Enable */
	unsigned	ms:1;				/* Multiple Selection */
	unsigned	rcd:1;				/* Read Cache Disable */
	unsigned	dr_reten_prior:4;		/* Demand Read Retention Priority */
	unsigned	write_reten_prior:4;		/* Write Retention Priority */
	unsigned	disable_prefetch_length:16;	/* Disable Prefetch Transfer Length */
	unsigned	min_prefetch:16;		/* Minumum number of blocks to prefetch */
	unsigned	max_prefetch:16;		/* Maximum number of blocks to prefetch */
	unsigned	max_prefetch_ceiling:16;	/* Upper limit of number of blocks for Max prefetch */
};

#define SCSI_MODE_NOTCH_PAGE	0x0C
struct scsi_mode_notch_page {
	unsigned	parameters_savable:1;
	unsigned	:1;
	unsigned	page_code:6;
	unsigned	page_parameter_length:8;
	unsigned	notched_drive:1;
	unsigned	logical_notch:1;
	unsigned	:6;
	unsigned	maximum_notches:16;
	unsigned	active_notch:16;
	unsigned	starting_boundary:32;
	unsigned	ending_boundary:32;
	unsigned	pages_notched_hi:32;
	unsigned	pages_notched_lo:32;
};

/* and dozens of others like the above, fill them in as you need them */

/* The structure returned by a SCSI_REQUEST_SENSE command */
struct scsi_request_sense {
	unsigned	valid:1;
	unsigned	error_code:7;
	unsigned	segment_number:8;
	unsigned	filemark:1;
	unsigned	end_of_media:1;
	unsigned	incorrect_length:1;
	unsigned	:1;
	unsigned	sense_key:4;
	unsigned char	information_bytes[4];
	unsigned char	additional_sense_length;
	unsigned char	command_information[4];
	unsigned char	additional_sense_code;
	unsigned char	additional_sense_code_qualifier;
	unsigned char	field_replaceable_unit_code;
	unsigned	sksv:1;
	unsigned	cd:1;
	unsigned	reserved:2;
	unsigned	bpv:1;
	unsigned	bit_pointer:3;
        unsigned	field_pointer:16;
	unsigned char	additional_sense_bytes[52];	/* Enough for all devices */
};

#define SCSI_SENSE_NONE			0x00
#define SCSI_SENSE_RECOVERED_ERROR	0x01
#define SCSI_SENSE_NOT_READY		0x02
#define SCSI_SENSE_MEDIUM_ERROR		0x03
#define SCSI_SENSE_HARDWARE_ERROR	0x04
#define SCSI_SENSE_ILLEGAL_REQUEST	0x05
#define SCSI_SENSE_UNIT_ATTENTION	0x06
#define SCSI_SENSE_DATA_PROTECT		0x07
#define SCSI_SENSE_BLANK_CHECK		0x08
#define SCSI_SENSE_VENDOR_UNIQUE	0x09
#define SCSI_SENSE_COPY_ABORTED		0x0a
#define SCSI_SENSE_ABORTED_COMMAND	0x0b
#define SCSI_SENSE_EQUAL		0x0c
#define SCSI_SENSE_VOLUME_OVERFLOW	0x0d
#define SCSI_SENSE_MISCOMPARE		0x0e
#define SCSI_SENSE_SOME			0x0f
/* software invented sense error codes for more specific error indications */
#define SCSI_SENSE_XTRA_NO_SENSE	0x10
#define SCSI_SENSE_XTRA_NO_TAPE		0x11
#define SCSI_SENSE_XTRA_NYREADY		0x12
#define SCSI_SENSE_XTRA_RECLENMISMATCH	0x13

/* SCSI sequential device density codes */
#define SCSI_SEQ_DENSITY_DEFAULT        0x00
#define SCSI_SEQ_DENSITY_NRZI_X3_22     0x01
#define SCSI_SEQ_DENSITY_GCR_QIC_120    0x0f
#define SCSI_SEQ_DENSITY_GCR_QIC_150    0x10
#define SCSI_SEQ_DENSITY_GCR_QIC_320    0x11
#define SCSI_SEQ_DENSITY_GCR_QIC_1350   0x12
#define SCSI_SEQ_DENSITY_DDS            0x13
#define SCSI_SEQ_DENSITY_NO_CHANGE      0x7f

#ifdef KERNEL
/*
 * The following are macros used to construct commands to pass
 * to the device dependent SCSI layer.
 */
extern struct scsi_cmd nullcmd;

#define SCSI_CMD6(_cc, _cmd)					\
			*(_cc) = nullcmd,			\
			(_cc)->cdb_len = 6,			\
			(_cc)->cdb.c6.opcode = (_cmd)

#define SCSI_CMD6_BUF(_cc, _cmd, _buf)				\
			*(_cc) = nullcmd,			\
			(_cc)->cdb_len = 6,			\
			(_cc)->cdb.c6.opcode = (_cmd),		\
			(_cc)->cdb.c6.block_addr = 0,		\
			(_cc)->cdb.c6.length = sizeof(_buf),	\
			(_cc)->data_len = sizeof(_buf),		\
			(_cc)->data_addr = (vm_offset_t)&(_buf)

#define SCSI_CMD6_BUF_ADDR(_cc, _cmd, _buf, _addr)		\
			*(_cc) = nullcmd,			\
			(_cc)->cdb_len = 6,			\
			(_cc)->cdb.c6.opcode = (_cmd),		\
			(_cc)->cdb.c6.block_addr = (_addr),	\
			(_cc)->cdb.c6.length = sizeof(_buf),	\
			(_cc)->data_len = sizeof(_buf),		\
			(_cc)->data_addr = (vm_offset_t)&(_buf)

#define SCSI_CMD10(_cc, _cmd)					\
			*(_cc) = nullcmd,			\
			(_cc)->cdb_len = 10,			\
			(_cc)->cdb.c10.opcode = (_cmd)

#define SCSI_CMD10_BUF(_cc, _cmd, _buf)				\
			*(_cc) = nullcmd,			\
			(_cc)->cdb_len = 10,			\
			(_cc)->cdb.c10.opcode = (_cmd),		\
			SCSI_SET_BADDR10(_cc, 0),		\
			SCSI_SET_BLEN10(_cc, sizeof(_buf)),	\
			(_cc)->data_len = sizeof(_buf),		\
			(_cc)->data_addr = (vm_offset_t)&(_buf)

#define SCSI_CMD10_BUF_ADDR(_cc, _cmd, _buf, _addr)		\
			*(_cc) = nullcmd,			\
			(_cc)->cdb_len = 10,			\
			(_cc)->cdb.c10.opcode = (_cmd),		\
			SCSI_SET_BADDR10(_cc, _addr),		\
			SCSI_SET_BLEN10(_cc, sizeof(_buf)),	\
			(_cc)->data_len = sizeof(_buf),		\
			(_cc)->data_addr = (vm_offset_t)&(_buf)

#define SCSI_CMD12_BUF(_cc, _cmd, _buf)				\
			*(_cc) = nullcmd,			\
			(_cc)->cdb_len = 12,			\
			(_cc)->cdb.c12.opcode = (_cmd),		\
			SCSI_SET_BADDR12(_cc, 0),		\
			SCSI_SET_BLEN12(_cc, sizeof(_buf)),	\
			(_cc)->data_len = sizeof(_buf),		\
			(_cc)->data_addr = (vm_offset_t)&(_buf)

#define SCSI_CMD12_BUF_ADDR(_cc, _cmd, _buf, _addr)		\
			*(_cc) = nullcmd,			\
			(_cc)->cdb_len = 12,			\
			(_cc)->cdb.c12.opcode = (_cmd),		\
			SCSI_SET_BADDR12(_cc, _addr),		\
			SCSI_SET_BLEN12(_cc, sizeof(_buf)),	\
			(_cc)->data_len = sizeof(_buf),		\
			(_cc)->data_addr = (vm_offset_t)&(_buf)

struct  rw_10_cmd {
        u_int   opcode          : 8;
        u_int   lun             : 3;
        u_int   dpo             : 1;            /* disable page out */
        u_int   fua             : 1;            /* force unit access (write to media before returning) */
        u_int   res0            : 2;
        u_int   ra              : 1;            /* relative addressing */
	u_int	blk_addr_hi	: 16;           /* hi 16 of 32 bit address */
	u_int	blk_addr_lo	: 16;           /* lo 16 of 32 bit address */
	u_int	res1		: 8;
	u_int	nblks_hi        : 8;
	u_int	nblks_lo        : 8;
        u_int   control         : 8;
};

#define	RW10_GET_BADDR(io)			\
	((u_int)(io)->blk_addr_hi << 16 | (u_int)(io)->blk_addr_lo)
#define	RW10_SET_BADDR(io, baddr)		\
	(io)->blk_addr_hi = (u_int)(baddr) >> 16,	\
	(io)->blk_addr_lo = (u_int)(baddr)
#define	RW10_GET_BLEN(io)			\
	((u_int)(io)->nblks_hi << 8 | (u_int)(io)->nblks_lo)
#define	RW10_SET_BLEN(io, blen)			\
	(io)->nblks_hi = (u_int)(blen) >> 8,	\
	(io)->nblks_lo = (u_int)(blen)

struct start_stop_unit_cmd {
	unsigned	opcode:8;
	unsigned	lun:3;
	unsigned	:4;
	unsigned	immediate:1;
	unsigned	:16;
	unsigned	:6;
	unsigned	loej:1;
	unsigned	start:1;
	unsigned	control:8;
};


struct mode_sense_6_cmd {
	unsigned	opcode:8;
	unsigned	lun:3;
	unsigned	:5;
	unsigned	pc:2;
	unsigned	page_code:6;
	unsigned	:8;
	unsigned	allocation_length:8;
	unsigned	control:8;
};

#define SCSI_MODE_SENSE_MAXLEN		255

#define SCSI_MODE_SENSE_PC_CURRENT	0x00
#define SCSI_MODE_SENSE_PC_CHANGEABLE	0x01
#define SCSI_MODE_SENSE_PC_DEFAULT	0x10
#define SCSI_MODE_SENSE_PC_SAVED	0x11

struct mode_select_6_cmd {
	unsigned	opcode:8;
	unsigned	lun:3;
	unsigned	pf:1;			/* Page Format CCS */
	unsigned	:3;
	unsigned	sp:1;			/* Save Page */
	unsigned	:16;
	unsigned	parameter_list_length:8;
	unsigned	control:8;
};

struct filemark_cmd {
	unsigned	opcode:8;
	unsigned	lun:3;
	unsigned	:4;
	unsigned	immediate:1;
	unsigned	count_hi:16;
	unsigned	count_lo:8;
	unsigned	control:8;
};

struct prevent_removal_cmd {
	unsigned	opcode:8;
	unsigned	lun:3;
	unsigned	:28;
	unsigned	prevent:1;
	unsigned	control:8;
};

struct rewind_cmd {
	unsigned	opcode:8;
	unsigned	lun:3;
	unsigned	:4;
	unsigned	immediate:1;
	unsigned	:24;
	unsigned	control:8;
};

struct space_cmd {
	unsigned	opcode:8;
	unsigned	lun:3;
	unsigned	:3;
	unsigned	code:2;
	unsigned	count_hi:16;
	unsigned	count_lo:8;
	unsigned	control:8;
};


struct scsi_queue {
	struct scsi_cmd	*head;
	struct scsi_cmd	*tail;
};

#define SCSI_NCONTROLLERS	4
#define SCSI_NTARGETS		8
#define SCSI_NUNITS		8

struct scsi_controller {
        int             bustype;        /* bustype */
	vm_offset_t	bus_addr;	/* address of csr registers */
	vm_offset_t	info;		/* controller information */
#ifdef	__STDC__
	int		(*init)(	/* initialize scsi controller*/
				struct scsi_controller *scp);
	int		(*shutdown)(	/* shutdown a scsi driver */
				struct scsi_controller *scp);
	int		(*command)(	/* submit a scsi command */
				struct scsi_controller *scp,
				int target,
				int action,
				struct scsi_cmd *cmd);
	unsigned long	(*minphys)(	/* return minphys() for controller */
				struct buf *bp);
#else
	int		(*init)();	/* initialize scsi controller*/
	int		(*shutdown)();	/* shutdown a scsi driver */
	int		(*command)();	/* submit a scsi command */
	unsigned long	(*minphys)();	/* return minphys() for controller */
#endif	/* __STDC__ */
	struct scsi_queue	target_queue[SCSI_NTARGETS];
};

#ifdef	MACH_KERNEL
extern zone_t	scsi_command_zone;
#endif
extern char	*scsi_sense_key_text[];
extern int	scsi_sense_key_count;

#endif /* KERNEL */

/*
 *  Miscellaneous SCSI Commands for establishing environment
 */
#define	SIOC_TEST_UNIT			_IO('S', 1)
#define	SIOC_INQUIRY			_IOR('S', 2, struct scsi_inquiry)

#ifdef notdef
#define	SIOC_CAPACITY			_IOR('S', 3, struct scsi_read_capacity)
#define	SIOC_CMD_MODE			_IOW('S', 4, int)
#define	SIOC_SET_CMD			_IOW('S', 5, struct scsi_cmd_parms)
#define	SIOC_FORMAT			_IOW('S', 6, int)
#define	SIOC_XSENSE			_IOR('S', 7, struct xsense)
#define	SIOC_RETURN_STATUS		_IOR('S', 8, int)
#define	SIOC_RESET_BUS			_IO('S', 9)
#endif

#define SIOC_SET_DEVICE			_IOW('S', 10, struct scsi_device)
#define SIOC_GET_DEVICE			_IOR('S', 11, struct scsi_device)
#define SIOC_ALLOCATE			_IOR('S', 12, int)

#ifdef notdef
#define	SIOC_MODE_SENSE_ERR_REC		_IOR('S', 0x41, struct error_recovery)
#define	SIOC_MODE_SENSE_DISCON		_IOR('S', 0x42, struct disc_recon)
#define	SIOC_MODE_SENSE_FORMAT		_IOR('S', 0x43, struct dad_form_parms )
#define	SIOC_MODE_SENSE_GEOMETRY	_IOR('S', 0x44, struct geometry_parms)
#define	SIOC_MODE_SELECT_ERR_REC	_IOW('S', 0x81, struct error_recovery)
#define	SIOC_MODE_SELECT_DISCON 	_IOW('S', 0x82, struct disc_recon)
#define	SIOC_MODE_SELECT_FORMAT		_IOW('S', 0x83, struct dad_form_parms )
#define	SIOC_MODE_SELECT_GEOMETRY	_IOW('S', 0x84, struct geometry_parms)
#endif /* notdef */

#ifdef KERNEL
#ifdef __STDC__

int
scsi_add(                  /* added scsi controller */
	int bus,
	struct scsi_controller *scp);

int
scsi_del(                  /* delete scsi controller */
	int bus);

int
scsi_command(              /* do scsi command */
        int action,
        struct scsi_cmd *cmd);

struct scsi_cmd *
scsi_command_alloc(        /* allocate scsi_cmd */
	void);

struct scsi_request_sense *
scsi_sense_alloc(          /* allocate scsi_request_sense */
	void);

void
scsi_command_free(         /* free scsi_cmd */
        struct scsi_cmd *cmd);

void
scsi_sense_free(          /* free scsi_request_sense */
        struct scsi_request_sense *sense);

void
scsi_error(                /* dump error info */
	dev_t dev,
        char *str,
	struct scsi_cmd *cmd);

void
scsi_flush(                /* flush command queue for scsi device */
        struct scsi_device sd);

unsigned long
(*scsi_minphys(            /* bus minphys() */
	struct scsi_device sd))(struct buf *bp);
#else
int  scsi_add();
int  scsi_del();
int  scsi_command();
struct scsi_cmd *scsi_command_alloc();
struct scsi_request_sense *scsi_sense_alloc();
void scsi_command_free();
void scsi_sense_free();
void scsi_error();
void scsi_flush();
unsigned long (*scsi_minphys())();
#endif /* __STDC__ */
#endif /* KERNEL */

#ifdef __cplusplus
    }
#endif

#endif /* _HP_SCSI_H_ */
