/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * MACH 3.0 and BSD support done by the Computer Systems Laboratory (CSL)
 * at The University of Utah.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: scsi_tape.c 1.18 94/11/10$
 */

#include "st.h"
#define NSCSI	NST	/* XXX */
#if NSCSI > 0

#include <sys/param.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/proc.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <sys/mtio.h>
#include <sys/errno.h>
#include <sys/malloc.h>
#ifdef	MACH_KERNEL
#include <hpsgc/mach_glue.h>
#include <machine/machspl.h>
#else
#include <hpsgc/bsd_glue.h>
#include <machine/spl.h>
#endif
#include <hpsgc/scsi.h>

/*
 * mt_dsreg values.  Consistant with OSF.
 */
#define	MT_ISCSI_FM	0x0001	/* filemark */
#define	MT_ISCSI_EOT	0x0002	/* end of partition/medium */
#define	MT_ISCSI_SM	0x0004	/* setmark */
#define	MT_ISCSI_BOT	0x0008	/* beginning of partition/medium */
#define	MT_ISCSI_EOD	0x0010	/* end of data */
#define	MT_ISCSI_POS	0x001f	/* composite of all "position" flags */

#define MT_ISCSI	0x10	/* XXX consistant with 300bsd, not OSF (8) */

#define MT_ISCSI_ATTN	0x0001	/* unit attention */
#define MT_ISCSI_ERROR	0x0002	/* error */

/* minor device breakdown */
#define STUNIT(d)	(minor(d)&0x07)			/* 3 bits for unit */
#define STREW(d)	((minor(d)>>3)&0x01)            /* 1 bit  for rewind on close */
#define STDENS(d)	((minor(d)>>4)&0xff)		/* 8 bits for density */

/* Convert unit/rewind/density to minor device number */
#define STMAKEDEV(u,r,d) (((u)&0x07)|(((r)&0x01)<<3)|(((d)&0xff)<<4))

#ifdef MACH_KERNEL
struct scsi_device scsi_sequential[] = {
        /* bus, target, lun,    unused */
	{ 0,	0,	0,	0 },		/* tape 0 */
	{ 0,	1,	0,	0 },		/* tape 1 */
	{ 0,	2,	0,	0 },		/* tape 2 */
	{ 0,	3,	0,	0 },		/* tape 3 */
	{ 0,	4,	0,	0 },		/* tape 4 */
	{ 0,	5,	0,	0 },		/* tape 5 */
	{ 0,	6,	0,	0 },		/* tape 6 */
};
#define NUNITS		(sizeof(scsi_sequential)/sizeof(scsi_sequential[0]))
#else
extern struct scsi_device scsi_sequential[];
#define NUNITS	NST
#endif

#define	INFOBYTES(bp)	(((bp)[0]<<24)|((bp)[1]<<16)|((bp)[2]<<8)|(bp)[3])

#ifdef  DBG
int stdebug = 0;
#endif

struct stinfo {
	unsigned long		block_size;	/* 0 indicates variable size mode */
	unsigned long		min_block;
	unsigned long		max_block;
	unsigned		alive:1;	/* device exists */
	unsigned		open:1;         /* device is opened */
	unsigned		online:1;       /* device is online */
	unsigned                needfmfm:1;     /* need to write two filemarks to terminate tape */
	unsigned		dontwait:1;	/* true if this device is opened with FBLOCK|FNONDELAY */
	unsigned		stickyfm:1;	/* true is tape is already positioned past filemark */
	unsigned		incorrect_length:1;	/* tape blocksize != requested size */
	struct scsi_inquiry	inquiry;
        unsigned                density_code:8;
	struct buf		raw;
        struct mtget            mt;
} stinfo[NUNITS];

#define ST_SPACE_RECORD    0
#define ST_SPACE_FILEMARK  1

int
sterror(cmd, info)
struct scsi_cmd *cmd;
struct stinfo *info;
{
	int     x;
	union {
		u_char	sense_key_specific[4];
		struct {
			u_int	fpv : 1;
			u_int	cd  : 1;
			u_int	res : 2;
			u_int	bpv : 1;
			u_int	bit : 3;
			u_int	fpf : 16;
			u_int	pad : 8;
		} data;
	} sks;

	if (!cmd->sense)
                return(SCSI_SENSE_XTRA_NO_SENSE);

#ifdef DBG 
	if (stdebug && (cmd->scsi_status == SCSI_CHECK_CONDITION)) {
		printf("sterror: CHECK CONDITION\n");
		printf("sterror: valid = %d, error code = 0x%x, segment = 0x%x\n", 
			cmd->sense->valid, cmd->sense->error_code, cmd->sense->segment_number);
		printf("sterror: filemark = %d, eom = %d, ili = %d, sense key = 0x%x\n", 
			cmd->sense->filemark, cmd->sense->end_of_media, cmd->sense->incorrect_length, cmd->sense->sense_key);
		if (cmd->sense->sense_key <= scsi_sense_key_count) printf("sterror: sense_key = %s\n", scsi_sense_key_text[cmd->sense->sense_key]);
		if (cmd->sense->valid) printf("sterror: info = 0x%x 0x%x 0x%x 0x%x\n",
			cmd->sense->information_bytes[0], cmd->sense->information_bytes[1], 
			cmd->sense->information_bytes[2], cmd->sense->information_bytes[3]);
		printf("sterror: add. sense length %d, cmd info = 0x%x 0x%x 0x%x 0x%x\n",
			cmd->sense->additional_sense_length, cmd->sense->command_information[0],
			cmd->sense->command_information[1], cmd->sense->command_information[2],
			cmd->sense->command_information[3]);
		printf("sterror: add. sense code 0x%x, add. sense code qualifier 0x%x, FRU code = 0x%x\n",
			cmd->sense->additional_sense_code, cmd->sense->additional_sense_code_qualifier,
			cmd->sense->field_replaceable_unit_code);
		printf("sterror: sense key specific: sksv = %d, cd = %d, bpv = %d, bit_pointer = %d, field_pointer = 0x%x\n",
			cmd->sense->sksv, cmd->sense->cd, cmd->sense->bpv, cmd->sense->bit_pointer, cmd->sense->field_pointer);
		if (cmd->sense->additional_sense_length) {
			printf("add sense:");
			for (x = 0; x < cmd->sense->additional_sense_length; ++x) {
				if ((x & 7) == 0)
					printf("\n");
				printf(" 0x%x", cmd->sense->additional_sense_bytes[x]);
			}
			printf("\n");
		}
	}
#endif

	/*
         * first pick off error conditions based on the additional sense codes
         * note that the pair ('additional_sense_code','additional_sense_code-qualifier')
         * is unique across sense keys except for the pair 0,0
         */

#define PAIR(a,b)      ((a)<<8|(b))

	switch (PAIR(cmd->sense->additional_sense_code, cmd->sense->additional_sense_code_qualifier)) {

	case PAIR(0,1):
fm:		info->mt.mt_dsreg |=  MT_ISCSI_FM;      /* filemark */
		info->mt.mt_dsreg &= ~MT_ISCSI_BOT;
		/*
		 * we know that the tape has moved on past this filemark,
		 * but in our internal mt_resid state we make it seem we
		 * did not for callers who inquire. just so that we remember
		 * that we are really past the filemark, we keep this flag.
		 */
		info->stickyfm = 1;
		return(0);

	case PAIR(0,2):
eot:		info->mt.mt_dsreg |=  MT_ISCSI_EOT;     /* end of tape */
		info->mt.mt_dsreg &= ~MT_ISCSI_BOT;
		return(0);

	case PAIR(0,3):
		info->mt.mt_dsreg |=  MT_ISCSI_SM;      /* save-set mark */
		info->mt.mt_dsreg &= ~MT_ISCSI_BOT;
		return(0);

	case PAIR(0,4):
		info->mt.mt_dsreg &= ~MT_ISCSI_POS;
		info->mt.mt_dsreg |=  MT_ISCSI_BOT;     /* beginning of tape */
		return(0);

	case PAIR(0,5):
eod:		info->mt.mt_dsreg |=  MT_ISCSI_EOD;     /* end of data */
		info->mt.mt_dsreg &= ~MT_ISCSI_BOT;
		return(0);

        case PAIR(0x3a,0):
		return(SCSI_SENSE_XTRA_NO_TAPE);        /* drive has no tape */

        case PAIR(0x04,0):
		return(SCSI_SENSE_XTRA_NYREADY);        /* drive in process of becoming ready */
        }

        /* pick off remaining errors based on sense key */
        switch (cmd->sense->sense_key) {

	case SCSI_SENSE_NONE:
		if (cmd->sense->filemark)
			goto fm;
		if (cmd->sense->end_of_media)
			goto eot;
		if (cmd->sense->incorrect_length) {
			info->incorrect_length = 1;
			return(0);
		}
                break;

	case SCSI_SENSE_BLANK_CHECK:
		goto eod;

	case SCSI_SENSE_UNIT_ATTENTION:
		info->mt.mt_erreg |= MT_ISCSI_ATTN;
		return (SCSI_SENSE_UNIT_ATTENTION);
        }
        /* remember we incurred some other error */
	info->mt.mt_erreg |= MT_ISCSI_ERROR;
        return(cmd->sense->sense_key);
}


void
stdone(cmd)
struct scsi_cmd *cmd;
{
	struct stinfo		*info;
	struct buf		*bp;
	struct scsi_device	sd;
	int			e, xfer;

	bp = (struct buf *)cmd->driver_info;
	info = &stinfo[STUNIT(bp->b_dev)];
	sd = cmd->device;

	if (cmd->flags & SCSI_FLUSH) {
		/* we have had some fatal error on this device */
                if (info->mt.mt_erreg) {
			bp->b_error = EIO;
			bp->b_flags |= B_ERROR;
		}
	} else
		switch(cmd->scsi_status) {

		case SCSI_GOOD:
			bp->b_resid = 0;
			break;

		case SCSI_CHECK_CONDITION:
			if (e = sterror(cmd, info)) {
				printf("st: %s\n", scsi_sense_key_text[e]);
				bp->b_flags |= B_ERROR;
				bp->b_error = EIO;
				scsi_flush(sd);
			} else {
				/* filemark, end-of-data, incomplete reads */
				if (INFOBYTES(cmd->sense->information_bytes) >= 0) {
					/* underlength: we got less than we asked for */
					bp->b_resid = INFOBYTES(cmd->sense->information_bytes);
					if (info->block_size)
						bp->b_resid *= info->block_size;
				} else {
					/* overlength: there is more in this tape block */
					info->mt.mt_resid = INFOBYTES(cmd->sense->information_bytes);
					if (info->block_size)
						info->mt.mt_resid *= info->block_size;
					info->mt.mt_resid += bp->b_bcount;
					/* we did get all we asked for */
					bp->b_resid = 0;
				}
			}
			break;

		case SCSI_CONDITION_MET:
		case SCSI_BUSY:
		case SCSI_INTERMEDIATE_GOOD:
		case SCSI_INTERMEDIATE_CONDITION_MET:
		case SCSI_RESERVATION_CONFLICT:
		case SCSI_QUEUE_FULL:
		default:
			/* This driver doesn't do any of this nonsense, so just panic */
			panic("stdone: unrecognized scsi_status");
		}

	xfer = bp->b_bcount - bp->b_resid;

	if (bp->b_flags & B_READ) {
		/* decrement residue in current tape block */
		info->mt.mt_resid -= xfer;
		if (info->mt.mt_resid < 0)
			info->mt.mt_resid = 0;
	} else if (xfer) {
		/* can write over a filemark */
		info->stickyfm = 0;
		info->mt.mt_dsreg &= ~MT_ISCSI_FM;
		info->mt.mt_resid = 0;
	}

	if (xfer)
		info->mt.mt_dsreg &= ~MT_ISCSI_BOT;

	if (cmd->sense) scsi_sense_free(cmd->sense);
	scsi_command_free(cmd);
	iodone(bp);
}


/* 
 * internal routine to wait until 'seconds' run out
 */
stwait(info, secs)
struct stinfo *info;
int *secs;
{
#define	WAITTICK 2
#define MAXTIMEOUT 30

	int wakeup();
	extern int hz;

	if (info->dontwait)
		return EWOULDBLOCK;

	if ((*secs -= WAITTICK) < 0) {
#ifdef  DBG
		if (stdebug&4) uprintf("stwait timeout...");
#endif
		return EIO;
	}

	/* wait WAITTICK seconds */
	timeout(wakeup, (caddr_t)info, hz * WAITTICK);
	return(sleep((caddr_t)info, (PZERO+1) | PCATCH) ? EINTR : 0);
}


/*
 * internal routine to rewind tape
 */
static int
strewind(dev)
dev_t dev;
{
        int unit = STUNIT(dev);
	struct stinfo *info = &stinfo[unit];
	struct scsi_cmd *cmd = scsi_command_alloc();
        int e, error = 0, secs = MAXTIMEOUT;
	int dsreg;

	*cmd = nullcmd;
	/* check if already done/working on it */
	if (!(info->mt.mt_dsreg & MT_ISCSI_BOT)) {
		dsreg = info->mt.mt_dsreg;
               	info->mt.mt_dsreg &= ~MT_ISCSI_POS;
       	        info->mt.mt_dsreg |=  MT_ISCSI_BOT;

		do {
#ifdef  DBG
			if (stdebug&4) uprintf("rewind...");
#endif
			SCSI_CMD6(cmd, SCSI_REWIND);
			((struct rewind_cmd *)(&cmd->cdb))->immediate = 1;
			cmd->device = scsi_sequential[unit];
			cmd->flags = SCSI_WAIT;
			scsi_command(SCSI_QUEUE_TAIL, cmd);
		} while ((cmd->scsi_status == SCSI_BUSY) && !(error = stwait(info, &secs)));

		if (!error && cmd->scsi_status != SCSI_GOOD && (e = sterror(cmd, info))) {
			/* tape drive would not rewind */
			uprintf("st%d: failed to rewind (%s)\n", unit, scsi_sense_key_text[e]);
			error = EIO;
		}

		if (!error) {
			/* no longer valid after tape movement */
			info->stickyfm = 0;
			info->mt.mt_resid = 0;
		} else {
			/* restore original position */
			info->mt.mt_dsreg = dsreg;
		}
	}
#ifdef  DBG
	else if (stdebug&4) uprintf("rewind reduced...");
#endif

	if (cmd->sense)
		scsi_sense_free(cmd->sense);
	scsi_command_free(cmd);
	return error;
}


/*
 * internal routine to space forward
 */
stforward(dev, count, code)
dev_t dev;
int count;
int code;
{
        int unit = STUNIT(dev);
	struct stinfo *info = &stinfo[unit];
	struct scsi_cmd *cmd = scsi_command_alloc();
        int e, error = 0, secs = MAXTIMEOUT;

#ifdef  DBG
	if (stdebug&4) uprintf("forward (%dx%d)...", count, code);
#endif
	*cmd = nullcmd;
	if (info->stickyfm && count-- > 0) {
#ifdef  DBG
		if (stdebug&4) uprintf("reduced...");
#endif
		info->mt.mt_dsreg &= ~(MT_ISCSI_FM|MT_ISCSI_BOT);
		info->stickyfm = 0;
		info->mt.mt_resid = 0;
		info->needfmfm = 0;
	}

	if (count > 0) {
		do {
			SCSI_CMD6(cmd, SCSI_SPACE);
			((struct space_cmd *)&cmd->cdb)->count_hi = count >> 8;
			((struct space_cmd *)&cmd->cdb)->count_lo = count;
			((struct space_cmd *)&cmd->cdb)->code = code;
			cmd->device = scsi_sequential[unit];
			cmd->flags = SCSI_WAIT;
			scsi_command(SCSI_QUEUE_TAIL, cmd);
		} while ((cmd->scsi_status == SCSI_BUSY) && !(error = stwait(info, &secs)));

		if (!error && cmd->scsi_status != SCSI_GOOD && (e = sterror(cmd, info))) {
			/* tape drive would not space forward */
			uprintf("st%d: failed to space forward (%s)\n", unit, scsi_sense_key_text[e]);
			error = EIO;
		}

		if (!error) {
			info->mt.mt_dsreg &= ~(MT_ISCSI_FM|MT_ISCSI_BOT);
			info->mt.mt_resid = 0;
			info->needfmfm = 0;
		}
	}

	if (cmd->sense)
		scsi_sense_free(cmd->sense);
	scsi_command_free(cmd);
        return error;
}


/*
 * internal routine to space backward
 */
static int
stbackward(dev, count, code)
dev_t dev;
int count;
int code;
{
        int unit = STUNIT(dev);
	struct stinfo *info = &stinfo[unit];
	struct scsi_cmd *cmd = scsi_command_alloc();
        int e, error = 0, secs = MAXTIMEOUT;

#ifdef  DBG
	if (stdebug&4) uprintf("backward (%dx%d)...", count, code);
#endif
	*cmd = nullcmd;
	if (info->stickyfm && count++ > 0) {
#ifdef  DBG
		if (stdebug&4) uprintf("increased...");
#endif
		info->mt.mt_dsreg |= MT_ISCSI_FM;
		info->stickyfm = 0;
		info->mt.mt_resid = 0;
		info->needfmfm = 0;
	}

	if (count > 0) {
		do {
			SCSI_CMD6(cmd, SCSI_SPACE);
			e = -(count);
			((struct space_cmd *)&cmd->cdb)->count_hi = e >> 8;
			((struct space_cmd *)&cmd->cdb)->count_lo = e;
			((struct space_cmd *)&cmd->cdb)->code = code;
			cmd->device = scsi_sequential[unit];
			cmd->flags = SCSI_WAIT;
			scsi_command(SCSI_QUEUE_TAIL, cmd);
		} while ((cmd->scsi_status == SCSI_BUSY) && !(error = stwait(info, &secs)));

		if (!error && cmd->scsi_status != SCSI_GOOD) {
			if (cmd->sense && cmd->sense->end_of_media && INFOBYTES(cmd->sense->information_bytes) == 1) {
				/* 
				 * if we backed to exactly the begining of the tape
				 * we are actually in the right place.
				 */
				info->mt.mt_dsreg |= MT_ISCSI_BOT;
			} else if (e = sterror(cmd, info)) {
				/* tape drive would not space backward */
				uprintf("st%d: failed to space backward (%s)\n", unit, scsi_sense_key_text[e]);
				error = EIO;
			}
		}

		if (!error) {
			info->mt.mt_dsreg |= MT_ISCSI_FM;
			info->mt.mt_resid = 0;
			info->needfmfm = 0;
		}
	}

	if (cmd->sense)
		scsi_sense_free(cmd->sense);
	scsi_command_free(cmd);
        return error;
}


/*
 * internal routine to write filemarks
 */
stfilemark(dev, count)
dev_t dev;
int count;
{
        int unit = STUNIT(dev);
	struct stinfo *info = &stinfo[unit];
	struct scsi_cmd *cmd = scsi_command_alloc();
        int e, error = 0, secs = MAXTIMEOUT;

	*cmd = nullcmd;
	do {
#ifdef  DBG
		if (stdebug&4) uprintf("filemark (%d)...", count);
#endif
		SCSI_CMD6(cmd, SCSI_WRITE_FILEMARK);
		((struct filemark_cmd *)&cmd->cdb)->count_hi = count >> 8;
		((struct filemark_cmd *)&cmd->cdb)->count_lo = count;
		((struct filemark_cmd *)&cmd->cdb)->immediate = 1;		
		cmd->device = scsi_sequential[unit];
		cmd->flags = SCSI_WAIT;
		scsi_command(SCSI_QUEUE_TAIL, cmd);
	} while ((cmd->scsi_status == SCSI_BUSY) && !(error = stwait(info, &secs)));

	if (!error && cmd->scsi_status != SCSI_GOOD && (e = sterror(cmd, info))) {
		/* tape drive would not write filemarks */
		uprintf("st%d: failed to write filemark (%s)\n", unit, scsi_sense_key_text[e]);
		error = EIO;
        }

	if (!error ) {
		info->mt.mt_dsreg &= ~MT_ISCSI_FM;
		info->stickyfm = 0;
		info->mt.mt_resid = 0;
		info->needfmfm = 0;
	}

	if (cmd->sense)
		scsi_sense_free(cmd->sense);
	scsi_command_free(cmd);
        return error;
}

stinit(dev)
dev_t dev;
{
	struct stinfo *info;
	struct scsi_cmd *cmd;
	struct scsi_device sd;
	int unit, secs, e, iosize, i;
	vm_offset_t v;
	struct scsi_inquiry *inquiry;
	struct scsi_request_sense *sense;

	iosize = NBPG;
	if (kmem_alloc(kernel_map, &v, iosize) != KERN_SUCCESS)
		panic("stinit: unable to allocate i/o page");

	cmd = scsi_command_alloc();
	*cmd = nullcmd;

	unit = STUNIT(dev);
	sd = scsi_sequential[unit];
	info = &stinfo[unit];
	info->alive = 0;

#ifdef  DBG
	if (stdebug&4) printf("st%d: sense...", unit);
#endif

	/* Request the sense in order to clear any unit attentions */
	sense = (struct scsi_request_sense *) v;
	SCSI_CMD6_BUF(cmd, SCSI_REQUEST_SENSE, *sense);
	cmd->device = sd;
	cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
	scsi_command(SCSI_QUEUE_TAIL, cmd);

	if (cmd->scsi_status == SCSI_DEVICE_NA) {
                if (cmd->sense)
			panic("stinit: sense data on non-available device!");
		scsi_command_free(cmd);
		kmem_free(kernel_map, v, iosize);
		return(ENXIO);
	}

#ifdef  DBG
	if (stdebug&4) uprintf("inquiry...");
#endif

	/* do an Inquiry to see if this device is what we expect */
	inquiry = (struct scsi_inquiry *) v;
	for (secs = 0; secs < 30; secs++) {
		SCSI_CMD6_BUF(cmd, SCSI_INQUIRY, *inquiry);
		cmd->device = sd;
		cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
		scsi_command(SCSI_QUEUE_TAIL, cmd);
		if (cmd->scsi_status != SCSI_BUSY)
			break;
		if (cmd->sense) scsi_command_free(cmd->sense);
		delay(1000000);
	}
	if (cmd->scsi_status != SCSI_GOOD && (e = sterror(cmd, info))) {
		if (cmd->sense) scsi_command_free(cmd->sense);
		scsi_command_free(cmd);
		kmem_free(kernel_map, v, iosize);
		return(ENXIO);
	}
	info->inquiry = *inquiry;

#ifdef  DBG
	if (stdebug&2) {
		printf("stinit: inquiry removable_media=0x%x iso_version=0x%x ecma_version=0x%x ansi_version=0x%x\n",
			inquiry->removable_media, inquiry->iso_version, inquiry->ecma_version, inquiry->ansi_version);
		printf("stinit: inquiry supports aen=0x%x relative_address=0x%x 32bit_xfers=0x%x 16bit_xfers=0x%x sync=0x%x linked cmds=0x%x cacheing=0x%x queueing=0x%x\n",
			inquiry->supports_relative_address, inquiry->supports_32bit_transfers, inquiry->supports_16bit_transfers,
			inquiry->supports_sync_transfers, inquiry->supports_linked_operations, inquiry->supports_cacheing, inquiry->supports_command_queuing);
		printf("stinit:\n");
		printf("\tvendor identification=");
		for (i = 0; i < sizeof(inquiry->vendor_identification); i++)
			printf("%c", inquiry->vendor_identification[i]);
		printf("\n");
		printf("\tproduct identification=");
		for (i = 0; i < sizeof(inquiry->vendor_identification); i++)
			printf("%c", inquiry->product_identification[i]);
		printf("\n");
		printf("\tproduct revision=");
		for (i = 0; i < sizeof(inquiry->vendor_identification); i++)
			printf("%c", inquiry->product_revision[i]);
		printf("\n");
	}
#endif

	e = (inquiry->peripheral_device_type == SCSI_SEQUENTIAL_ACCESS)
		? 0 : EIO;
	if (cmd->sense) scsi_command_free(cmd->sense);
	scsi_command_free(cmd);
	kmem_free(kernel_map, v, iosize);
	if (!e) {
		info->alive = 1;
		printf("st%d: vendor id=", unit);
		for (i = 0; i < sizeof(inquiry->vendor_identification); i++)
			printf("%c", inquiry->vendor_identification[i]);
		printf(" product id=");
		for (i = 0; i < sizeof(inquiry->product_identification); i++)
			printf("%c", inquiry->product_identification[i]);
		printf(" product revision=");
		for (i = 0; i < sizeof(inquiry->product_revision); i++)
			printf("%c", inquiry->product_revision[i]);
		printf("\n");
	}
	return(e);
}

int
stopen(dev, flag)
dev_t dev;
int flag;
{
	struct stinfo *info;
	struct scsi_cmd *cmd = (struct scsi_cmd *)0;
	struct scsi_device sd;
	int e, s, i, unit, secs, error = 0;

	/* I/O structures */
	struct scsi_inquiry *inquiry;
	struct scsi_block_limits *block_limits;
	struct scsi_request_sense *sense;
	struct stmode {
		struct {
			unsigned char	data_length;
			unsigned char	medium_type;
			unsigned 	wp:1;
			unsigned        buffered_mode:3;
			unsigned	speed:4;
			unsigned char	block_descriptor_length:8;
		} header;
		struct scsi_mode_block block;
	} *mode;

	vm_offset_t v = 0;		/* I/O area */

#define IOSIZE	\
	(((sizeof(*inquiry)+sizeof(*block_limits)+sizeof(*sense)+sizeof(*mode)) + NBPG-1) & ~(NBPG-1))

        if ((unit = STUNIT(dev)) >= NUNITS)
                return (ENXIO);

#ifdef  DBG
	if (stdebug&4) uprintf("stopen...");
#endif

	sd = scsi_sequential[unit];
	info = &stinfo[unit];
	if (!info->alive)
		return (ENXIO);

	s = splbio();

	/* cannot open already opened devices */
	if (info->open) {
		splx(s);
		return (EBUSY);
	}
	info->open = TRUE;
	splx(s);

	info->mt.mt_type = MT_ISCSI;
	info->dontwait = 0;

	if (kmem_alloc(kernel_map, &v, IOSIZE) != KERN_SUCCESS)
		panic("stopen: unable to allocate i/o page");

	cmd = scsi_command_alloc();
	*cmd = nullcmd;

#ifdef  DBG
	if (stdebug&4) uprintf("sense...");
#endif

	/* Request the sense in order to clear any unit attentions */
	sense = (struct scsi_request_sense *)v;
	SCSI_CMD6_BUF(cmd, SCSI_REQUEST_SENSE, *sense);
	cmd->device = sd;
	cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
	scsi_command(SCSI_QUEUE_TAIL, cmd);

	if (cmd->scsi_status == SCSI_DEVICE_NA) {
                if (cmd->sense)
			panic("stopen: sense data on non-available device!");
		error = ENXIO;
		goto open_done;
	}

	/* do a TEST UNIT READY operation to find tape availability */
	secs = MAXTIMEOUT;
        while (TRUE) {
#ifdef  DBG
		if (stdebug&4) uprintf("ready...");
#endif
		SCSI_CMD6(cmd, SCSI_TEST_UNIT_READY);
		cmd->device = sd;
		cmd->flags = SCSI_WAIT;
		scsi_command(SCSI_QUEUE_TAIL, cmd);
		if (cmd->scsi_status != SCSI_GOOD && (e = sterror(cmd, info))) {

                        if (info->mt.mt_erreg & MT_ISCSI_ATTN)
                                break;

                        if (e == SCSI_SENSE_XTRA_NYREADY || e == SCSI_BUSY) {
				if (error = stwait(info, &secs))
					goto open_done;
                        } else {
				uprintf("st%d: %s\n", unit, scsi_sense_key_text[e]);
				error = EIO;
				goto open_done;
                        }
                } else
                        /* device ready */
                        break;
	}

	/* if device has been reset or media has been changed, rewind the tape */
	if (info->mt.mt_erreg & MT_ISCSI_ATTN) {
		info->mt.mt_erreg = 0;
		if (error = strewind(dev))
			goto open_done;
	} else
		info->mt.mt_erreg = 0;

	inquiry = &info->inquiry;

#ifdef  DBG
	if (stdebug&4) uprintf("blimits...");
#endif

	/* find out the possible block sizes */
	block_limits = (struct scsi_block_limits *)v;
	SCSI_CMD6(cmd, SCSI_READ_BLOCK_LIMITS);
	cmd->data_addr = (vm_offset_t)block_limits;
	cmd->data_len = sizeof(*block_limits);
	cmd->device = sd;
	cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
	scsi_command(SCSI_QUEUE_TAIL, cmd);
	if (cmd->scsi_status != SCSI_GOOD && (e = sterror(cmd, info))) {
		uprintf("st%d: failed to provide block limits (%s)\n", unit, scsi_sense_key_text[e]);
		error = EIO;
		goto open_done;
	}

	info->min_block = block_limits->min_block_length;
	info->max_block = block_limits->max_block_length;
	if (info->min_block == info->max_block) {
		/*
		 * We require a block size that is a multiple of our default block length
		 */
		if (info->min_block & (SCSI_SEQUENTIAL_DEFAULT_BLOCK_LENGTH-1)) {
			uprintf("st%d: device block size is not a multiple of %d\n", unit, SCSI_SEQUENTIAL_DEFAULT_BLOCK_LENGTH);
			error = EIO;
			goto open_done;
		}
	}

#ifdef  DBG
	if (stdebug&4) uprintf("sense mode...");
#endif

	/* do a sense mode to see if the device is write protected */
	mode = (struct stmode *)v;
	SCSI_CMD6_BUF(cmd, SCSI_MODE_SENSE_6, *mode);
	cmd->device = sd;
	cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
	scsi_command(SCSI_QUEUE_TAIL, cmd);
	if (cmd->scsi_status != SCSI_GOOD && (e = sterror(cmd, info))) {
		uprintf("st%d: %s\n", unit, scsi_sense_key_text[e]);
		error = EIO;
		goto open_done;
	}
	/* verify received data length */
	if (mode->header.data_length < (sizeof(mode->header) - sizeof(mode->header.data_length))) {
		uprintf("st%d: failed to obtain device specific parameters\n", unit);
		error = EIO;
		goto open_done;
	}
	if ((flag&FWRITE) && mode->header.wp) {
		uprintf("st%d: tape is write protected\n", unit);
		error = EIO;
		goto open_done;
	}
        if (mode->header.block_descriptor_length < sizeof(mode->block)) {
                /* legal response, but we want to know the block length */
		uprintf("st%d: failed to obtain block descriptor\n", unit);
		error = EIO;
		goto open_done;
        }
	/* set block length and density */
	info->block_size   = mode->block.block_length;
	info->density_code = mode->block.density_code;
#ifdef  DBG
	if (stdebug&4) uprintf("(bs=%x,dens=%x)...", info->block_size, info->density_code);
#endif
	if (mode->header.buffered_mode != 1 || mode->block.number_of_blocks != 0 ||
        (STDENS(dev) != 0 && STDENS(dev) != mode->block.density_code)) {
#ifdef  DBG
		if (stdebug&4) uprintf("select mode...");
#endif
                /* clear reserved fields */
                mode->header.data_length = 0;
                mode->header.medium_type = 0;
                mode->header.wp = 0;
                mode->block.reserved = 0;
                /* set fields of interest */
                mode->header.buffered_mode = 1;         /* report GOOD as soon as data is in buffer */
                mode->header.speed = 0;                 /* default device speed */
                if (mode->block.density_code != SCSI_SEQ_DENSITY_DDS)
        		mode->block.density_code = STDENS(dev); /* tape density but only when not DAT */
		mode->block.number_of_blocks = 0;
		mode->header.block_descriptor_length = sizeof(mode->block);
		SCSI_CMD6_BUF(cmd, SCSI_MODE_SELECT_6, *mode);
		cmd->device = sd;
		cmd->flags = SCSI_DATA_OUT | SCSI_WAIT;
		cmd->data_len = sizeof(*mode);
		/* don't save parameters to media */
		((struct mode_select_6_cmd *)(&cmd->cdb))->sp = 0;
		((struct mode_select_6_cmd *)(&cmd->cdb))->parameter_list_length = cmd->data_len;
		scsi_command(SCSI_QUEUE_TAIL, cmd);
		if (cmd->scsi_status != SCSI_GOOD && (e = sterror(cmd, info))) {
			uprintf("st%d: failed to set mode (%s)\n", unit, scsi_sense_key_text[e]);
			error = EIO;
			goto open_done;
		}
	}

	/* if removeable media then lock in the media */
	if (info->inquiry.removable_media) {
#ifdef  DBG
		if (stdebug&4) uprintf("lock...");
#endif
		/* we don't care if we get an error */
		SCSI_CMD6(cmd, SCSI_PREVENT_REMOVAL);
		((struct prevent_removal_cmd *)(&cmd->cdb))->prevent = 1;
		cmd->device = sd;
		cmd->flags = SCSI_WAIT;
		scsi_command(SCSI_QUEUE_TAIL, cmd);
	}

open_done:
	if (cmd) {
		if (cmd->sense) scsi_sense_free(cmd->sense);
		scsi_command_free(cmd);
	}
	if (v)
		kmem_free(kernel_map, v, IOSIZE);
	if (error) {
		info->open = FALSE;
		info->mt.mt_type = 0;
	}
        info->online = info->open;
	info->needfmfm = 0;
#ifdef  DBG
	if (stdebug&4) uprintf("stopen done (%d)\n", error);
#endif
	return (error);
}


int
stclose(dev, flag)
dev_t dev;
int flag;
{
	struct stinfo *info;
	struct scsi_device sd;
	struct scsi_request_sense sense;
        int unit, rval, x, e, error = 0;

        if ((unit = STUNIT(dev)) >= NUNITS)
                return (ENXIO);

#ifdef  DBG
	if (stdebug&4) uprintf("stclose...");
#endif

	sd = scsi_sequential[unit];
	info = &stinfo[unit];

	if (!info->open) {
		error = ENODEV;
	}

	if (info->mt.mt_erreg & MT_ISCSI_ATTN) {
		/* 
		 * the device has been reset or media replaced.
		 * this means that it is unsafe to write file marks, etc.
		 */
		error = EIO;
#ifdef  DBG
		if (stdebug&4) uprintf("EIO...");
#endif
	}

	if (!error && info->online && (flag & FWRITE) && info->needfmfm) {
			/* we were writing: write two filemarks and backup over one of them if we dont rewind */
                        if (!(error = stfilemark(dev, 2)) && !STREW(dev))
				error = stbackward(dev, 1, ST_SPACE_FILEMARK);
	}
#ifdef  DBG
	  else if (stdebug&4) {
		if (!info->online)  uprintf("!online...");
		if (!(flag&FWRITE)) uprintf("!fwrite...");
		if (!info->needfmfm) uprintf("!needfmfm...");
	}
#endif

	if (!error) {
		if (STREW(dev)) {
			/* we are a rewind device, do so */
			error = strewind(dev);
		}
	}

	/* we are positioned after filemark for next open */
	if (info->stickyfm) {
		info->stickyfm = 0;
		info->mt.mt_dsreg &= ~MT_ISCSI_FM;
	}
	info->mt.mt_resid = 0;

	/* error or not, unlock the media if removeable, but do not bother to wait  */
	if (info->inquiry.removable_media) {
		struct scsi_cmd *cmd = scsi_command_alloc();
#ifdef  DBG
		if (stdebug&4) uprintf("unlock...");
#endif
		SCSI_CMD6(cmd, SCSI_PREVENT_REMOVAL);
		((struct prevent_removal_cmd *)(&cmd->cdb))->prevent = 0;
		cmd->device = sd;
		cmd->flags = SCSI_CALL;
		cmd->io_done = scsi_command_free;
		scsi_command(SCSI_QUEUE_TAIL, cmd);
	}
#ifdef  DBG
	if (stdebug&4) uprintf("stclose done (%d) stat=%x\n", error, info->mt.mt_dsreg);
#endif
	info->online = info->open = FALSE;
	return error;
}


/*
 * internal routine only so we can use physio() in stread/stwrite
 */
static int
ststrategy(bp)
struct buf *bp;
{
	struct stinfo *info;
	struct scsi_cmd *cmd;
	struct scsi_device sd;
	struct iocmd {
		unsigned	opcode:8;
		unsigned	lun:3;
		unsigned	:4;
		unsigned	fixed:1;
		unsigned	length_hi:16;
		unsigned	length_lo:8;
		unsigned	control:8;
	} *cp;
	int x, unit;
	unsigned       len;
        vm_offset_t    pa;

        if ((unit = STUNIT(bp->b_dev)) >= NUNITS)
                return (ENXIO);

	sd = scsi_sequential[unit];
	info = &stinfo[unit];

	/* 
	 * if we are in fixed block mode then we only allow
         * operations that are a multiple of our block size.
	 */
	if (info->block_size && (bp->b_bcount&(info->block_size-1)) != 0) {
		bp->b_error = EIO;
		bp->b_flags |= B_ERROR;
none:
		bp->b_resid = bp->b_bcount;
		iodone(bp);
		return;
	}

	/*
	 * if we have read a tape block which did not fill the user
	 * request, we will return to cause physio() to finish up.
	 */
	if (info->incorrect_length)
		goto none;

	if (info->mt.mt_dsreg & MT_ISCSI_EOT) {
		/* out of space */
		bp->b_error = ENOSPC;
		bp->b_flags |= B_ERROR;
		goto none;
	}

	if (bp->b_flags & B_READ && info->mt.mt_dsreg & (MT_ISCSI_FM|MT_ISCSI_SM|MT_ISCSI_EOD)) {
		/* no more data to transfer */
		goto none;
	}

	cmd = scsi_command_alloc();
	*cmd = nullcmd;
        cmd->device = sd;

	if (bp->b_flags&B_READ) 
		cmd->flags = SCSI_CALL | SCSI_DATA_IN;
	else
		cmd->flags = SCSI_CALL | SCSI_DATA_OUT;
	cmd->data_len = bp->b_bcount;
	cmd->data_addr = (vm_offset_t)bp->b_un.b_addr;
	if (bp->b_proc)
		cmd->data_map = bp;

        if (bp->b_flags & B_PHYS)
		cmd->flags |= SCSI_PHYSIO;

	cmd->io_done = stdone;
	cmd->driver_info = (vm_offset_t)bp;
	cp = (struct iocmd *)&cmd->cdb;
	if (bp->b_flags&B_READ) 
		cp->opcode = SCSI_READ;
	else
		cp->opcode = SCSI_WRITE;
	info->needfmfm = (cp->opcode == SCSI_WRITE);
	if (info->block_size) {
		cp->fixed = 1;
		len = bp->b_bcount / info->block_size;
	}
	else {
		cp->fixed = 0;
		len = bp->b_bcount;
	}
	cp->length_hi = len >> 8;
	cp->length_lo = len;

        /* no longer at beginning of tape */
        info->mt.mt_dsreg &= ~MT_ISCSI_BOT;

	scsi_command(SCSI_QUEUE_TAIL, cmd);

}


int
stread(dev, uio)
dev_t dev;
struct uio *uio;
{
	int unit;
	int retval;

#ifdef  DBG
	if (stdebug&4) uprintf("<");
#endif

        if ((unit = STUNIT(dev)) >= NUNITS)
                return (ENXIO);

	stinfo[unit].incorrect_length = 0;
	retval = physio(ststrategy, &stinfo[unit].raw, dev, B_READ, scsi_minphys(scsi_sequential[unit]), uio);
	return(retval);
}


int
stwrite(dev, uio)
dev_t dev;
struct uio *uio;
{
	int unit;
	int retval;

#ifdef  DBG
	if (stdebug&4) uprintf(">");
#endif

        if ((unit = STUNIT(dev)) >= NUNITS)
                return (ENXIO);

	stinfo[unit].incorrect_length = 0;
	retval = physio(ststrategy, &stinfo[unit].raw, dev, B_WRITE, scsi_minphys(scsi_sequential[unit]), uio);
	return(retval);
}


int
stioctl(dev, op, arg, flag)
dev_t dev;
int op;
caddr_t arg;
int flag;
{
	struct mtget *mtget;
	struct mtop *mtop;
	struct stinfo *info;
	int count, unit, error = 0;

        if ((unit = STUNIT(dev)) >= NUNITS)
                return (ENXIO);

	info = &stinfo[unit];

#ifdef  DBG
	if (stdebug&4) uprintf("ioctl (op=%x)...", op);
#endif
	switch (op) {

	case MTIOCGET:		/* tape status */
		mtget = (struct mtget *) arg;
		*mtget = info->mt;
		break;

	case MTIOCTOP:		/* tape operation */
		mtop = (struct mtop *) arg;
		count = mtop->mt_count;
		if (count <= 0)
			return (EINVAL);
#ifdef  DBG
	if (stdebug&4) uprintf("tape operation (op=%x)...", mtop->mt_op);
#endif
		switch (mtop->mt_op) {

		case MTWEOF:		/* write an end-of-file record */
                        error = stfilemark(dev, count);
			break;

		case MTFSF:		/* forward space file */
                        error = stforward(dev, count, ST_SPACE_FILEMARK);
			break;

		case MTBSF:		/* backward space file */
                        error = stbackward(dev, count, ST_SPACE_FILEMARK);
                        break;

		case MTFSR:		/* forward space record */
                        error = stforward(dev, count, ST_SPACE_RECORD);
			break;

		case MTBSR:		/* backward space record */
                        error = stbackward(dev, count, ST_SPACE_RECORD);
                        break;

		case MTOFFL:		/* rewind & place offline */
                        info->online = 0;
		case MTREW:		/* rewind */
			error = strewind(dev);
			break;

		case MTNOP:		/* no op */
			break;

		default:
			error = ENXIO;
		}
		break;

	default:
		error = ENXIO;
	}
#ifdef  DBG
	if (stdebug&4) uprintf("ioctl done (%d) stat=%x\n", error, info->mt.mt_dsreg);
#endif
	return (error);
}


/*
 * stmakedev - called when RB_ASKME is used from setconf()
 * Given a device address return a dev_t for it
 */
dev_t
stmakedev(maj, ctrlr, unit, rewind, density)
int     maj;
int     ctrlr;
int     unit;
int     rewind;
int     density;
{
        return(makedev(maj, STMAKEDEV(unit,rewind,density)));
}

#endif	/* NSCSI > 0 */
