/* 
 * Copyright (c) 1990,1991,1992,1993, 1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: cswtch.s 1.26 94/12/16$
 *	Author: Bob Wheeler & Leigh Stoller of University of Utah CSL
 */

/*
 *	Spl and context switch routines for the hppa.
 */

#include <machine/asm.h>
#include <assym.s>
#include <machine/machspl.h>
#include <machine/psl.h>
#include <mach/machine/vm_param.h>

#include <machine/break.h>

	.space	$PRIVATE$
	.subspa	$DATA$

	.import cur_pcb,data
	.import intstack_base,data
	.import istackptr,data


	.space	$TEXT$
	.subspa $CODE$


/*
 *	splx(level)
 *
 *	Set the spl level and return the previous level. If the level is 
 *	SPLHIGH then the interrupt enable bit PSW-I is cleared.
 */

	.export	splx,entry
	.proc
	.callinfo
splx
	.entry
	rsm	PSW_I,r0
	mfctl	eiem,ret0
	mtctl	arg0,eiem
	comiclr,= SPLHIGH,arg0,0
	ssm	PSW_I,r0
	bv,n	0(r2)
	.exit
	.procend


/*
 * void load_context(new_thread)
 *	thread	*new_thread
 *
 * Load the context for the first kernel thread, and go.
 */
	.export	load_context,entry
	.proc
	.callinfo no_calls

load_context

	.entry
	/*
	 * Since this is the first thread, we came in on the interrupt
	 * stack. The first thread never returns, so there is no need to
	 * worry about saving its frame, hence we can reset the istackptr
	 * back to its base.
	 */

	ldil	L%intstack_base,t1
	ldo	R%intstack_base(t1),t1

	ldil	L%istackptr,t2
	stw	t1,R%istackptr(t2)

	/*
	 * get new thread pointer and set the cur_pcb pointer
	 */
	ldw	THREAD_TOP_ACT(arg0),arg0
	ldw	ACT_PCB(arg0),t1
	ldil	L%cur_pcb,t2
	stw	t1,R%cur_pcb(t2)

	/*
	 * Restore the registers and return into the thread.
	 */
	ldw	SS_GOTO(t1),r2
	ldw	SS_R3(t1),r3
	ldw	SS_R4(t1),r4
	ldw	SS_R5(t1),r5
	ldw	SS_R6(t1),r6
	ldw	SS_R7(t1),r7
	ldw	SS_R8(t1),r8
	ldw	SS_R9(t1),r9
	ldw	SS_R10(t1),r10
	ldw	SS_R11(t1),r11
	ldw	SS_R12(t1),r12
	ldw	SS_R13(t1),r13
	ldw	SS_R14(t1),r14
	ldw	SS_R15(t1),r15
	ldw	SS_R16(t1),r16
	ldw	SS_R17(t1),r17
	ldw	SS_R18(t1),r18
	bv	0(r2)
	ldw	SS_R30(t1),r30

	.exit
	.procend


/*
 *  thread_t thread_machine_switch_context(old_thread, new_thread)
 *	thread_t	old_thread, new_thread
 *
 * Switch from one thread to another.
 */
	.export	thread_machine_switch_context,entry
	.proc
	.callinfo

thread_machine_switch_context

	.entry
	/*
	 * Save the old thread's registers.
	 */
	ldw	THREAD_TOP_ACT(arg0),t1
	ldw	ACT_PCB(t1),t1
	stw	r2,SS_GOTO(t1)
	stw	r3,SS_R3(t1)
	stw	r4,SS_R4(t1)
	stw	r5,SS_R5(t1)
	stw	r6,SS_R6(t1)
	stw	r7,SS_R7(t1)
	stw	r8,SS_R8(t1)
	stw	r9,SS_R9(t1)
	stw	r10,SS_R10(t1)
	stw	r11,SS_R11(t1)
	stw	r12,SS_R12(t1)
	stw	r13,SS_R13(t1)
	stw	r14,SS_R14(t1)
	stw	r15,SS_R15(t1)
	stw	r16,SS_R16(t1)
	stw	r17,SS_R17(t1)
	stw	r18,SS_R18(t1)
	stw	r30,SS_R30(t1)
	mfctl	pidr1,t2
	stw	t2,SS_CR8(t1)

	/*
	 * Make the new thread the current thread.
	 */
	ldw	THREAD_TOP_ACT(arg1),arg1
	ldw	ACT_PCB(arg1),t1
	ldil	L%cur_pcb,t2
	stw	t1,R%cur_pcb(t2)

	/*
	 * Restore the new thread's registers.
	 */
	ldw	SS_GOTO(t1),r2
	ldw	SS_R3(t1),r3
	ldw	SS_R4(t1),r4
	ldw	SS_R5(t1),r5
	ldw	SS_R6(t1),r6
	ldw	SS_R7(t1),r7
	ldw	SS_R8(t1),r8
	ldw	SS_R9(t1),r9
	ldw	SS_R10(t1),r10
	ldw	SS_R11(t1),r11
	ldw	SS_R12(t1),r12
	ldw	SS_R13(t1),r13
	ldw	SS_R14(t1),r14
	ldw	SS_R15(t1),r15
	ldw	SS_R16(t1),r16
	ldw	SS_R17(t1),r17
	ldw	SS_R18(t1),r18
	ldw	SS_R30(t1),r30
	ldw	SS_CR8(t1),t2
	bv	0(r2)
	mtctl	t2,pidr1

	.exit
	.procend


/*
 *  thread_machine_start_self()
 *
 * Start a new thread by calling thread_start_self(start_pos).
 */
	.export	thread_machine_start_self,entry
	.proc
	.callinfo

thread_machine_start_self

	.entry
	.import thread_start_self,code
	ldil	L%thread_start_self,r1
	ldo	R%thread_start_self(r1),r1
	bv	0(r1)
	copy	r10,arg0

	.exit
	.procend


	.end
