/* 
 * Copyright (c) 1990,1991,1992,1993,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: hp800_init.c 1.37 95/03/02$
 */

/*
 * Low machine initialization routine.
 *
 * Some boot mechanism brought the kernel into physical memory and transfered
 * control to a low-level routine in locore. This routine then set up a 
 * stack pointer and a global data pointer for this routine.
 * 
 * This routine is running in real mode, interrupts are off and the cache
 * is consistent. We need to set up the VM system and perform other tasks 
 * before going to virtual mode.
 * 
 * This routine returns the page number of the next available physical address.
 */

#include <mach/machine/vm_types.h>
#include <mach/machine/kern_return.h>
#include <mach/vm_prot.h>
#include <machine/pmap.h>

#include <mach/vm_param.h>
#include <machine/pdc.h>
#include <machine/regs.h>
#include <machine/act.h>
#include <machine/cpu.h>
#include <machine/iodc.h>
#include <machine/iomod.h>

#include <md.h>
#if NMD > 0
#include <hpcio/mdvar.h>
#endif

/* Interrupt stack pointer; zero while in interrupts */
vm_offset_t istackptr = 0;

/*
 * Random variables that have to go into the data segment of the kernel
 */
int cputype = 0;
double fpu_zero = 0.0;
int copr_sfu_config = 0;
int fpcopr_version = 0;

struct {
	struct parisc_saved_state st;
	int extra[16];
} tmp_saved_state = { 0 };

/*
 * memory limits
 */
extern u_int mem_size;		/* from busconf */
extern u_int io_size;		/* from busconf */

int vm_on = 0;
vm_offset_t avail_start = 0;
vm_offset_t avail_end = 0;
vm_offset_t avail_next = 0;
unsigned int avail_remaining = 0;
vm_offset_t virtual_avail = 0;
vm_offset_t virtual_end = 0;

extern long text_start;
extern long text_end;
extern long equiv_end;
extern long io_end;

extern long intstack_top;

/*
 * pointers in kernel load image
 */
extern char etext;
extern char edata;
extern char end;
extern void start_text();

vm_offset_t phys_first_addr, phys_last_addr;
vm_offset_t emalloc_base, emalloc_limit;

/*
 * location of hpmc trap vector code in locore.s. Must look like a function
 * so the linker knows that it's in $TEXT$
 */
extern void i_hpmach_chk();
int (*pdc)() = (int (*)())0;

/*
 * Size of instruction and data cache lines in this machine. 
 */
int dcache_line_size = 0;
int dcache_block_size = 0;
int dcache_line_mask = 0;
int dcache_size   = 0;
int dcache_base   = 0;
int dcache_stride = 0;
int dcache_count  = 0;
int dcache_loop   = 0;

int icache_line_size = 0;
int icache_block_size = 0;
int icache_line_mask = 0;
int icache_base   = 0;
int icache_stride = 0;
int icache_count  = 0;
int icache_loop   = 0;

struct	iodc_data cons_iodc = { 0 };
struct	device_path cons_dp = { 0 };

extern char version[];

parisc_init()
{
	int hpmc_br_instr;
	int *p = (int *) i_hpmach_chk;
	register struct mapping *mp;
	int i;
	vm_offset_t addr;
	int pdcerr;
	struct pdc_hpa pdc_hpa;
	struct pdc_coproc pdc_coproc;
	struct pdc_cache pdc_cache;
	struct pdc_model pdc_model;
	struct pdc_iodc_read pdc_iodc;
	extern int crashdump();
#ifdef	BTLB
	struct pdc_btlb pdc_btlb;
#endif
#ifdef	HPT
	struct pdc_hwtlb pdc_hwtlb;
	extern struct hpt_entry *hpt_table;
	extern int usehpt;
#endif	

	pdc = (int (*)())PAGE0->mem_pdc;

	delay_init();
	pdc_console_init(1);
	printf("%s\n", version);

	/*
	 * Determine what the boot program is using as its console
	 * so that we can use the same device.
	 */
	pdcerr = (*PAGE0->mem_pdc)(PDC_IODC, PDC_IODC_READ, &pdc_iodc,
	                           PAGE0->mem_cons.pz_hpa, PDC_IODC_INDEX_DATA,
	                           &cons_iodc, sizeof(cons_iodc));
	if (pdcerr == 0)
		bcopy((char *)&PAGE0->mem_cons.pz_dp, (char *)&cons_dp,
		      sizeof(struct device_path));
	else
		printf("Warning: can't id console boot device (PDC Ret'd %d)\n",
		       pdcerr);

#ifdef PDCDUMP
	/*
	 * Setup the transfer of control addr to point to the crash dump
	 * initialization code.
	 */
	PAGE0->ivec_toc    = crashdump;
#else
	/* XXX point ivec_toc somewhere else */
#endif
	PAGE0->ivec_toclen = 0;

	/*
	 * get cache parameters from the PDC
	 */
	(*PAGE0->mem_pdc)(PDC_CACHE, PDC_CACHE_DFLT, &pdc_cache);

	dcache_line_size = pdc_cache.dc_conf.cc_line * 16;
	dcache_line_mask = dcache_line_size - 1;
	dcache_block_size = dcache_line_size * pdc_cache.dc_conf.cc_block;

	dcache_size = pdc_cache.dc_size;
	dcache_base = pdc_cache.dc_base;
	dcache_stride = pdc_cache.dc_stride;
	dcache_count = pdc_cache.dc_count;
	dcache_loop = pdc_cache.dc_loop;

	icache_line_size = pdc_cache.ic_conf.cc_line * 16;
	icache_line_mask = icache_line_size - 1;
	icache_block_size = icache_line_size * pdc_cache.ic_conf.cc_block;

	icache_base = pdc_cache.ic_base;
	icache_stride = pdc_cache.ic_stride;
	icache_count = pdc_cache.ic_count;
	icache_loop = pdc_cache.ic_loop;

	/*
	 * purge TLBs and flush caches
	 */
	ptlball(&pdc_cache);
	
#ifdef	BTLB
        /*
         * get block tlb information for clearing
         */
	pdcerr = (*pdc)(PDC_BLOCK_TLB, PDC_BTLB_DEFAULT, &pdc_btlb);
	
        if (pdcerr != 0)
                printf("Warning: PDC_BTLB call Ret'd %d\n", pdcerr);

	switch (pdc_btlb.finfo.num_c) {
	/* S-Chip specific */
	case 0:
		for (i = 0; i < pdc_btlb.finfo.num_i; i++)
			purge_block_itlb(i);
		for (i = 0; i < pdc_btlb.finfo.num_d; i++)
			purge_block_dtlb(i);
		break;
	/* L-Chip specific */
	case 8:
		for (i = 0; i < pdc_btlb.finfo.num_c; i++)
			purge_L_block_ctlb(i);
		break;
	/* T-Chip specific */
	case 16:
		for (i = 0; i < pdc_btlb.finfo.num_c; i++)
			purge_block_ctlb(i);
		break;
	default:
		panic("unrecognized block-TLB, cannot purge block TLB(s)");
		/* NOTREACHED */
	}
#endif
	fcacheall();

	/*
	 * get the cpu type
	 */
	(*PAGE0->mem_pdc)(PDC_MODEL, PDC_MODEL_INFO, &pdc_model);
	cputype = pdc_model.hvers >> 4;

	switch (cputype) {

	case HP_705:
	case HP_710:
        case HP_715_33:
        case HP_715_50:
        case HP_715T_50:
        case HP_715S_50:
        case HP_715_75:
	case HP_720:
	case HP_725:
	case HP_730:
	case HP_735:
	case HP_735_125:
	case HP_750:
	case HP_755:
		break;

	case HP_712_60:
	case HP_712_80:
	case HP_712_3:
	case HP_712_4:
	case HP_715_100:
		break;

	case HP_810:
	case HP_820:
	case HP_825:
	case HP_832:
	case HP_840:
	case HP_850:
	case HP_855:
	case HP_815:
	case HP_835:
	case HP_845:
		break;

	case HP_850_OLD:
		cputype = HP_850;
		break;

	default:
		cputype = 0;
		break;
	}

	cpuinfo(&pdc_cache);
#if 1
	if (dcache_line_size != CACHE_LINE_SIZE)
		printf("WARNING: data cache line size = %d bytes, %s\n",
		       dcache_line_size, "THIS IS *VERY* BAD!");
#endif
	/*
	 * Get the instruction to do branch to PDC_HPMC from PDC.  If
	 * successful, then insert the instruction at the beginning
	 * of the HPMC handler.
	 */
	if ((*PAGE0->mem_pdc)(PDC_INSTR, PDC_INSTR_DFLT, &hpmc_br_instr) == 0)
		p[0] = hpmc_br_instr;
	else
		p[0] = 0;

	/* 
	 * Now compute the checksum of the hpmc interrupt vector entry
	 */
	p[5] = -(p[0] + p[1] + p[2] + p[3] + p[4] + p[6] + p[7]);

	/*
	 * configure the devices including memory. Passes back size of 
	 * physical memory in mem_size.
	 */
	busconf();

	/* 
	 * Zero out BSS of kernel before doing anything else. The location
	 * pointed to by &edata is included in the data section.
	 */
	bzero((vm_offset_t) &edata + 4, (vm_offset_t) &end - 
	      (vm_offset_t) &edata - 4);

        /*
         * Locate any coprocessors and enable them by setting up the CCR.
         * SFU's are ignored (since we dont have any).  Also, initialize
         * the floating point registers here.
         */
        if ((pdcerr = (*pdc)(PDC_COPROC, PDC_COPROC_DFLT, &pdc_coproc)) < 0)
                printf("Warning: PDC_COPROC call Ret'd %d\n", pdcerr);
        copr_sfu_config = pdc_coproc.ccr_enable;
        mtctl(CR_CCR, copr_sfu_config & CCR_MASK);
        fprinit(&fpcopr_version);
	fpcopr_version = (fpcopr_version & 0x003ff800) >> 11;
        mtctl(CR_CCR, 0);

        /*
         * Clear the FAULT light (so we know when we get a real one)
         * PDC_COPROC apparently turns it on (for whatever reason).
         */
        pdcerr = PDC_OSTAT(PDC_OSTAT_RUN) | 0xCEC0;
        (void) (*pdc)(PDC_CHASSIS, PDC_CHASSIS_DISP, pdcerr);

#ifdef TIMEX
	/*
	 * Enable the quad-store instruction.
	 */
	pdcerr = (*pdc)(PDC_MODEL, PDC_MODEL_ENSPEC,
			&pdc_model, pdc_model.pot_key);
	if (pdcerr < 0)
		printf("Warning: PDC enable FP quad-store Ret'd %d\n", pdcerr);
#endif

	/*
	 * on the hp800 the value in &etext is a pointer to the last word
	 * in the text section. Similarly &edata and &end are pointers to
	 * the last words in the section. We want to change this so that 
	 * these pointers point past the sections that they terminate.
	 */
	text_start = (long) trunc_page((vm_offset_t) &start_text);
	text_end = (long) round_page((vm_offset_t) &etext + 4);

	/*
	 * before we go to all the work to initialize the VM see if we really 
	 * linked the image past the end of the PDC/IODC area.
	 */
	if (text_start < 0x10800)
		panic("kernel text mapped over PDC and IODC memory");

	/*
	 * find ranges of physical memory that isn't allocated to the kernel
	 */
	avail_start = round_page((vm_offset_t) &end);
	avail_end = trunc_page(mem_size);
	phys_first_addr = 0;
	phys_last_addr = avail_end;

	/* 
	 * Zero the rest of memory rounding to next page boundary.
	 */
#if NMD > 0
	if (avail_end > (MDMAXBYTES + MDLEAVEBYTES)) {
		long tsize;

		tsize = avail_end - trunc_page(MDMAXBYTES) - avail_start;
		bzero(avail_start, tsize);
	} else
#endif
	bzero(avail_start, (avail_end - avail_start));

	/*
	 * bootstrap the rest of the virtual memory system
	 */
#ifdef MAXMEMBYTES
	if ((avail_end - avail_start) > MAXMEMBYTES) {
		mem_size  = trunc_page(MAXMEMBYTES);
		avail_end = mem_size;
	}
#endif

#ifdef HPT
	/*
	 * If we want to use the HW TLB support, ensure that it exists.
	 */
	if (usehpt &&
	    !((*pdc)(PDC_TLB, PDC_TLB_INFO, &pdc_hwtlb) == 0 &&
	      (pdc_hwtlb.min_size || pdc_hwtlb.max_size))) {
		printf("Warning: no HW TLB support\n");
		usehpt = 0;
	}
#endif

	pmap_bootstrap(&avail_start, &avail_end);

#if NMD > 0
	if (avail_end > (MDMAXBYTES + MDLEAVEBYTES)) {
		avail_end -= trunc_page(MDMAXBYTES);
		mdbase = avail_end;
	}
#endif

	/*
	 * set limits on virtual memory and kernel equivalenced memory
	 */
	virtual_avail = avail_end;
	virtual_end = 0x3fff0000;

	/*
	 * pmap_bootstrap allocated memory for VM data structures that
	 * must be equivalently mapped.
	 */
	equiv_end = (long) round_page(avail_start);
	io_end = 0xF0000000;	/* XXX */

#ifdef	BTLB
	/*
	 * Do block mapping. We are mapping from 0, up through the first
	 * power of 2 address above the end of the equiv region. This 
	 * means some memory gets block mapped that should not be, but
	 * so be it (we make the text writable also :-)). We do this to
	 * conserve block entries since we hope to use them for other
	 * purposes (someday).
	 *
	 * XXX must be at least 512k for all current BTLBs.
	 * Should never be a problem with Mach...
	 */
	addr = equiv_end;
	if (addr != 1 << log2(addr))
		addr = 1 << log2(addr);
	if (addr < 512 * 1024)
		addr = 512 * 1024;
	/*
	 * Readjust avail_{next,remaining} to skip the space between
	 * the real equiv_end and the rounded value.  We use this space
	 * for other equivalently-mapped data structures.
	 */
	avail_next = addr;
	avail_remaining = atop(avail_end - avail_next);
	emalloc_base = equiv_end;
	emalloc_limit = addr;
	equiv_end = addr;

	if (pdc_btlb.finfo.num_c)
		printf("%d combined", pdc_btlb.finfo.num_c);
	else
		printf("%d/%d I/D", pdc_btlb.finfo.num_i, pdc_btlb.finfo.num_d);
	printf(" BTLB entries found.  Block mapping up to 0x%x (0x%x)\n",
	       emalloc_limit, emalloc_base);

	/*
	 * XXX L-CHIP vs T-CHIP vs S-CHIP difference in Block TLB insertion.
	 */
	switch (pdc_btlb.finfo.num_c) {
	/* S-CHIP */
	case 0:
		pmap_block_map(0, addr, VM_PROT_ALL, 0, BLK_ICACHE);
		pmap_block_map(0, addr, VM_PROT_READ|VM_PROT_WRITE,
			       0, BLK_DCACHE);
		break;
	/* L-CHIP */
	case 8:
		pmap_block_map(0, addr, VM_PROT_ALL, 0, BLK_LCOMBINED);
		break;
	/* T-CHIP */
	case 16:
		pmap_block_map(0, addr, VM_PROT_ALL, 0, BLK_COMBINED);
		break;
	default:
		panic("unrecognized block-TLB, cannot map kernel");
		/* NOTREACHED */
	}
#else
	/* XXX resurrect old code here */
	panic("Need block TLBs right now");
#endif

#ifdef	HPT
	/*
	 * Turn on the HW TLB assist.
	 */
	if (usehpt) {
		pdcerr = (*pdc)(PDC_TLB, PDC_TLB_CONFIG, &pdc_hwtlb, hpt_table,
				sizeof(struct hpt_entry) * PARISC_HASHSIZE,
				PDC_TLB_WORD3);
		if (pdcerr) {
			printf("Warning: HW TLB init failed (%d), disabled\n",
			       pdcerr);
			usehpt = 0;
		} else
			printf("HW TLB initialized (%d entries at 0x%x)\n",
			       PARISC_HASHSIZE, hpt_table);
	}
#endif
	/*
	 * map the PDC and IODC area for kernel read/write
	 * XXX - should this be read only?
	 */
	(void) pmap_map(0, 0, text_start, VM_PROT_READ | VM_PROT_WRITE);

	/*
	 * map the kernel text area.
	 */
#ifdef KGDB
	(void) pmap_map(text_start, text_start, text_end, 
			VM_PROT_READ | VM_PROT_EXECUTE | VM_PROT_WRITE);
#else
	(void) pmap_map(text_start, text_start, text_end, 
			VM_PROT_READ | VM_PROT_EXECUTE);
#endif
	/*
	 * map the data section of the kernel
	 */
	(void) pmap_map(text_end, text_end, equiv_end,
			VM_PROT_READ | VM_PROT_WRITE);

#ifndef IO_HACK
	/*
	 * map the I/O pages
	 */
	(void) pmap_map(trunc_page(io_size), trunc_page(io_size),
			0, VM_PROT_READ | VM_PROT_WRITE);
#endif

#if 0
	/*
	 * map the breakpoint page
	 */
	(void) pmap_map(break_page, break_page, break_page+PARISC_PAGE_SIZE,
			VM_PROT_READ | VM_PROT_EXECUTE);
#endif

	/*
	 * map the interrupt stack red zone.
	 */
	addr = trunc_page((vm_offset_t) &intstack_top);
	(void) pmap_map(addr, addr, addr + PAGE_SIZE, VM_PROT_READ);

	vm_on = 1;
}

cpuinfo(cp)
	struct pdc_cache *cp;
{
	extern u_int itick_per_n_usec, n_usec;
	register int mhzint;		/* integer part of clock speed */
	register char *mhzfra;		/* fractional part of clock speed */
	int cpuspd;

	/*
	 * Determine our clock speed in a machine-independent way.
	 * We carry it out to 3 decimal places and round up to 2.
	 * At the end of this block, `mhzint' and `mhzfra' are set.
	 */
	{
		register char *nums, *cpfra;
		register int nindx;

		nums = "0123456789";
		mhzfra = cpfra = ".00";
		if ((mhzint = (itick_per_n_usec*1000/n_usec) % 1000 + 5) > 9) {
			cpfra++;
			*cpfra++ = nums[(mhzint / 100) % 10];
			if ((nindx = (mhzint / 10) % 10) != 0)
				*cpfra++ = nums[nindx];
		}
		*cpfra = '\0';
		mhzint = itick_per_n_usec / n_usec;
	}

	cpuspd = mhzint;
	printf("HP9000/");
	switch (cputype) {
	case HP_705:
		printf("705");
		break;
	case HP_710:
		printf("710");
		break;
	case HP_720:
		printf("720");
		break;
	case HP_730:
		printf("730");
		break;
	case HP_750:
		printf("750");
		break;

	case HP_715_33:
	case HP_715_50:
	case HP_715T_50:
	case HP_715S_50:
	case HP_715_75:
		printf("715");
		break;
	case HP_725:
		printf("725");
		break;
	case HP_735:
	case HP_735_125:
		printf("735");
		break;
	case HP_755:
		printf("755");
		break;

	case HP_712_60:
	case HP_712_80:
		printf("712");
		break;

	case HP_835:
		printf("834/835");
		break;
	case HP_845:
		printf("845");
		break;

	default:
		printf("ID%x", cputype);
		cpuspd = 0;
		break;
	}
	printf(" (%d%sMhz PA-RISC", mhzint, mhzfra);
	if (cp->dc_conf.cc_sh == 0)
		printf(", %dK Icache, %dK Dcache", cp->ic_size / 1024,
		       cp->dc_size / 1024);
	else
		printf(", %dK shared I/D cache", cp->ic_size / 1024);
	if (cp->dt_conf.tc_sh == 0)
		printf(", %d entry ITLB, %d entry DTLB", cp->it_size,
		       cp->dt_size);
	else
		printf(", %d entry shared TLB", cp->it_size);
	printf(")\n");
	if (cpuspd == 0)
		printf("*** WARNING: unrecognized CPU type, good luck :-)\n");
}
