/* 
 * Copyright (c) 1991, 1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: cioreg.h 1.3 92/05/22$
 * 	Author: Bob Wheeler, Leigh Stoller, University of Utah CSL
 */

/*
 *	CIO architected structures and constants
 */

#define v_int	volatile int

/*
 * per-controller data structure
 */
struct cio_chan {
	int		   type;	/* Device type */
	int		   ciounit;	/* CIO bus this controller is on */
	int		   channel;	/* slot in CIO this controller is on */
	int		   alive;	/* controller autoconfigured? */
	int		   notified;	/* driver has been notified */
	int		   int_level;	/* level to interrupt for completion */
	struct bus         *device;	/* device drivers */
	struct ciodio      *dio;	/* Direct I/O address for device */
	struct ciosubc     *subc;	/* Subchannel address for device */
	caddr_t		   chain;	/* Pointer to chain ram quad */
	struct ciohpa	   *hpa;	/* CIO hard physical address space */
	struct cio_request *active;	/* active request */
	int		   debug;	/* print CIO debugging stuff */
};


/*
 * per-request structure
 */
struct cio_request {
	struct	dmaquad     *cmdq;	/* pointer to DMA quad */
	struct	statquad    *stat;	/* status quad for completion */
	struct  dmaquad	    *quad;	/* dma quads for scatter/gather */
	int		    ciounit;	/* cio unit to queue request on */
	int		    channel;	/* cio channel for request */
	int		    polled;     /* device needs ARQ mask set */
};

#define CIOBYTES		16*NBPG
#define	CIO_QUAD_ALIGN		16
#define	CIO_BUFFER_ALIGN	64
#define	CIO_MIN_BUFFER		64

#define CIO_QUAD_SIZE		16	/* size of a "real" dma quad */
#define CIO_TIME_OUT		150000	/* timeout for sync CIO in 10 usec */



/*
 * CIO Direct I/O space structure
 *
 * The CIO defines certain bits in the address to be control lines! Thus
 * this structure is a bit strange. In general you want to use the normal
 * entry. If you use cend it will assert the C_END line on the bus. Using
 * cbyte will assert the C_BYT line and using cendbyte asserts both C_END
 * and C_BYTE. 
 */
struct  cioquad {
	v_int	rsvd1[3];  
	v_int	normal;
	v_int	rsvd2[3];  
	v_int 	cend;
	v_int	rsvd3[3];  
	v_int	cbyte;
	v_int	rsvd4[3];  
	v_int	cendbyte;
};

struct	ciodio {
	struct	cioquad	data;
	struct	cioquad	ctlsens;
	struct	cioquad	order;
	struct	cioquad	statcmd;
};

/*
 * CIO subchannel state structure
 *
 * This is the subchannel state structure defined in the SPA of the CIO card.
 */
struct	ciosubc {
	v_int	reserved1[4];		/* words 0-3 are reserved */
	v_int	currquad_link;		/* current quad's link field */
	v_int	currquad_cmd;		/* current quad's command field */
	v_int	currquad_addr;		/* current quad's address field */
	v_int	currquad_count;		/* current quad's count field */
	v_int	reserved2[5];		/* word 8-12 are reserved */
	v_int 	status;			/* subchannel status register */
	v_int 	control;		/* subchannel control register */
	v_int	reserved3;		/* word 15 is reserved */
	v_int	subch_data[8];		/* RTS return buffer */
	v_int	reserved4[40];		/* words 24-63 are reserved */
};


/*
 * fields defined in ciosubc subchannel status field
 */
#define SUBC_STAT_SSTAT	0xf0000000	/* subch status (see SSTAT values) */
#define SUBC_STAT_RDY	0x00100000	/* subchannel ready bit */
#define SUBC_STAT_LCN	0x000ffff0	/* current logical channel */

#define SSTAT_UNKNOWN	0		/* AES, LCD, ERT or unknown RTS code */
#define SSTAT_PARITY	1		/* Parity error on CIO read data */
#define SSTAT_NOMYAD	2		/* no MYAD asserted */
#define SSTAT_INTERNAL	4		/* internal error */
#define SSTAT_DMA	5		/* unknown DMA command */
#define SSTAT_SRQ	6		/* SRQ on inactive sub channel */
#define SSTAT_RTS	7		/* RTS overrun */

/*
 * fields defined in the ciosubc subchannel control register
 */
#define SUBC_CTRL_PARITY 0x80000000	/* subchannel supports parity */
#define SUBC_CTRL_LOGMUX 0x40000000	/* subchannel is log multiplexed */
#define SUBC_CTRL_SCHAIN 0x20000000	/* subchannel/logchannel execution */
#define SUBC_CTRL_BREAK	 0x10000000	/* logical channel break pending */
#define SUBC_CTRL_RTS	 0x08000000	/* channel receiving RTS data */
#define SUBC_CTRL_BLKSIZ 0x0000ffff	/* blocksize for this subchannel */

/*
 * Each CIO card has 16 subchannels associated with it.
 */
#define MAX_CIO_CHAN	16
#define CIO_CHAIN_SIZE	(MAX_CIO_CHAN * 16)
#define CIO_CHAIN_ALIGN	(MAX_CIO_CHAN * 16)

/*
 * CIO soft physical address space.
 *
 * Each channel has a direct I/O space and a channel subspace defined in the 
 * SPA of the CIO card.
 */
struct ciospa {
	struct ciodio   direct[16];	/* Direct I/O pages */
	struct ciosubc	subc[16];	/* Subchannel I/O pages */
};



/*
 * CIO hard physical address space
 *
 * This is the same as the iomod structure defined in iomod.h Some of the 
 * auxiliary register set is device specific. 
 */

struct ciohpa {
	u_int	reserved0;
	v_int	io_eim;		/* (RW) Controls EIM destination */
	v_int	io_dc_add_data;	/* write address of IODC to read IODC data */
	u_int	reserved1[6];
	caddr_t	io_spa_address;	/* (WO) SPA space; 0-20:addr, 24-31:iodc_spa */
	u_int	reserved2[2];
	v_int	io_command;	/* (WO) module commands (see below) */
	v_int	io_status;	/* (RO) error returns (see below) */
	u_int	reserved3[2];
	v_int	interrupt;	/* (RW) interrupt source register */
	caddr_t	chain_ram_base;	/* start address of chain RAM */
	v_int	sub_mask_clr;	/* (RW) clear bits in the subch mask reg */
	v_int	sub_mask_set;	/* (RW) set bits in the subch mask reg */
	v_int	diagnostic;	/* (RW) diagnostic use (reserved) */
	u_int	reserved4[2];
	caddr_t	nmi_address;	/* (RW) address to send MNI transaction to */
	v_int	nmi_data;	/* (RW) data to send for NMI */
};


/*
 * CIO HPA io_status bits
 *
 * These are general status bits for the whole CIO card. You must check these
 * bits after each direct I/O command to the card since no interrupt is 
 * generated for direct I/O errors.
 */

#define IO_STATUS_DBYTE	 0x80000000	/* DA Asserted D_BYT */
#define IO_STATUS_DEND	 0x40000000	/* DA Asserted D_END */
#define IO_STATUS_BR	 0x20000000	/* Burst request */
#define IO_STATUS_MYAD	 0x10000000	/* DA Asserted MYAD */
#define IO_STATUS_PEHI	 0x08000000	/* Parity error in high byte */
#define IO_STATUS_PELO	 0x04000000	/* Parity error in low byte */
#define IO_STATUS_FAIL	 0x02000000	/* No power */
#define IO_STATUS_LOST   0x01000000	/* lost power in the past */
#define IO_STATUS_MODDEP 0x00ff0000	/* module dependent diagnostic bits */
#define IO_STATUS_ME	 0x00000080	/* module error */
#define IO_STATUS_RY	 0x00000040	/* ready for command */
#define IO_STATUS_MSTAT	 0x0000003f	/* module status (see MERR values) */

#define MERR_DEPEND	0x00		/* catastrophic internal error */
#define MERR_LOOPBACK	0x04		/* loopback error */
#define MERR_DATA	0x05		/* data parity error */
#define MERR_PROTOCOL	0x06		/* protocol error on bus */
#define MERR_ADDRESS	0x07		/* no slave responded to address */
#define MERR_NOARQ	0x18		/* ARQ asserted but no one responded */


/*
 * CIO DMA
 *
 * These structures are used for CIO DMA. See the description below...
 *
 * One big problem is that all the DMA quads must have physical addresses and
 * we often have to add quads into the chain to do scatter/gather operations
 * when the data crosses physical page boundaries. As a result, we actually 
 * keep two copies of the chain information around, the first copy is loaded
 * with virtual addresses. The second copy is the copy handed to the hardware 
 * with physical addresses, additional quads for scatter/gather, different
 * counts for scatter/gather and so forth. The routine prepchain sets up the 
 * scatter/gather stuff and the physical addresses.
 */

/*
 * dmaquads are the main quads linked together to specify a DMA operation to
 * the CIO. When the command is CIODMA_LNKST then use the linkquad instead
 * of the dmaquad.
 */
struct dmaquad {
	/*
	 * physical addressed quad information
	 */
	struct	dmaquad *plink;		/* physical address of next quad */
	u_int	pcommand;		/* DMA command, see ciodma codes */
	caddr_t	paddress;		/* physical address of data buffer */
	int     pcount;			/* byte count to transfer */

	/*
	 * virtual addressed quad information
	 */
	struct	dmaquad *link;		/* virtual address of next quad */
	u_int	command;		/* DMA command, see ciodma codes */
	caddr_t	address;		/* virtual address of data buffer */
        int 	count;			/* byte count to transfer */
};

/*
 * This is the quad to use when the command is CIODMA_LNKST. The status of
 * the last dmaquad will be placed into the quad pointed to by stat and this
 * quad will then be linked into the head of status quads in the semaphore
 * quad pointed to by semq.
 */
struct linkquad {
	/*
	 * physical addressed quad information
	 */
	struct dmaquad 	*plink;		/* physical address of next quad */
	u_int	 	pcommand;	/* DMA command must be CIODMA_LNKST */
	struct semaquad *psema;		/* physical pointer to semaquad */
	struct statquad *pstat;		/* physical pointer to statquad */

	/*
	 * virtual addressed quad information
	 */
	struct dmaquad 	*link;		/* virtual address of next quad */
	u_int	 	command;	/* DMA command must be CIODMA_LNKST */
	struct semaquad *sema;		/* virtual pointer to semaquad */
	struct statquad *stat;		/* virtual pointer to statquad */
};

/*
 * This is a semaphore quad. The first field head_sema is an hp800 ldcws
 * semaphore. When 0, the hardware is updating the linked list of status
 * quads pointed to by head_next.
 */
struct semaquad {
	int	 head_sema;		     /* Semaphore for hardware. */
	volatile struct statquad *head_next; /* Link address for hardware */
	int	 undefined[2];
};

/*
 * The status quad is filled in after processing a linkquad. The link field
 * is filled in with the value from the head_next field of the semaphore quad
 * and the value of head_next is filled in with the address to the stat quad.
 * In essence putting the statquad at the head of the linked list of status
 * quads pointed to by head_next.
 *
 * The status value is filled in with the subchannel status register and the 
 * last_link field and count fileds are  filled in with the values of the link
 * and count fields of the previous dmaquad entry.
 */
struct statquad {
	volatile struct statquad *link;	/* link to next return */
	int		status;		/* status of subc after last quad */
	struct dmaquad 	*last_link;	/* value of link in last quad */
	int		count;		/* Count from previous quad */
};

/*
 * value in link to indicate the end of a dma quad chain, the final pointer
 * should point to itself with the eoc (end of chain bit) set
 */
#define	CIO_EOC(address) ((struct dmaquad *) ((u_int) address | 0x1))

/*
 * alignment requirements for quads and data
 */
#define	CIO_QUAD_ALIGN		16
#define	CIO_BUFFER_ALIGN	64
#define	CIO_MIN_BUFFER		64

/*
 * CIO DMA commands
 */
#define	CIODMA_IDY	0x08000000	/* request id from device adapter */
#define	CIODMA_DLD	0x10000004	/* download device adapter */
#define	CIODMA_PAU	0x20000000	/* pause device adapter */
#define	CIODMA_DIS	0x30000000	/* disconnect the log/sub channel */
#define	CIODMA_RS	0x40000000	/* read status of device */
#define	CIODMA_WC	0x50000004	/* write control message to device */
#define	CIODMA_RD	0x60000000	/* read data from device */
#define	CIODMA_WD	0x70000004	/* write data to device */
#define	CIODMA_RTS	0x88000000	/* read transparent status */
#define	CIODMA_WTC	0x98000004	/* write transparent control */
#define CIODMA_RDS	0xA0000000	/* read download status */
#define	CIODMA_CLC	0xB8000004	/* connect logical channel */
#define CIODMA_LNKST    0x00000008	/* return status of dma operation */

/*
 * predicates to test for commands
 */
#define IS_CIODMA_IDY(command)   (((command) & 0xf8000000) == CIODMA_IDY)
#define IS_CIODMA_DLD(command)   (((command) & 0xf0000004) == CIODMA_DLD)
#define IS_CIODMA_PAU(command)   (((command) & 0xf0000000) == CIODMA_PAU)
#define IS_CIODMA_DIS(command)   (((command) & 0xf0000000) == CIODMA_DIS)
#define IS_CIODMA_RS(command)    (((command) & 0xf0000000) == CIODMA_RS)
#define IS_CIODMA_WC(command)    (((command) & 0xf0000004) == CIODMA_WC)
#define IS_CIODMA_RD(command)    (((command) & 0xf0000000) == CIODMA_RD)
#define IS_CIODMA_WD(command)    (((command) & 0xf0000004) == CIODMA_WD)
#define IS_CIODMA_RTS(command)   (((command) & 0xf8000000) == CIODMA_RTS)
#define IS_CIODMA_WTC(command)   (((command) & 0xf8000004) == CIODMA_WTC)
#define IS_CIODMA_RDS(command)   (((command) & 0xf0000000) == CIODMA_RDS)
#define IS_CIODMA_CLC(command)   (((command) & 0xf8000004) == CIODMA_CLC)
#define IS_CIODMA_LNKST(command) (((command) & 0xf0000008) == CIODMA_LNKST)


/*
 * For DLD and RDS commands, "or" in CIO_CODE to the command to download 
 * or read status about code transfers.
 */
#define CIO_CODE	0x01000000

/*
 * For RS command, "or" in CIO_DISC to the command to disconnect the channel
 * after the status is read.
 */
#define CIO_DISC	0x01000000

/*
 * For DLD, RS, WC, RD, WD, and RDS commands, "or" in CIO_BYTE to have the
 * operation occurs in byte mode. Byte mode should be used whenever the 
 * count is an odd value. Default mode is 16 bits.
 */
#define CIO_BYTE	0x08000000

/*
 * For LNKST command, "or" in CIO_INTER to generate an interrupt when the 
 * status linking is complete.
 */
#define CIO_INTR	0x00000080

/*
 * For all commands that read/write data "or" in CIO_CONT to cause the next
 * quad to be a scatter/gather continuation of this quad
 */
#define CIO_CONT	0x00000002


/*
 * CIO Level 1 Constants
 *
 * The device adapter can be accessed through the direct I/O space for the 
 * specific channel. These constants define values for the direct I/O 
 * operations.
 */

/*
 * read sense
 *
 * These bits are defined for the sense status obtain from the direct I/O space
 * from the channel adapter in the stlsens field.
 */
#define CIO_SENSE_ARQ	0x01		/* attention requested */
#define CIO_SENSE_ARE	0x04		/* attention request enabled */
#define CIO_SENSE_LV1	0x08		/* level 1 device present on channel */
#define CIO_SENSE_NMI	0x10		/* device is asserting NMI */
#define CIO_SENSE_PRE	0x20		/* non-level 1 device present */
#define CIO_SENSE_PST	0x40		/* passed self test */
#define CIO_SENSE_RFC	0x80		/* ready for command */

/*
 * write control operation
 *
 * These commands can be issued to the direct I/O space of the channel adapter
 * in the ctlsens field. Level 1 devices must issue appropriate C_BYT and C_END
 * lines.
 */
#define CIO_DA_CTRL_ARD		0x01	/* attention request disabled */
#define CIO_DA_CTRL_ARE		0x02	/* attention request enabled */
#define CIO_DA_CTRL_RQA		0x04	/* request atten when sense[rfc]=1 */
#define CIO_DA_CTRL_NMK		0x08	/* non-maskable int acknowledge */
#define CIO_DA_CTRL_DEN		0x10	/* device enabled */
#define CIO_DA_CTRL_DCL		0x20	/* device clear */
#define CIO_DA_CTRL_PEN		0x40	/* parity enable */

/*
 * CIO subchannel commands.
 *
 * These commands are used with the level 1 write_command, read_status and 
 * write_order commands to access level 2 device adapter functions.
 * You must "or" in the value of the subchannel that you want. These commands
 * are written to the direct I/O space of the channel in the statcmd field.
 */
#define	CIO_SUBC_CMD_DSC	0x00	/* destroy subchannel */
#define	CIO_SUBC_CMD_AEK	0x10	/* async event acknowledge */
#define	CIO_SUBC_CMD_CSC	0x20	/* connect subchannel */
#define CIO_SUBC_CMD_RSC	0x40	/* resume subchannel */
#define CIO_SUBC_CMD_ETA	0x50	/* ETA acknowledge */


/*
 * CIO subchannel status codes.
 *
 * These fields describe the status returned from the direct I/O space of a 
 * channel in the statcmd field. 
 */
#define CIO_SUBC_STATUS_RFC	0x00	/* ready for command */
#define CIO_SUBC_STATUS_AES	0x10	/* async event sensed */
#define CIO_SUBC_STATUS_DPE	0x20	/* data parity error */
#define CIO_SUBC_STATUS_SCD	0x30	/* subchannel destroyed */
#define CIO_SUBC_STATUS_SCR	0x50	/* subchannel request */
#define CIO_SUBC_STATUS_DOD	0x60	/* dead or dying */
#define CIO_SUBC_STATUS_PER	0x70	/* protocol error */
#define CIO_SUBC_STATUS_ERT	0x80	/* error trap */


/*
 * packet returned by CIODMA_IDY interrogation of the device adapter
 */
struct idy {
	unsigned char	card_id;
	unsigned char	firmware_id;
	unsigned short	firmware_rev;
	unsigned char	mpx:	2;
	unsigned char	psa:	1;
	unsigned char	rev:	3;
	unsigned char	mod:	2;
	unsigned char	ports_less_1;
	unsigned char	reqs_less_1;
	unsigned char	max_bsize_hi;
	unsigned char	max_bsize_lo;
};

/*
 * card IDY values for devices that we know about
 */
#define	CIOIDY_HPIB	2
#define CIOIDY_LAN	6
#define CIOIDY_MUX	7

/*
 * values of mpx field
 */
#define MPX_SUBONLY	0		/* subchannel multiplexing only */
#define MPX_LOGONLY	1		/* logchannel multiplexing only */
#define MPX_UNKNOWN	3		/* information not available */

/*
 * values of mod field
 */
#define MOD_WORDONLY	0		/* word mode transfers only */
#define MOD_BYTEONLY	1		/* byte mode transfers only */
#define MOD_WORDBYTE	2		/* both byte and word transfers */
