/* 
 * Copyright (c) 1991,1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: lanreg.h 1.2 91/09/25$
 * 	Author: Bob Wheeler, Leigh Stoller, University of Utah CSL
 */

/*
 * 	HP 27125 CIO LAN structures and constants
 */

#define CIOLAN_MAX_TU		(ETHERMTU + sizeof(struct ether_header))
#define CIOLAN_MIN_TU		(64 + sizeof(struct ether_header))

/*
 * values that can be or'ed into CIODMA_RS commands to get back various 
 * types of status
 */
#define CIOLAN_RS_ALL		0x00000000
#define CIOLAN_RS_INTERFACE	0x01000000
#define CIOLAN_RS_STATS		0x02000000
#define CIOLAN_RS_CONFIG	0x04000000

#define CIOLAN_RD_ARM		0x02000000

/*
 * lan header, same information as the first part of a lan_interface status
 */
struct lan_header {
	u_short  xmit_buffers;
	u_short  rec_buffers;
	u_short  status;
	struct   ether_header hdr;
	u_char   data[ETHERMTU];
};

#define CIOLAN_BUFFER_SIZE	(sizeof(struct lan_header))

/*
 * lan interface status
 */
struct lan_interface {
	u_short  xmit_buffers;
	u_short  rec_buffers;
	u_short  status;
	u_short  selftest;
	u_short  filters;
	u_char   hw_addr[6];
};

/*
 * interface exception status bits
 */
#define CIOLAN_IE_STF	0x8000
#define CIOLAN_IE_NOL	0x4000
#define CIOLAN_IE_PFB	0x0200
#define CIOLAN_IE_NTH	0x0100
#define CIOLAN_IE_E8	0x0080

/*
 * status bits in read header status
 */
#define CIOLAN_RH_ERR	0x0080
#define CIOLAN_RH_CRC	0x0040
#define CIOLAN_RH_FRF	0x0002
#define CIOLAN_RH_ODD	0x0001

/*
 * selftest/loopback status error bits
 */
#define CIOLAN_SS_LSF	0x8000
#define CIOLAN_SS_TSF	0x4000
#define CIOLAN_SS_NRF	0x2000
#define CIOLAN_SS_EIF	0x1000
#define CIOLAN_SS_BAB	0x0800
#define CIOLAN_SS_NHB	0x0400
#define CIOLAN_SS_MEM	0x0200
#define CIOLAN_SS_BTX	0x0100
#define CIOLAN_SS_U	0x0080
#define CIOLAN_SS_L	0x0040
#define CIOLAN_SS_LC	0x0020
#define CIOLAN_SS_R	0x0010
#define CIOLAN_SS_FF	0x0008
#define CIOLAN_SS_ORX	0x0004
#define CIOLAN_SS_CRC	0x0002
#define CIOLAN_SS_BRX	0x0001

/*
 * packet filters
 */
#define CIOLAN_FLT_P    0x0001		/* Promiscuous */
#define CIOLAN_FLT_B    0x0002		/* Broadcast */
#define CIOLAN_FLT_M	0x0004		/* Multicast */
#define CIOLAN_FLT_BAD  0x0010		/* Bad packets */

/*
 * configuration macros
 */
#define lan_go_offline(bp);				\
	do {						\
		*(bp)++ = 0x80;				\
		*(bp)++ = 0x01;				\
	   } while(0)

#define lan_go_online(bp);				\
	do {						\
		*(bp)++ = 0x80;				\
		*(bp)++ = 0x02;				\
	   } while(0)

#define lan_soft_reset(bp);				\
	do {						\
		*(bp)++ = 0x7f;				\
		*(bp)++ = 0xff;				\
	   } while(0)

#define lan_reset_stats(bp);				\
	do {						\
		*(bp)++ = 0x80;				\
		*(bp)++ = 0x00;				\
	   } while(0)

#define lan_reset_stats(bp);				\
	do {						\
		*(bp)++ = 0x80;				\
		*(bp)++ = 0x00;				\
	   } while(0)

#define lan_arm_receive(bp, limit, timeout);		\
	do {						\
		*(bp)++ = 0x00;				\
		*(bp)++ = 0x05;				\
		*(bp)++ = (limit);			\
		*(bp)++ = (timeout);			\
	   } while(0)

#define lan_loopback_test(bp);				\
	do {						\
		*(bp)++ = 0x7F;				\
		*(bp)++ = 0xFE;				\
	   } while(0)

#define lan_set_filter(bp, filter);			\
	do {						\
		*(bp)++ = 0x00;				\
		*(bp)++ = 0x02;				\
		*(bp)++ = 0x00;				\
		*(bp)++ = (filter);			\
	   } while(0)
