/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/* 
 * $RCSfile: ec596var.h,v $
 *
 * $Revision: 1.2 $
 * $Date: 1994/07/21 18:17:39 $
 */

#ifndef ec596var_h
#define ec596var_h

/*
 * The ETHERTYPE_NTRAILER packet types starting at ETHERTYPE_TRAIL
 * have (type-ETHERTYPE_TRAIL)*512 bytes of data followed by an ETHER
 * type (as given above) and then the (variable-length) header.
 */
#define	ETHERTYPE_TRAIL		0x1000	/* Trailer packet */
#define	ETHERTYPE_NTRAILER	0x0010

#define LAN_PROM_ADDR		0xf0810000	/* addr. of LAN PROM */
#define OFFSET_STATION_ADDR	0x0000		/* offset of network addr. */

#ifndef NET_ADD_MULTICAST_ADDRESS
# define NET_ADD_MULTICAST_ADDRESS ('n'<<16|3)
#endif
#ifndef NET_DELETE_MULTICAST_ADDRESS
# define NET_DELETE_MULTICAST_ADDRESS ('n'<<16|4)
#endif

/*
 * Each interface is described by one of the following structures,
 * which routing code uses to locate it.
 *
 * Ethernet Controller Software Interface
 */
typedef struct {
    struct ifnet ec_if;		   /* Common Ethernet structures */
    lock_data_t	init_lock;	   /* Controller-initialization lock */
    lock_data_t	q_lock;		   /* Non-queued controller access lock */
    lock_data_t	q_lock_init;	   /* Saved initial value of "q_lock", */
				   /*	pointing to dynamically-allocated */
				   /*	"simple_lock" (cf. "ecinit()" below) */
    /* "Transmit" completion status */
    struct {
	unsigned long late_collision;
	unsigned long no_carrier_sense;
	unsigned long lost_clear_to_send;
	unsigned long dma_underrun;
	unsigned long deferred;
	unsigned long too_many_collisions;
    } xmit_status;
    struct {			   /* Reasons to reinitialize the chip */
	int scb_not_ready;	   /* # times SCB not ready */
	int scb_corrupted;	   /* # times SCB status word corrupted */
	int tcb_corrupted;	   /* # times TCB forward link corrupted */
	int nop_completed;	   /* # times NOP CB completed */
	int xmit_timeout;	   /* # times transmit hung */
	int rfd_not_completed;	   /* # times RFD not marked completed */
	int ru_not_ready;	   /* # times RU not ready */
    } reason;
    rfd_t *rfd_head, *rfd_tail;	   /* ->RFD list */
    rbd_t *rbd_head, *rbd_tail;	   /* ->RBD list */
    tcb_t *tcb_head, *tcb_tail;	   /* ->TCB list */
    ecsc_t *ecsc;		   /* ->data cache aligned information */
    unsigned long signature;	   /* 82596 ROM signature (for chip version) */
    unsigned long dogwalks;	   /* # watchdog-timer intervals */
    unsigned long ru_noresources;  /* # times RU ran out of buffers */
    int nmc;			   /* # valid entries in multi-cast list */
    int request_tdr;		   /* Requests for TDR */
    int retransmit;		   /* Force retransmission of this packet */
    int nretransmit;		   /* # times we did it */
    int ufotransmit;		   /* # times xmit no OK nor error bits set */
    int txtimestamp;		   /* Latest tx activity internal clk value */
    int txtimelap;		   /* # internal clock ticks to wait */
    tbd_t tbd_list[NTCBBUF][NTBDBUF]; /* TBD list for each TCB */
    swap_t rscerrs;		   /* Latest watch-dog resource error count */
    swap_t ovrnerrs;		   /* Latest watch-dog over-run error count */
    short if_flags;		   /* Chip configuration, among: */
#define IFF_CHIPCONFIG (IFF_LOOPBACK|IFF_PROMISC|IFF_ALLMULTI)
    eaddr_t ec_addr;		   /* Hardware Ethernet address */
    eaddr_t mc[NMULTICAST];	   /* Packed multi-cast address list */
    unsigned short
      	      initialized   :1,	   /* A flag to say we are up to it */
	      rxtimeout	    :1,	   /* Receiver restart-timer running */
	      cbtimeout	    :1,	   /* "eccbstart()" time-out */
	      cbabort	    :1,	   /* "eccbstart()" abort */
	      thin_lan	    :1,	   /* 0/1 <=> thick/thin LAN */
	      scberrs	    :1,	   /* <=> Observing error counters? */
	      force_selftest:1;	   /* <=> Self-test requested? */
    volatile boolean_t txactive;   /* Transmitter active */
    volatile boolean_t q_op;	   /* Queued operation mode */
    volatile boolean_t txsetup;	   /* "Tranmit"s are being set up */
    boolean_t recover;		   /* Request watchdog to recover interface */
    warning_t last_warning;	   /* Last warning printed */
} ecsi_t;

#endif /* ec596var_h */
