/* 
 * Copyright (c) 1990,1991,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: md.c 1.6 94/12/14$
 *	Author: Mike Hibler, University of Utah CSL
 */

/*
 * Way-simple memory disk driver.
 * This file should move to some machine-independent dir.
 */
#include <md.h>
#if NMD > 0

#if	MACH_KERNEL
#include <sys/types.h>
#include <device/param.h>
#include <device/errno.h>
#include <device/buf.h>
#include <device/conf.h>
#else
#include "param.h"
#include "systm.h"
#include "errno.h"
#include "buf.h"
#include "conf.h"
#include "machine/vmparam.h"
#include "machine/pte.h"
#endif

#include <machine/mdvar.h>	/* really belongs in some MI directory */

#include <boot_ufs/defs.h>	/* XXX for FS check in mdinit */
#include <boot_ufs/fs.h>	/* XXX for FS check in mdinit */

#ifndef DEBUG
#define DEBUG
#endif

#ifdef DEBUG
int mddebug = 0x00;
#define MDB_FOLLOW	0x01
#define MDB_INIT	0x02
#define MDB_IO		0x04
#define MDB_DATA	0x08
#endif

struct	buf mdbuf[NMD];

#define	mdunit(x)	((minor(x) >> 3) & 0x7)	/* for consistency */

struct md_softc {
	int		 sc_flags;		/* flags */
	caddr_t		 sc_vaddr;		/* kernel VA start of md */
	daddr_t		 sc_size;		/* size of md (blocks) */
} md_softc[NMD];

#if MACH_KERNEL
vm_offset_t mdbase;
#endif

/* sc_flags */
#define	MDF_ALIVE	0x01
#define MDF_INITED	0x02

mdinit(lowpage)
	int lowpage;
{
	int unit = 0;			/* XXX only one for now */
	register struct md_softc *ms = &md_softc[unit];
	register int i;

#if MACH_KERNEL
	register vm_offset_t va, pa;
	extern vm_map_t kernel_map;

#ifdef DEBUG
	if (mddebug & MDB_INIT)
		printf("mdinit: attempt %x bytes at %x\n",
		       MDMAXBYTES, mdbase);
#endif
	pa = mdbase;
	if (pa == 0)
		return(0);
	va = mdbase = kmem_alloc_pageable(kernel_map, MDMAXBYTES);
	if (va == 0)
		return(0);
#ifdef DEBUG
	if (mddebug & MDB_INIT)
		printf("mdinit: allocated va at %x\n", va);
#endif
	for (i = 0; i < MDMAXBYTES; i += PAGE_SIZE) {
		pmap_enter(pmap_kernel(), va, pa,
			 VM_PROT_READ|VM_PROT_WRITE, TRUE);
		va += PAGE_SIZE;
		pa += PAGE_SIZE;
	}
#ifdef DEBUG
	if (mddebug & MDB_INIT)
		printf("mdinit: mapped\n");
#endif
#else
	register int j, base;
	int vpage;
	extern int freemem;
	extern struct pte mdpt[];

	/* XXX let system have a little memory */
	lowpage += btoc(MDLEAVEBYTES);

	i = btoc(MDMAXBYTES);
	if (lowpage + i > physmem)
		return(0);

	physmem -= i;
	freemem = maxmem = physmem;

	vpage = btoc(mdbase);
	for (j = 0, base = physmem; j < i; j++, base++) {
#ifdef hp300
		*(int *)(&mdpt[j]) = PG_V | PG_KW | ctob(base);
#else
		*(int *)(&mdpt[j]) = PG_V | PG_KW | base;
#endif
#ifdef hp9000s800
		validate(ctob(vpage++), KERNELSPACE, 0,
			 PDE_AR_KRW, (unsigned) base);
#endif
	}
	i = ctob(i);
#endif

	ms->sc_vaddr = mdbase;
	ms->sc_size = btodb(i);
	printf("md%d: %d bytes at %x, ",
	       unit, dbtob(ms->sc_size), ms->sc_vaddr);
	if (((struct fs *)(mdbase+BBSIZE))->fs_magic == FS_MAGIC)
		printf("has");
	else
		printf("no");
	printf(" filesystem\n");
	ms->sc_flags = MDF_ALIVE | MDF_INITED;
	return(1);
}

mdopen(dev, flags)
	dev_t dev;
{
	int unit = mdunit(dev);
	register struct md_softc *ms = &md_softc[unit];

#ifdef DEBUG
	if (mddebug & MDB_FOLLOW)
		printf("mdopen(%x, %x)\n", dev, flags);
#endif
	if (unit >= NMD || (ms->sc_flags & MDF_ALIVE) == 0)
		return(ENXIO);
	return(0);
}

mdclose(dev)
	dev_t dev;
{
#ifdef DEBUG
	if (mddebug & MDB_FOLLOW)
		printf("mdclose(%x)\n", dev);
#endif
	return(0);
}

mdstrategy(bp)
	register struct buf *bp;
{
	register int unit = mdunit(bp->b_dev);
	register struct md_softc *ms = &md_softc[unit];
	register caddr_t base;
	register daddr_t bn;
	register int sz;

#ifdef DEBUG
	if (mddebug & (MDB_FOLLOW|MDB_IO|MDB_DATA))
		printf("mdstrategy(%x): u%d b%d c%d a%x\n",
		       bp, unit, bp->b_blkno, bp->b_bcount, bp->b_un.b_addr);
#endif

	if ((ms->sc_flags & MDF_INITED) == 0) {
		bp->b_error = ENXIO;
		bp->b_flags |= B_ERROR;
		goto done;
	}
	bn = bp->b_blkno;
	sz = (bp->b_bcount + DEV_BSIZE - 1) / DEV_BSIZE;
	if (bn < 0 || bn + sz > ms->sc_size) {
		sz = ms->sc_size - bn;
		if (sz == 0) {
			bp->b_resid = bp->b_bcount;
			goto done;
		}
		if (sz < 0) {
			bp->b_error = EINVAL;
			bp->b_flags |= B_ERROR;
			goto done;
		}
		bp->b_bcount = dbtob(sz);
	}
	/*
	 * Copy the data.
	 * XXX raise spl here?  I think not.
	 * XXX instrumentation.
	 */
	base = ms->sc_vaddr + dbtob(bp->b_blkno);

#ifdef DEBUG
	if (mddebug & MDB_IO)
		printf(" bcopy %x -> %x\n",
		       (bp->b_flags&B_READ) ? base : bp->b_un.b_addr,
		       (bp->b_flags&B_READ) ? bp->b_un.b_addr : base);
#endif

	if (bp->b_flags & B_READ)
		bcopy(base, bp->b_un.b_addr, bp->b_bcount);
	else
		bcopy(bp->b_un.b_addr, base, bp->b_bcount);
#ifdef DEBUG
	if (mddebug & MDB_DATA) {
		long *lp;

		if (bp->b_flags & B_READ)
			base = bp->b_un.b_addr;
		lp = (long *)base;
		printf(" %x(%x): %x %x %x %x %x %x %x %x\n",
		       base, pmap_extract(pmap_kernel(), base),
		       lp[0], lp[1], lp[2], lp[3], lp[4], lp[5], lp[6], lp[7]);
	}
#endif
	bp->b_resid = 0;
done:
	iodone(bp);

#ifdef DEBUG
	if (mddebug & MDB_IO) {
		printf("mdstrategy: done");
		if (bp->b_flags & B_ERROR)
			printf(", error = %d", bp->b_error);
		printf("\n");
	}
#endif
}

mdread(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	register int unit = mdunit(dev);

#ifdef DEBUG
	if (mddebug & MDB_FOLLOW)
		printf("mdread(%x, %x)\n", dev, uio);
#endif
	return(physio(mdstrategy, &mdbuf[unit], dev, B_READ, minphys, uio));
}

mdwrite(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	register int unit = mdunit(dev);

#ifdef DEBUG
	if (mddebug & MDB_FOLLOW)
		printf("mdwrite(%x, %x)\n", dev, uio);
#endif
	return(physio(mdstrategy, &mdbuf[unit], dev, B_WRITE, minphys, uio));
}

/*ARGSUSED*/
mdioctl(dev, cmd, data, flag)
	dev_t dev;
	int cmd;
	caddr_t data;
	int flag;
{
#if	MACH_KERNEL
#else
	return(EINVAL);
#endif
}

#if	MACH_KERNEL
#else
mdsize(dev)
	dev_t dev;
{
	int unit = mdunit(dev);
	register struct md_softc *ms = &md_softc[unit];

	if (unit >= NMD || (ms->sc_flags & MDF_INITED) == 0)
		return(-1);
	return(ms->sc_size);
}

/*ARGSUSED*/
mddump(dev)
{
	/* yeah, right! */
	return(ENXIO);
}
#endif
#endif
