/* 
 * Copyright (c) 1990,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: _setjmp.s 1.8 94/12/14$
 */

#include <mach/machine/asm.h>

	.space	$TEXT$
	.subspa $CODE$

/*
 * The PA jmp_buf is 48 words arranged as follows:
 *
 *	 0- 9:	sigcontext
 *	10-26:	callee save GRs (r3-r18) and DP (r27)
 *	   27:	callee save SRs (sr3)
 *	28-47:	callee save FRs (fr12-fr21)
 */

/*
 * int
 * setjmp(env)
 *	jmp_buf env;
 * 
 * This routine does not restore signal state.
 */

	.export	setjmp,entry
	.export	_setjmp,entry
	.proc
	.callinfo
setjmp
_setjmp
	.entry

	/*
	 * save sp and rp in sigcontext, skip the rest
	 */
	stw	sp,8(arg0)
	stw	rp,24(arg0)
	ldo	40(arg0),arg0

	/*
	 * save dp and the callee saves registers
	 */
	stwm	r3,4(arg0)
	stwm	r4,4(arg0)
	stwm	r5,4(arg0)
	stwm	r6,4(arg0)
	stwm	r7,4(arg0)
	stwm	r8,4(arg0)
	stwm	r9,4(arg0)
	stwm	r10,4(arg0)
	stwm	r11,4(arg0)
	stwm	r12,4(arg0)
	stwm	r13,4(arg0)
	stwm	r14,4(arg0)
	stwm	r15,4(arg0)
	stwm	r16,4(arg0)
	stwm	r17,4(arg0)
	stwm	r18,4(arg0)
	stwm	r27,4(arg0)

	mfsp	sr3,t1
	stwm	t1,4(arg0)

	/*
	 * If the address pointed to by arg0 is aligned then this is easy
	 */
	bb,>=,n	arg0,29,$setjmp_aligned

	/*
	 * this is harder, We know that sp is currently aligned to a double
	 * word, use this double word as a temporary area
	 */
	fstds	fr12,0(sp)
	ldw	0(sp),t1
	stwm	t1,4(arg0)
	ldw	4(sp),t1
	stwm	t1,4(arg0)

	fstds	fr13,0(sp)
	ldw	0(sp),t1
	stwm	t1,4(arg0)
	ldw	4(sp),t1
	stwm	t1,4(arg0)

	fstds	fr14,0(sp)
	ldw	0(sp),t1
	stwm	t1,4(arg0)
	ldw	4(sp),t1
	stwm	t1,4(arg0)

	fstds	fr15,0(sp)
	ldw	0(sp),t1
	stwm	t1,4(arg0)
	ldw	4(sp),t1
	stwm	t1,4(arg0)

	fstds	fr16,0(sp)
	ldw	0(sp),t1
	stwm	t1,4(arg0)
	ldw	4(sp),t1
	stwm	t1,4(arg0)

	fstds	fr17,0(sp)
	ldw	0(sp),t1
	stwm	t1,4(arg0)
	ldw	4(sp),t1
	stwm	t1,4(arg0)

	fstds	fr18,0(sp)
	ldw	0(sp),t1
	stwm	t1,4(arg0)
	ldw	4(sp),t1
	stwm	t1,4(arg0)

	fstds	fr19,0(sp)
	ldw	0(sp),t1
	stwm	t1,4(arg0)
	ldw	4(sp),t1
	stwm	t1,4(arg0)

	fstds	fr20,0(sp)
	ldw	0(sp),t1
	stwm	t1,4(arg0)
	ldw	4(sp),t1
	stwm	t1,4(arg0)

	fstds	fr21,0(sp)
	ldw	0(sp),t1
	stwm	t1,4(arg0)
	ldw	4(sp),t1
	b	$setjmp_exit
	stwm	t1,4(arg0)

$setjmp_aligned
	fstds,ma fr12,8(arg0)
	fstds,ma fr13,8(arg0)
	fstds,ma fr14,8(arg0)
	fstds,ma fr15,8(arg0)
	fstds,ma fr16,8(arg0)
	fstds,ma fr17,8(arg0)
	fstds,ma fr18,8(arg0)
	fstds,ma fr19,8(arg0)
	fstds,ma fr20,8(arg0)
	fstds    fr21,0(arg0)

$setjmp_exit
	bv	0(rp)
	copy	r0,ret0
	.exit
	.procend



/*
 * void 
 * longjmp(env, val)
 *	jmp_buf env;
 *	int val;
 * 
 * This routine does not retore signal state.
 * This routine does not override a zero val.
 */

	.export	longjmp,entry
	.export	_longjmp,entry
	.proc
	.callinfo
longjmp
_longjmp

	.entry

	/*
	 * restore sp and rp
	 */
	ldw	8(arg0),sp
	ldw	24(arg0),rp
	ldo	40(arg0),arg0

	/*
	 * restore callee saves registers
	 */
	ldwm	4(arg0),r3
	ldwm	4(arg0),r4
	ldwm	4(arg0),r5
	ldwm	4(arg0),r6
	ldwm	4(arg0),r7
	ldwm	4(arg0),r8
	ldwm	4(arg0),r9
	ldwm	4(arg0),r10
	ldwm	4(arg0),r11
	ldwm	4(arg0),r12
	ldwm	4(arg0),r13
	ldwm	4(arg0),r14
	ldwm	4(arg0),r15
	ldwm	4(arg0),r16
	ldwm	4(arg0),r17
	ldwm	4(arg0),r18
	ldwm	4(arg0),r27

	ldwm	4(arg0),t1
	mtsp	t1,sr3

	/*
	 * If the address pointed to by arg0 is aligned then this is easy
	 */
	bb,>=,n	arg0,29,$longjmp_aligned

	/*
	 * this is harder, We know that sp is currently aligned to a double
	 * word, use this double word as a temporary area
	 */

	ldwm	4(arg0),t1
	stw	t1,0(sp)
	ldwm	4(arg0),t1
	stw	t1,4(sp)
	fldds	0(sp),fr12

	ldwm	4(arg0),t1
	stw	t1,0(sp)
	ldwm	4(arg0),t1
	stw	t1,4(sp)
	fldds	0(sp),fr13

	ldwm	4(arg0),t1
	stw	t1,0(sp)
	ldwm	4(arg0),t1
	stw	t1,4(sp)
	fldds	0(sp),fr14

	ldwm	4(arg0),t1
	stw	t1,0(sp)
	ldwm	4(arg0),t1
	stw	t1,4(sp)
	fldds	0(sp),fr15

	ldwm	4(arg0),t1
	stw	t1,0(sp)
	ldwm	4(arg0),t1
	stw	t1,4(sp)
	fldds	0(sp),fr16

	ldwm	4(arg0),t1
	stw	t1,0(sp)
	ldwm	4(arg0),t1
	stw	t1,4(sp)
	fldds	0(sp),fr17

	ldwm	4(arg0),t1
	stw	t1,0(sp)
	ldwm	4(arg0),t1
	stw	t1,4(sp)
	fldds	0(sp),fr18

	ldwm	4(arg0),t1
	stw	t1,0(sp)
	ldwm	4(arg0),t1
	stw	t1,4(sp)
	fldds	0(sp),fr19

	ldwm	4(arg0),t1
	stw	t1,0(sp)
	ldwm	4(arg0),t1
	stw	t1,4(sp)
	fldds	0(sp),fr20

	ldwm	4(arg0),t1
	stw	t1,0(sp)
	ldwm	4(arg0),t1
	stw	t1,4(sp)
	b	$longjmp_exit
	fldds	0(sp),fr21

$longjmp_aligned
	fldds,ma 8(arg0),fr12
	fldds,ma 8(arg0),fr13
	fldds,ma 8(arg0),fr14
	fldds,ma 8(arg0),fr15
	fldds,ma 8(arg0),fr16
	fldds,ma 8(arg0),fr17
	fldds,ma 8(arg0),fr18
	fldds,ma 8(arg0),fr19
	fldds,ma 8(arg0),fr20
	fldds	0(arg0),fr21

$longjmp_exit
	bv	0(rp)
	copy	arg1,ret0
	.exit
	.procend
