/*
 * Mach Operating System
 * Copyright (c) 1989 Carnegie Mellon University.
 * Copyright (c) 1990,1991,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON, THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF
 * THIS SOFTWARE IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY
 * OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF
 * THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 *
 * 	Utah $Hdr: thread.c 1.7 94/12/03$
 */

#include <mach/cthreads.h>
#include "cthread_internals.h"
#include <mach.h>

vm_size_t cthread_stack_size = 0x40000;
vm_address_t cthread_stack_base = 0x50000000;	/* XXX leave room for a heap */

/*
 * Set up the initial state of a MACH thread so that it will invoke
 * routine(child) when it is resumed.
 */
void
cproc_setup(register cproc_t child, thread_t thread, void (*routine)(cproc_t))
{
	struct parisc_thread_state state;
	kern_return_t r;
	unsigned size;
	extern int get_dp();

	/*
	 * Set up parisc call frame and registers.
	 */
	size = PARISC_THREAD_STATE_COUNT;
	MACH_CALL(thread_get_state(thread, PARISC_THREAD_STATE, 
				   (thread_state_t) &state, &size), r);

	/*
	 * set the PC queue to point to routine.
	 */
	state.iioq_head = (int) routine;
	state.iioq_tail = (int) routine + 4;

	/*
	 * setup the first argument to routine to be the address of child.
	 */
	state.r26 = (int) child;

	/*
	 * establish a user stack for this thread. There is no guarentee
	 * that the stack is aligned correctly. We have to align it to
	 * a double word boundary.
	 *
	 * stack_base doesn't account for self-pointer, hence "+4"
	 */
	state.r30 = child->stack_base + 4 + 7 & 0xfffffff8;

	/*
	 * clear off the first 48 bytes of the stack, this is going to be 
	 * our fake stack marker, we don't want any routine to think there
	 * is a stack frame below us.
	 */
	bzero(state.r30, 48);

	/*
	 * now set the stack pointer 48 bytes deeper, 32 for a frame marker 
	 * and 16 for the standard 4 arguments.
	 */
	state.r30 += 48;

	/*
	 * setup the dp register to be the same as ours
	 */
	state.r27 = get_dp();

	/*
	 * Now we have to initialize the floating point status registers
	 * since they will be in a random state and could cause a trap.
	 */
	state.fr0 = 0.0;
	state.fr1 = 0.0;
	state.fr2 = 0.0;
	state.fr3 = 0.0;

	/*
	 * the psw is fine it is set up as CQPDI by the kernel. Set the
	 * registers into the saved state of the thread.
	 */
	MACH_CALL(thread_set_state(thread, PARISC_THREAD_STATE,
				   (thread_state_t) &state,
				   PARISC_THREAD_STATE_COUNT),r);
}

#if 0
/* defined in lock.s */
int
cthread_sp()
{
	int x;

	return (int) &x;
}
#endif
