/*
 * Mach Operating System
 * Copyright (c) 1994-1987 Carnegie Mellon University.
 * Copyright (c) 1993,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON, THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF
 * THIS SOFTWARE IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY
 * OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF
 * THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */
/*
 *	File:	thread.h
 *	Author:	Avadis Tevanian, Jr.
 *
 *	This file contains the structure definitions for threads.
 *
 */

#ifndef	_KERN_THREAD_H_
#define _KERN_THREAD_H_

#include <mach_ipc_compat.h>
#include <hw_footprint.h>
#include <mach_fixpri.h>
#include <mach_host.h>
#include <net_atm.h>
#include <mach/boolean.h>
#include <mach/thread_info.h>
#include <mach/thread_status.h>
#include <mach/machine/vm_types.h>
#include <mach/message.h>
#include <mach/port.h>
#include <mach/vm_prot.h>
#include <kern/ast.h>
#include <kern/cpu_number.h>
#include <kern/queue.h>
#include <kern/pc_sample.h>
#include <kern/processor.h>
#include <kern/sched_prim.h>	/* event_t, continuation_t */
#include <mach/message.h> /* XXX */
#include <mach/vm_param.h>
#include <kern/queue.h> /* XXX */
#include <kern/time_out.h>
#include <kern/timer.h>
#include <kern/time_out.h>
#include <kern/lock.h>
#include <kern/sched.h>
#include <kern/task.h>		/* for current_space(), current_map() */
#include <kern/shuttle.h>
#include <kern/act.h> /* XXX */
#include <ipc/ipc_kmsg_queue.h>

#include "thread.h"/*XXX*/

struct pmap;
struct Act;

struct thread {
	struct Shuttle	shuttle;

	struct_id_decl;			/* structure ID for sanity checking */


	/* Blocking information */
	int		wait_event;	/* event we are waiting on */
	kern_return_t	wait_result;	/* outcome of wait -
					   may be examined by this thread
					   WITHOUT locking */
	boolean_t	wake_active;	/* someone is waiting for this
					   thread to become suspended */



	/* Self-preservation */
	decl_simple_lock_data(,lock)
	int		ref_count;	/* number of references to me */

	int		state;		/* Current state: */
#define TH_WAIT			0x01
#define TH_RUN			0x04
	/* TH_UNINT should be handled by an uninterruptible wait() simply
	   ignoring interrupt requests. */
#define TH_UNINT		0x08	/* thread is waiting uninteruptibly */
#define TH_IDLE			0x80	/* thread is an idle thread */

#define	TH_SCHED_STATE	(TH_WAIT|TH_RUN|TH_UNINT)

	/* General-purpose pointer field whose use depends on what the
	   thread happens to be doing */
	void		*message;

	/* Topmost act for this thread */
	struct Act	*top_act;

	/* Machine-dependent thread state */
	MachineThread	mthread;



	/* XXX information specific to the traditional scheduler */

	/* Thread bookkeeping */
	queue_chain_t	pset_threads;	/* list of all threads in proc set*/

	/* Scheduling information */
	int		priority;	/* thread's priority */
	int		max_priority;	/* maximum priority */
	int		sched_pri;	/* scheduled (computed) priority */
#if	MACH_FIXPRI
	int		sched_data;	/* for use by policy */
	int		policy;		/* scheduling policy */
#endif	/* MACH_FIXPRI */
	int		depress_priority; /* depressed from this priority */
	unsigned int	cpu_usage;	/* exp. decaying cpu usage [%cpu] */
	unsigned int	sched_usage;	/* load-weighted cpu usage [sched] */
	unsigned int	sched_stamp;	/* last time priority was updated */

	/* Timing data structures */
	timer_data_t	user_timer;	/* user mode timer */
	timer_data_t	system_timer;	/* system mode timer */
	timer_save_data_t user_timer_save;  /* saved user timer value */
	timer_save_data_t system_timer_save;  /* saved sys timer val. */
	unsigned int	cpu_delta;	/* cpu usage since last update */
	unsigned int	sched_delta;	/* weighted cpu usage since update */

	/* Time-outs */
	timer_elt_data_t timer;		/* timer for thread */
	timer_elt_data_t depress_timer;	/* timer for priority depression */

	/* Processor data structures */
	processor_set_t	processor_set;	/* assigned processor set */
	processor_t	bound_processor;	/* bound to processor ?*/
#if	MACH_HOST
	boolean_t	may_assign;	/* may assignment change? */
	boolean_t	assign_active;	/* someone waiting for may_assign */
#endif	MACH_HOST

#if	NCPUS > 1
	processor_t	last_processor; /* processor this last ran on */
#if	MACH_LOCK_MON
	unsigned	lock_stack;	/* number of locks held */
#endif  MACH_LOCK_MON
#endif	NCPUS > 1



	/* XXX get rid of this by using lighter-weight primitives */

	/* IPC data structures */
	struct thread *ith_next, *ith_prev;
	mach_msg_return_t ith_state;
	union {
		mach_msg_size_t msize;		/* max size for recvd msg */
		struct ipc_kmsg *kmsg;		/* received message */
	} data;
	mach_port_seqno_t ith_seqno;		/* seqno of recvd message */

	/* This queue is used only when destroying messages:
	   it prevents nasty recursion problems when destroying one message
	   causes other messages to be destroyed.
	   This queue should always be empty under normal circumstances.
	   See ipc_kmsg_destroy() for more details.  */
	struct ipc_kmsg_queue ith_messages;



	/* XXX continuation crap that can be gotten rid of soon.  */

	/* State saved when thread's stack is discarded */
	union {
		struct {
			struct ipc_port *port;
			int exc;
			int code;
			int subcode;
		} exception;
		void *other;		/* catch-all for other state */
	} saved;



	/* XXX The rest is goo that I'm not sure yet what to do with.  */

	/* VM global variables */

	boolean_t	vm_privilege;	/* Can use reserved memory? */

	decl_simple_lock_data(, ith_lock_data)

	mach_port_t ith_mig_reply;	/* reply port for mig */
	struct ipc_port *ith_rpc_reply;	/* reply port for kernel RPCs */

#if	NET_ATM
	nw_ep_owned_t   nw_ep_waited;
#endif	/* NET_ATM */
#if	MACH_PROF
	/* Profiling data structures */
	boolean_t	thread_profiled;/* is thread being profiled ? */
	boolean_t	thread_profiled_own;
					/* is thread being profiled 
					 * on its own ? */
	struct prof_data *profil_buffer;/* prof struct if so */
#endif	MACH_PROF

	/* Ast/Halt data structures */
	boolean_t	active;		/* how alive is the thread */
	int		ast;    	/* ast's needed.  See ast.h */
};

/* typedef of thread_t is in kern/kern_types.h */
typedef struct thread_shuttle	*thread_shuttle_t;
#define THREAD_NULL		((thread_t) 0)
#define THREAD_SHUTTLE_NULL	((thread_shuttle_t)0)

#define	ith_msize	data.msize
#define	ith_kmsg	data.kmsg
#define	ith_wait_result	wait_result

#define	ith_port	saved.exception.port
#define ith_exc		saved.exception.exc
#define ith_exc_code	saved.exception.code
#define ith_exc_subcode	saved.exception.subcode

#define ith_other	saved.other

#ifndef	_KERN_KERN_TYPES_H_
typedef struct thread *thread_t;

#define THREAD_NULL	((thread_t) 0)

typedef	mach_port_t *thread_array_t;
#endif	/* _KERN_KERN_TYPES_H_ */


extern thread_t		active_threads[NCPUS];	/* active threads */
extern vm_offset_t	active_stacks[NCPUS];	/* active kernel stacks */

#ifdef KERNEL
/*
 *	User routines
 */

extern kern_return_t	thread_create(
	task_t		parent_task,
	Act		**new_act);
extern kern_return_t	thread_info(
	thread_t	thread,
	int		flavor,
	thread_info_t	thread_info_out,
	natural_t	*thread_info_count);
extern kern_return_t	thread_assign(
	thread_t	thread,
	processor_set_t	new_pset);
extern kern_return_t	thread_assign_default(
	thread_t	thread);
extern void		thread_deallocate(
	thread_t	thread);
#endif

#define		thread_lock(th)		simple_lock(&(th)->lock)
#define		thread_unlock(th)	simple_unlock(&(th)->lock)


extern void		reaper_thread(void);

void		thread_freeze();
void		thread_doassign();
void		thread_unfreeze();


#ifdef	DEBUG
void		thread_verify(struct thread *thread);
#else	!DEBUG
#define		thread_verify(thread)
#endif	!DEBUG


/* Exported to kernel to create kernel threads */
thread_t	kernel_thread();

/* Exported to startup.c */
void		thread_init(void);
void		reaper_thread();

/* Machine-dependent code must define the following functions */
void		thread_machine_init();
kern_return_t	thread_machine_create(thread_t thread, Act *act, __dead void (*start_position)());
void		thread_machine_destroy(thread_t thread);
kern_return_t   thread_machine_set_state();
kern_return_t   thread_machine_get_state();

#endif	/* _KERN_THREAD_H_ */
